from __future__ import division

import json
import logging
import pprint
import sys
from itertools import chain

import os
import re
from matplotlib import cm

import effusion
from effusion import *


def build():
    func_name = 'build'

    effusion.dataset.set_phase_and_role(effusion.args.phase, 'training')

    query_uniprot = None
    if effusion.args.query_uniprot:
        query_uniprot = effusion.args.query_uniprot
    if effusion.args.task_number is not None:
        if query_uniprot is not None:
            raise ValueError('Query uniprot specified multiple ways')

        if effusion.args.uniprots is not None:
            query_uniprots = open(effusion.args.uniprots).read().split()
        else:
            query_uniprots = effusion.dataset.evaluation_proteins
        query_uniprot = query_uniprots[effusion.args.task_number]
    assert query_uniprot is not None

    build_func = None
    if effusion.args.build_method == 'widenet':
        build_func = effusion.ProteinNetwork.build_widenet
    elif effusion.args.build_method == 'direct_net':
        build_func = effusion.ProteinNetwork.build_direct_net


    # do not load prebuilt or save nets (for calculating stats)
    # g = build_func(
    #     query_uniprot,
    #     net_cache=None,
    #     do_string=False)

    net_filename = os.path.join(effusion.dataset.networks_dir, query_uniprot + '.ssfa.json.gz')
    g = build_func(
        query_uniprot,
        net_cache=net_filename,
        do_string=True)

    # change to evaluation role. we're putting the answers in the network
    if not effusion.args.no_write_visual_files:
        effusion.dataset.set_phase_and_role('discovery', 'evaluation')
        annotations_dict = effusion.Annotation.get_annotations_for_uniprots(g.vs['name'], include_implied=False)
        g.set_evidence(annotations_dict)
        g.export_cyjs(query_uniprot + '.cyjs')


def predict():
    func_name = 'predict'

    effusion.dataset.set_phase_and_role(effusion.args.phase, 'training')

    query_uniprot = None
    if effusion.args.query_uniprot:
        query_uniprot = effusion.args.query_uniprot
    if effusion.args.task_number is not None:
        if query_uniprot is not None:
            raise ValueError('query uniprot specified multiple ways')

        if effusion.args.uniprots is not None:
            query_uniprots = open(effusion.args.uniprots).read().split()
        else:
            query_uniprots = effusion.dataset.evaluation_proteins
        query_uniprot = query_uniprots[effusion.args.task_number]
    assert query_uniprot is not None

    module_path, class_name = effusion.args.prediction_class.rsplit('.', 1)
    prediction_class = sys.modules[module_path].__dict__[class_name]
    method = prediction_class(dataset=effusion.dataset)

    experiment_preds_path = os.path.join(effusion.dataset.preds_dir, class_name)
    if effusion.args.experiment_name:
        experiment_preds_path += '.' + effusion.args.experiment_name

    raw_preds = method.predict_one(query_uniprot)

    try:
        os.mkdir(experiment_preds_path)
    except OSError as e:
        if e.errno == 17:  # File exists
            pass
        else:
            raise
    with open(os.path.join(experiment_preds_path, query_uniprot + '.preds.json'), 'w') as f:
        json.dump(raw_preds, f, indent=4, default=lambda o: o.__dict__)

    postprocess_args = {
        'subset_type': 'molecular_function',
        'do_filter_purified': False,
        'do_break_ties': True,
        'prop_down_type': None,  # 'smart_prop_down',
        'max_predictions': 200

    }  # dataset, uniprot, raw_preds, subset_type, do_filter_purified, do_break_ties, prop_down_type, max_predictions
    preds = effusion.postprocess(effusion.dataset,
                                 query_uniprot,
                                 raw_preds,
                                 **postprocess_args)
    for p in preds:
        print str(p), p.prob

    collecting_query_info = False
    if collecting_query_info == True:
        effusion.dataset.set_phase_and_role(effusion.args.phase, 'evaluation')
        annotations = effusion.Annotation.get_goas_for_uniprot(uniprot=query_uniprot,
                                                               include_negative=False,
                                                               include_direct=True,
                                                               include_implied=True,
                                                               include_moonlight=False)
        true_ontology_mf_ids = frozenset({a.go_id for a in annotations if effusion.dataset.gene_ontology.term4id(
            a.go_id).aspect == effusion.GoAspectEnum.molecular_function})
        true_ontology_mf_subgraph = effusion.dataset.gene_ontology.subgraph(true_ontology_mf_ids,
                                                                            keep_ancestors=True)  # can't imagine keep_ancestors is useful here
        method.current_query_info['true_ontology_mf_ic'] = true_ontology_mf_subgraph.get_ic()
        method.current_query_info['true_ontology_mf_leaves'] = len(true_ontology_mf_subgraph.vs.select(_outdegree=0))
        method.current_query_info['true_ontology_mf_terms'] = len(true_ontology_mf_subgraph.vs)

        candidate_ontology_mf_ids = {v['name'] for v in method.candidate_ontology.vs if
                                     v['term'].aspect == effusion.GoAspectEnum.molecular_function}
        go_ids = candidate_ontology_mf_ids - true_ontology_mf_ids
        method.current_query_info['maximum_fp'] = len(go_ids)
        method.current_query_info['maximum_fp_ic'] = sum(
            effusion.dataset.gene_ontology.term4id(go_id).conditional_ic for go_id in go_ids)

        go_ids = true_ontology_mf_ids - candidate_ontology_mf_ids
        method.current_query_info['maximum_fn'] = len(go_ids)
        method.current_query_info['maximum_fn_ic'] = sum(
            effusion.dataset.gene_ontology.term4id(go_id).conditional_ic for go_id in go_ids)

        go_ids = true_ontology_mf_ids & candidate_ontology_mf_ids
        method.current_query_info['maximum_tp'] = len(go_ids)
        method.current_query_info['maximum_tp_ic'] = sum(
            effusion.dataset.gene_ontology.term4id(go_id).conditional_ic for go_id in go_ids)

        query_info_header = sorted(method.current_query_info.keys())
        print 'header:', '\t'.join(k for k in query_info_header)
        print 'query_info:', '\t'.join(str(method.current_query_info[k]) for k in query_info_header)


def evaluate_in_aggregate():
    # all methods
    # methods = {'BlastMethod$', 'SSMethod\..*$'}

    # best methods
    # methods = {'BlastMethod$', 'SSMethod\..*\.ai_cond$', 'SSMethod.td.ai_ijgp$'}

    # eval_metrics = [('wru', 'wmi'),
    #                 ('nru', 'nmi')]

    # 'partial' takes the intersection of the eval_uniprots
    # 'full' tales the union of all the proteins with predictions.
    # so, often all eval_uniprots, but possibly less if, e.g., we only ran the first 1000 preds
    # eval_set = 'full'

    # just testing
    plot_performance(methods={'BlastMethod$', 'SSMethod\.td\.ai_cond$'},
                     eval_metrics=[('recall', 'precision')],
                     eval_set='full',
                     subset_type='molecular_function',
                     name='test')
    sys.exit(1)

    # plot best td + bu. all evaluation metrics (on other plots), but some of them go in supplementary
    plot_performance(methods={'BlastMethod$', 'RandomBlastMethod$', 'SSMethod\..*\.ai_cond$'},
                     # eval_metrics=[('wtp', 'wfp'), ('recall', 'precision'), ('wrc', 'wpr'), ('sw-wrc', 'sw-wpr')],
                     eval_metrics=[('tp', 'fp')],
                     eval_set='full',
                     subset_type='molecular_function')

    # plot best td + bu on partial set
    plot_performance(methods={'BlastMethod$', 'RandomBlastMethod$', 'SSMethod\..*\.ai_cond$'},
                     # eval_metrics=[('wtp', 'wfp'), ('recall', 'precision'), ('wrc', 'wpr'), ('sw-wrc', 'sw-wpr')],
                     eval_metrics=[('wtp', 'wfp')],
                     eval_set='partial',
                     subset_type='molecular_function')

    # # plot best td + bu on partial set. catalytic
    plot_performance(methods={'BlastMethod$', 'RandomBlastMethod$', 'SSMethod\..*\.ai_cond$'},
                     # eval_metrics=[('wtp', 'wfp'), ('recall', 'precision'), ('wrc', 'wpr'), ('sw-wrc', 'sw-wpr')],
                     eval_metrics=[('wtp', 'wfp')],
                     eval_set='partial',
                     subset_type='catalytic')

    # # plot all inference engines
    plot_performance(methods={'BlastMethod$',
                              'RandomBlastMethod$',
                              'SSMethod.bu.ai$',
                              'SSMethod.bu.ai_cond$',
                              'SSMethod.bu.ai_ijgp$',
                              'SSMethod.bu.ai_ijgp_cond$',
                              'SSMethod.bu.bp$',
                              'SSMethod.bu.glc$',
                              'SSMethod.bu.ijgp$',
                              'SSMethod.bu.samplesearch$',
                              'SSMethod.bu.trwbp$',
                              'SSMethod.td.ai$',
                              'SSMethod.td.ai_cond$',
                              'SSMethod.td.ai_ijgp$',
                              'SSMethod.td.ai_ijgp_cond$',
                              'SSMethod.td.bp$',
                              'SSMethod.td.glc$',
                              'SSMethod.td.ijgp$',
                              'SSMethod.td.samplesearch$',
                              'SSMethod.td.trwbp$'},
                     # eval_metrics=[('wtp', 'wfp'), ('recall', 'precision'), ('wrc', 'wpr'), ('sw-wrc', 'sw-wpr')],
                     eval_metrics=[('wtp', 'wfp')],
                     eval_set='full',  # because a lot of them just failed
                     subset_type='molecular_function',
                     name='engines')

    # plot all blastlike and semi-supervised. not really fair because edges don't account for distance
    plot_performance(methods={'BlastMethod$', 'RandomBlastMethod$', 'SSMethod\.td\.ai_cond$', 'BlastLikeMethod',
                              'SupervisedNetworkMethod'},
                     # eval_metrics=[('wtp', 'wfp'), ('recall', 'precision'), ('wrc', 'wpr'), ('sw-wrc', 'sw-wpr')],
                     eval_metrics=[('wtp', 'wfp')],
                     eval_set='partial',
                     # because we want to see realistic performance, ignoring that the MST for a few networks were too big
                     subset_type='molecular_function',
                     name='supervised')

    # no more negs
    plot_performance(methods={'BlastMethod$', 'RandomBlastMethod$', 'SSMethod.td.ai_cond$', 'SSMethod.bu.ai_cond$',
                              '.*\.no_more_negs'},
                     # eval_metrics=[('wtp', 'wfp'), ('recall', 'precision'), ('wrc', 'wpr'), ('sw-wrc', 'sw-wpr')],
                     eval_metrics=[('wtp', 'wfp')],
                     eval_set='partial',
                     # because we want to see realistic performance, ignoring that the MST for a few networks were too big
                     subset_type='molecular_function',
                     name='no_more_negs')

    # no pseudocounts
    plot_performance(methods={'BlastMethod$', 'RandomBlastMethod$', 'SSMethod.td.ai_cond$', 'SSMethod.bu.ai_cond$',
                              '.*\.no_pseudocounts'},
                     # eval_metrics=[('wtp', 'wfp'), ('recall', 'precision'), ('wrc', 'wpr'), ('sw-wrc', 'sw-wpr')],
                     eval_metrics=[('wtp', 'wfp')],
                     eval_set='partial',
                     # because we want to see realistic performance, ignoring that the MST for a few networks were too big
                     subset_type='molecular_function',
                     name='no_pseudocounts')

    # no more negs
    plot_performance(methods={'BlastMethod$', 'RandomBlastMethod$', 'SSMethod.td.ai_cond$', 'SSMethod.bu.ai_cond$',
                              '.*\.not_weighted'},
                     # eval_metrics=[('wtp', 'wfp'), ('recall', 'precision'), ('wrc', 'wpr'), ('sw-wrc', 'sw-wpr')],
                     eval_metrics=[('wtp', 'wfp')],
                     eval_set='partial',
                     # because we want to see realistic performance, ignoring that the MST for a few networks were too big
                     subset_type='molecular_function',
                     name='not_weighted')


def plot_performance(methods, eval_metrics, eval_set, subset_type, name=None):
    # todo i have it here because i don't feel like installing numpy on the cluster
    import pandas as pd

    postprocess_args = {
        'subset_type': subset_type,
        'do_filter_purified': False,
        'do_break_ties': True,
        'prop_down_type': None,  # 'smart_prop_down',
        'max_predictions': 200
    }

    effusion.dataset.set_phase_and_role(effusion.args.phase, 'evaluation')

    experiment_name = os.path.split(effusion.dataset.preds_dir)[-1]
    if name:
        experiment_name += '.' + name

    eval_uniprots = set(effusion.dataset.evaluation_proteins)
    if False:  # testing
        print 'testing code!!'
        eval_uniprots = set(list(eval_uniprots)[0:200])

    if eval_set == 'full':
        # uniprots_in_common = set()
        uniprots_in_common = set(effusion.dataset.evaluation_proteins)
    elif eval_set == 'partial':
        uniprots_in_common = set(effusion.dataset.evaluation_proteins)
    else:
        raise ValueError

    num_experiments = 0

    # ('recall', 'precision'), ('wpr', 'wrc'), ('sw-pr', 'sw-rc')
    method_metric_accumulator_map = {
        'tp': effusion.evaluation.make_method_accumulator_class(
            protein_accumulator_class=effusion.evaluation.make_protein_accumulator_class(
                numerator_measure=effusion.evaluation.TruePositive(),
                denominator_measure=effusion.evaluation.Constant(),
                term_weighted=False,
                default_value=1),
            sample_weighted=False),
        'fp': effusion.evaluation.make_method_accumulator_class(
            protein_accumulator_class=effusion.evaluation.make_protein_accumulator_class(
                numerator_measure=effusion.evaluation.FalsePositive(),
                denominator_measure=effusion.evaluation.Constant(),
                term_weighted=False,
                default_value=1),
            sample_weighted=False),
        'wtp': effusion.evaluation.make_method_accumulator_class(
            protein_accumulator_class=effusion.evaluation.make_protein_accumulator_class(
                numerator_measure=effusion.evaluation.TruePositive(),
                denominator_measure=effusion.evaluation.Constant(),
                term_weighted=True,
                default_value=1),
            sample_weighted=False),
        'wfp': effusion.evaluation.make_method_accumulator_class(
            protein_accumulator_class=effusion.evaluation.make_protein_accumulator_class(
                numerator_measure=effusion.evaluation.FalsePositive(),
                denominator_measure=effusion.evaluation.Constant(),
                term_weighted=True,
                default_value=1),
            sample_weighted=False),
        'recall': effusion.evaluation.make_method_accumulator_class(
            protein_accumulator_class=effusion.evaluation.make_protein_accumulator_class(
                numerator_measure=effusion.evaluation.TruePositive(),
                denominator_measure=effusion.evaluation.ConditionPositive(),
                term_weighted=False,
                default_value=0),
            sample_weighted=False),
        'precision': effusion.evaluation.make_method_accumulator_class(
            protein_accumulator_class=effusion.evaluation.make_protein_accumulator_class(
                numerator_measure=effusion.evaluation.TruePositive(),
                denominator_measure=effusion.evaluation.PredictedConditionPositive(),
                term_weighted=False,
                default_value=1),
            sample_weighted=False),
        'wrc': effusion.evaluation.make_method_accumulator_class(
            protein_accumulator_class=effusion.evaluation.make_protein_accumulator_class(
                numerator_measure=effusion.evaluation.TruePositive(),
                denominator_measure=effusion.evaluation.ConditionPositive(),
                term_weighted=True,
                default_value=0),
            sample_weighted=False),
        'wpr': effusion.evaluation.make_method_accumulator_class(
            protein_accumulator_class=effusion.evaluation.make_protein_accumulator_class(
                numerator_measure=effusion.evaluation.TruePositive(),
                denominator_measure=effusion.evaluation.PredictedConditionPositive(),
                term_weighted=True,
                default_value=1),
            sample_weighted=False),
        'sw-wrc': effusion.evaluation.make_method_accumulator_class(
            protein_accumulator_class=effusion.evaluation.make_protein_accumulator_class(
                numerator_measure=effusion.evaluation.TruePositive(),
                denominator_measure=effusion.evaluation.ConditionPositive(),
                term_weighted=True,
                default_value=0),
            sample_weighted=True),
        'sw-wpr': effusion.evaluation.make_method_accumulator_class(
            protein_accumulator_class=effusion.evaluation.make_protein_accumulator_class(
                numerator_measure=effusion.evaluation.TruePositive(),
                denominator_measure=effusion.evaluation.PredictedConditionPositive(),
                term_weighted=True,
                default_value=1),
            sample_weighted=True)
    }

    method_to_preds_dict = {}
    for dirpath, dirnames, filenames in os.walk(effusion.dataset.preds_dir, followlinks=True):
        if dirpath == effusion.dataset.preds_dir:
            continue

        method_name = os.path.basename(dirpath)

        matches_pattern = False
        for method_regex in methods:
            matches_pattern = True if re.match(method_regex, method_name) else False
            if matches_pattern:
                break
        if not matches_pattern:
            continue

        # replace filesystem name with display name
        method_name = method_name.replace('SSMethod', 'Effusion')
        method_name = method_name.replace('SupervisedNetworkMethod', 'Effusion.supervised')
        method_name = method_name.replace('BlastLikeMethod', 'Effusion.blast_like')
        method_name = method_name.replace('RandomBlast', 'Random BLAST')
        method_name = method_name.replace('Blast', 'BLAST')
        method_name = method_name.replace('Method', '')
        # method_name = method_name.replace('')

        if '.' in method_name:
            method_name_split = method_name.split('.')
            method_name = '%s (%s)' % (method_name_split[0], ', '.join(method_name_split[1:]))

        logging.info('Working on %s' % method_name)

        num_experiments += 1

        # non-root, because some methods that fail return just the root. or just the MST heuristic didn't work out
        num_queries_with_non_root_predictions = 0

        uniprots_for_method = set()

        preds_dict = {}
        for filename in filenames:
            if not filename.endswith('.preds.json'):
                continue
            uniprot = filename.replace('.preds.json', '')

            if uniprot not in eval_uniprots:
                continue

            with open(os.path.join(dirpath, filename)) as preds_file:
                raw_preds = json.load(preds_file)

                preds = effusion.postprocess(effusion.dataset,
                                             uniprot,
                                             raw_preds,
                                             **postprocess_args)
                go_ids_predicted = {p.go_id for p in preds}

                query_has_non_root_predictions = len(go_ids_predicted - effusion.dataset.gene_ontology.root_ids) > 0

                if query_has_non_root_predictions:
                    uniprots_for_method.add(uniprot)

                num_queries_with_non_root_predictions += query_has_non_root_predictions

                preds_dict[uniprot] = preds

        logging.info(
            'Method predicted at least one non-root annotation for num proteins: %s' % num_queries_with_non_root_predictions)

        method_to_preds_dict[method_name] = effusion.ImmutableDict(preds_dict)

        if eval_set == 'full':
            uniprots_in_common |= uniprots_for_method
        elif eval_set == 'partial':
            uniprots_in_common &= uniprots_for_method
        else:
            raise ValueError

        eval_uniprots = uniprots_in_common

    print 'Evaluating against %d uniprots' % len(eval_uniprots)

    # get true label dict  (cutoff date will be applied)
    annotations_dict = effusion.Annotation.get_annotations_for_uniprots(uniprots=eval_uniprots,
                                                                        include_negative=False,
                                                                        include_direct=True,
                                                                        include_implied=True,
                                                                        include_moonlight=False)
    all_Ts = {}
    for uniprot, annotations in annotations_dict.items():
        label = effusion.Label(effusion.dataset.gene_ontology, annotations,
                               catalytic_only=subset_type == 'catalytic',
                               mf_only=subset_type == 'molecular_function',
                               bp_only=subset_type == 'biological_process')
        all_Ts[uniprot] = label

        if False:  # print the evaluation protein stats
            true_ontology_mf_ids = frozenset({a.go_id for a in annotations if effusion.dataset.gene_ontology.term4id(
                a.go_id).aspect == effusion.GoAspectEnum.molecular_function})
            true_ontology_mf_subgraph = effusion.dataset.gene_ontology.subgraph(true_ontology_mf_ids,
                                                                                keep_ancestors=True)  # can't imagine keep_ancestors is useful here
            # if date proceeded cutoff and it is an MF id
            query_had_mf_evidence_during_training = any((a.date < effusion.dataset.cutoff_dates['validation']
                                                         and a.go_id in true_ontology_mf_ids) for a in
                                                        annotations)  # hack
            print uniprot, true_ontology_mf_subgraph.get_ic(), len(
                true_ontology_mf_subgraph.vs.select(_outdegree=0)), len(
                true_ontology_mf_subgraph.vs), query_had_mf_evidence_during_training

    for eval_metric in eval_metrics:
        logging.info('Plotting metric %s' % str(eval_metric))

        ax = None

        num_plots = num_experiments
        num_plotted = 0
        colormap = cm.get_cmap('Spectral')

        def legend_order_func(method_name_and_preds_dict):
            method_name = method_name_and_preds_dict[0]
            if 'Effusion' in method_name:
                hypo0_base1 = 0
            elif 'BLAST' in method_name:
                hypo0_base1 = 1
            else:
                raise ValueError
            return hypo0_base1, method_name

        for method_name, preds_dict in sorted(method_to_preds_dict.iteritems(), key=legend_order_func):
            # evaluation set options: (1) all proteins with exp (2) only proteins with pred (3) only proteins where all methods have preds

            filtered_preds_dict = {k: v for k, v in preds_dict.items() if k in eval_uniprots}
            merged_ordered_Ps = sorted(chain(*filtered_preds_dict.values()), key=lambda p: p.prob, reverse=True)

            df = pd.DataFrame()

            metric_accumulators = [
                method_metric_accumulator_map[eval_metric[0]](all_Ts),
                method_metric_accumulator_map[eval_metric[1]](all_Ts),
            ]

            for i, mop_item in enumerate(merged_ordered_Ps):
                tau = mop_item.prob

                metric_accumulators[1].accumulate(mop_item)
                metric_accumulators[0].accumulate(mop_item)

                df = df.append({eval_metric[0]: metric_accumulators[0].get_val(),
                                eval_metric[1]: metric_accumulators[1].get_val(),
                                'tau': tau},
                               ignore_index=True)

            df = df.drop_duplicates('tau',
                                    'last')  # merge rows with the same tau. we already have a penalty. let's not make it arbitrary on order of sibling leaves evaluated for blast which was annotated to both...
            df = df.drop_duplicates(eval_metric[0],
                                    'first')  # x axis is first time we acheived that score. don't cound first good prediction and all the bad ones after it. may actually be really important!!!

            color = colormap(num_plotted / num_plots)
            if method_name == 'BLAST':
                color = 'black'
            elif method_name == 'Effusion (td, ai_cond)':
                color = 'blue'
            elif method_name == 'Effusion (bu, ai_cond)':
                color = 'orange'
            if num_plots > 5:
                alpha = .7
            else:
                alpha = 1
            ax = df.plot(ax=ax, x=eval_metric[0], y=eval_metric[1], color=color, alpha=alpha, label=method_name)

            num_plotted += 1

        ax.set_ylabel(eval_metric[1])
        # ax.set_xlim(0, 1)
        # ax.set_ylim(0, .2)


        ax.legend(fontsize='small')
        fig = ax.get_figure()
        fig.savefig('%s.%s.%s.%s.svg' % (experiment_name,
                                         subset_type,
                                         eval_set,
                                         '_'.join(eval_metric)))


def evaluate_per_protein():
    # per_protein_analysis('BlastMethod',
    #                      'SSMethod.td.ai_cond',
    #                      'molecular_function',
    #                      ('wtp', 'wfp'),
    #                      'standard',
    #                      'blast-td')

    per_protein_analysis('BlastMethod',
                         'SSMethod.td.ai_cond',
                         'catalytic',
                         ('wtp', 'wfp'),
                         'standard',
                         'blast-td')

    # per_protein_analysis('SSMethod.td.ai_cond.no_more_negs',
    #                      'SSMethod.td.ai_cond',
    #                      'molecular_function',
    #                      ('p_ic', 'mi'),
    #                      'siblings',
    #                      'negs-td')
    #
    # per_protein_analysis('SSMethod.td.ai_cond.no_more_negs',
    #                      'SSMethod.bu.ai_cond',
    #                      'molecular_function',
    #                      ('p_ic', 'mi'),
    #                      'siblings',
    #                      'negs-bu')
    #
    # per_protein_analysis('SSMethod.td.ai_cond.not_weighted',
    #                      'SSMethod.td.ai_cond',
    #                      'molecular_function',
    #                      ('p_ic', 'mi'),
    #                      'weighting',
    #                      'weighting-td')

    # per_protein_analysis('SSMethod.td.ai_cond.no_pseudocounts',
    #                      'SSMethod.td.ai_cond',
    #                      'molecular_function',
    #                      ('p_ic', 'mi'),
    #                      'pseudocounts',
    #                      'pseudocounts-td')

    # per_protein_analysis('SSMethod.td.ai_cond.no_pseudocounts',
    #                      'SSMethod.bu.ai_cond',
    #                      'molecular_function',
    #                      ('p_ic', 'mi'),
    #                      'pseudocounts',
    #                      'pseudocounts-bu')


def per_protein_analysis(baseline_method, hypothesis_method, subset_type, protein_eval_metric, example_finding_details,
                         experiment_name):
    # todo i have it here because i don't feel like installing numpy on the cluster
    import pandas as pd
    from numpy import trapz

    pd.set_option('display.expand_frame_repr', False)
    pd.set_option('display.max_columns', 500)
    pd.set_option('display.width', 1000)

    metrics_str = '-'.join(protein_eval_metric)
    outpath = '.'.join([experiment_name, subset_type, metrics_str, example_finding_details, 'tsv'])
    outfile = open(outpath, 'w')

    eval_set = 'partial'

    metric_accumulator_map = {
        'wtp': effusion.evaluation.make_protein_accumulator_class(
            numerator_measure=effusion.evaluation.TruePositive(),
            denominator_measure=effusion.evaluation.Constant(),
            term_weighted=True,
            default_value=0),
        'wfp': effusion.evaluation.make_protein_accumulator_class(
            numerator_measure=effusion.evaluation.FalsePositive(),
            denominator_measure=effusion.evaluation.Constant(),
            term_weighted=True,
            default_value=0),
    }

    # 'wtp': effusion.evaluation.make_method_accumulator_class(
    #     protein_accumulator_class=effusion.evaluation.make_protein_accumulator_class(
    #         numerator_measure=effusion.evaluation.TruePositive(),
    #         denominator_measure=effusion.evaluation.Constant(),
    #         term_weighted=True,
    #         default_value=1),
    #     sample_weighted=False),
    # 'wfp': effusion.evaluation.make_method_accumulator_class(
    #     protein_accumulator_class=effusion.evaluation.make_protein_accumulator_class(
    #         numerator_measure=effusion.evaluation.FalsePositive(),
    #         denominator_measure=effusion.evaluation.Constant(),
    #         term_weighted=True,
    #         default_value=1),
    #     sample_weighted=False),

    methods = {baseline_method, hypothesis_method}

    postprocess_args = {
        'subset_type': subset_type,
        'do_filter_purified': False,
        'do_break_ties': True,
        'prop_down_type': None,  # 'smart_prop_down',
        'max_predictions': 200
    }

    effusion.dataset.set_phase_and_role(effusion.args.phase, 'evaluation')

    eval_uniprots = set(effusion.dataset.evaluation_proteins)
    # eval_uniprots = {'Q4ZIL6'}  # eff_vs_blast example
    # eval_uniprots = {'Q921C5'}  # td_vs_bu example
    # eval_uniprots = {'P34945'}  # weighting example
    # eval_uniprots = {'Q6UWY2'}  # pseudocounts example
    # eval_uniprots = {'Q99497'}  # failed example
    if eval_set == 'full':
        uniprots_in_common = set()
    elif eval_set == 'partial':
        uniprots_in_common = set(effusion.dataset.evaluation_proteins)
    else:
        raise ValueError

    method_to_preds_dict = {}

    num_experiments = 0

    for dirpath, dirnames, filenames in os.walk(effusion.dataset.preds_dir, followlinks=True):
        if dirpath == effusion.dataset.preds_dir:
            continue

        method_name = os.path.basename(dirpath)

        if method_name not in methods:
            continue

        num_experiments += 1

        logging.info('Working on %s' % method_name)

        # non-root, because some methods that fail return just the root. or just the MST heuristic didn't work out
        num_queries_with_non_root_predictions = 0

        uniprots_for_method = set()

        preds_dict = {}
        for filename in filenames:
            if not filename.endswith('.preds.json'):
                continue
            uniprot = filename.replace('.preds.json', '')

            if uniprot not in eval_uniprots:
                continue

            with open(os.path.join(dirpath, filename)) as preds_file:
                raw_preds = json.load(preds_file)

                preds = effusion.postprocess(effusion.dataset,
                                             uniprot,
                                             raw_preds,
                                             **postprocess_args)
                go_ids_predicted = {p.go_id for p in preds}

                query_has_non_root_predictions = len(go_ids_predicted - effusion.dataset.gene_ontology.root_ids) > 0

                if query_has_non_root_predictions:
                    uniprots_for_method.add(uniprot)

                    # prob = next(p.prob for p in preds if p.go_id == 'GO:0003674')
                    # if prob < .99:
                    #     print uniprot, 'GO:0003674', prob, 'failed inference'

                num_queries_with_non_root_predictions += query_has_non_root_predictions

                preds_dict[uniprot] = preds

        logging.info(
            'Method predicted at least one non-root annotation for num proteins: %s' % num_queries_with_non_root_predictions)

        method_to_preds_dict[method_name] = effusion.ImmutableDict(preds_dict)

        if eval_set == 'full':
            uniprots_in_common |= uniprots_for_method
        elif eval_set == 'partial':
            uniprots_in_common &= uniprots_for_method
        else:
            raise ValueError

        eval_uniprots = uniprots_in_common

    assert num_experiments == 2

    logging.info('Evaluating against %d uniprots' % len(eval_uniprots))

    # get true label dict  (cutoff date will be applied)
    annotations_dict = effusion.Annotation.get_annotations_for_uniprots(uniprots=eval_uniprots,
                                                                        include_negative=False,
                                                                        include_direct=True,
                                                                        include_implied=True,
                                                                        include_moonlight=False)
    all_Ts = {}
    for uniprot, annotations in annotations_dict.items():
        label = effusion.Label(effusion.dataset.gene_ontology, annotations,
                               catalytic_only=subset_type == 'catalytic',
                               mf_only=subset_type == 'molecular_function',
                               bp_only=subset_type == 'biological_process')
        all_Ts[uniprot] = label

        if False:  # print the evaluation protein stats
            true_ontology_mf_ids = frozenset({a.go_id for a in annotations if effusion.dataset.gene_ontology.term4id(
                a.go_id).aspect == effusion.GoAspectEnum.molecular_function})
            true_ontology_mf_subgraph = effusion.dataset.gene_ontology.subgraph(true_ontology_mf_ids,
                                                                                keep_ancestors=True)  # can't imagine keep_ancestors is useful here
            outfile.write('%f\t%d\t%d\n' % true_ontology_mf_subgraph.get_ic(),
                          len(true_ontology_mf_subgraph.vs.select(_outdegree=0)), len(true_ontology_mf_subgraph.vs))

    class ProteinResult(object):
        def __init__(self, example_finding_details=None):
            default_fields = [
                ('uniprot', None),
                ('true_ic', None),
                ('aucs', None),
                ('min_y_diff', 0),
                ('max_y_diff', 0),
                ('largest_correct_u', set()),
                ('largest_incorrect_v', set()),
                ('best_mrca_ic', 0),
                ('best_mrca_catalytic', None),
                ('best_mrca_ui_vi', None),
                ('best_mrca_base_ui_increase', 0),
                ('best_mrca_ydiff', None)]
            sibling_fields = [
                ('best_siblings_hypo_vi_decrease', 0),
                ('best_siblings_vi', None),
                ('best_siblings_vi_depth', None),
                ('best_siblings_predicted_correctly', None)
            ]
            weighting_fields = [
                ('best_params_ui', None),
                ('best_params_hypo_ui_prob', None),
                ('best_params_base_ui_prob', None),
                ('best_params_hypo_ui_base_ui_diff', 0),
                ('best_params_p11_raw', None),
                ('best_params_p11', None),
                ('best_params_p11_diff', 0)
            ]
            pseudocount_fields = [
                ('best_params_ui', None),
                ('best_params_hypo_ui_prob', None),
                ('best_params_base_ui_prob', None),
                ('best_params_hypo_ui_base_ui_diff', None),
                ('best_params_p11_raw', None),
                ('best_params_p11', None),
                ('best_params_p11_diff', None)
            ]

            fields_and_default_values = (default_fields +
                                         (sibling_fields if example_finding_details == 'siblings' else []) +
                                         (weighting_fields if example_finding_details == 'weighting' else []) +
                                         (pseudocount_fields if example_finding_details == 'pseudocounts' else []))

            for field, init_value in fields_and_default_values:
                self.__dict__[field] = init_value

            super(ProteinResult, self).__setattr__('fields',
                                                   [field for field, default_value in fields_and_default_values])

        def __setattr__(self, key, value):
            if key not in self.fields:
                raise KeyError
            super(ProteinResult, self).__setattr__(key, value)

        def __str__(self):
            return '\t'.join(str(self.__dict__[h]) for h in self.fields)

        def header(self):
            return '\t'.join(self.fields)

    outfile.write(ProteinResult(example_finding_details).header() + '\n')
    for uniprot in eval_uniprots:
        result = ProteinResult(example_finding_details)
        result.uniprot = uniprot

        hypo_preds_for_protein = method_to_preds_dict[hypothesis_method][uniprot]
        base_preds_for_protein = method_to_preds_dict[baseline_method][uniprot]

        # filter to common terms
        go_ids_in_common = {p.go_id for p in hypo_preds_for_protein} & {p.go_id for p in base_preds_for_protein}
        true_pos_go_ids_for_protein = {go_id for go_id, val in all_Ts[uniprot].items() if val}
        result.true_ic = effusion.i(true_pos_go_ids_for_protein)
        hypo_preds_for_protein = filter(lambda x: x.go_id in go_ids_in_common, hypo_preds_for_protein)
        base_preds_for_protein = filter(lambda x: x.go_id in go_ids_in_common, base_preds_for_protein)

        if len(eval_uniprots) == 1:
            header = (
                ('GO Term', 's'), ('Posterior Probability', 'f'), ('In Standard', 's'), ('TP Information (Bits)', 'f'),
                ('FP Information (Bits, AKA Misinformation)', 'f'))
            outfile.write('hypo preds for protein:\n')
            outfile.write(' '.join(['%%%ds' % max(len(h[0]), 10) for h in header]) % tuple(h[0] for h in header) + '\n')
            outfile.write(' '.join(['%%%ds' % max(len(h[0]), 10) for h in header]) % tuple(
                '-' * max(len(h[0]), 10) for h in header) + '\n')
            tp_ic = 0
            fp_ic = 0
            for p in hypo_preds_for_protein:
                ic = effusion.dataset.gene_ontology.term4id(p.go_id).conditional_ic
                if p.go_id in true_pos_go_ids_for_protein:
                    mark = u'\u2713'
                    tp_ic += ic
                else:
                    mark = u'\u2717'
                    fp_ic += ic
                outfile.write((u' '.join(['%%%d%s' % (len(h[0]), h[1]) for h in header]) % (
                    p.go_id, p.prob, mark, tp_ic, fp_ic)).encode('utf8') + '\n')
            outfile.write('\nbase preds for protein:\n')
            outfile.write(' '.join(['%%%ds' % max(len(h[0]), 10) for h in header]) % tuple(h[0] for h in header) + '\n')
            outfile.write(' '.join(['%%%ds' % max(len(h[0]), 10) for h in header]) % tuple(
                '-' * max(len(h[0]), 10) for h in header) + '\n')
            tp_ic = 0
            fp_ic = 0
            for p in base_preds_for_protein:
                ic = effusion.dataset.gene_ontology.term4id(p.go_id).conditional_ic
                if p.go_id in true_pos_go_ids_for_protein:
                    mark = u'\u2713'
                    tp_ic += ic
                else:
                    mark = u'\u2717'
                    fp_ic += ic
                outfile.write((u' '.join(['%%%d%s' % (max(len(h[0]), 10), h[1]) for h in header]) % (
                    p.go_id, p.prob, mark, tp_ic, fp_ic)).encode('utf8') + '\n')
        hypo_df = pd.DataFrame()

        metric_accumulators = [
            metric_accumulator_map[protein_eval_metric[0]](true_pos_go_ids_for_protein),
            metric_accumulator_map[protein_eval_metric[1]](true_pos_go_ids_for_protein),
        ]

        hypo_preds_at_points = {}
        predicted_go_ids_so_far = set()
        for i, pred in enumerate(hypo_preds_for_protein):
            tau = pred.prob

            predicted_go_ids_so_far.add(pred.go_id)

            metric_accumulators[0].accumulate(pred.go_id)
            metric_accumulators[1].accumulate(pred.go_id)

            hypo_df = hypo_df.append({protein_eval_metric[0]: metric_accumulators[0].get_val(),
                                      protein_eval_metric[1]: metric_accumulators[1].get_val(),
                                      'tau': tau,
                                      'preds_key': i},
                                     ignore_index=True)
            hypo_preds_at_points[i] = frozenset(predicted_go_ids_so_far)

        hypo_df = hypo_df.drop_duplicates('tau',
                                          'last')  # merge rows with the same tau. we already have a penalty. let's not make it arbitrary on order of sibling leaves evaluated for blast which was annotated to both...
        hypo_df = hypo_df.round({protein_eval_metric[0]: 8})  # otherwise, indexes can be very slightly off
        hypo_df = hypo_df.drop_duplicates(protein_eval_metric[0], 'first')
        # alternatively, something like to take the biggest or smallest
        # df.sort_values('B', ascending=False).drop_duplicates('A').sort_index()
        hypo_df = hypo_df.set_index(protein_eval_metric[0])

        base_df = pd.DataFrame()

        metric_accumulators = [
            metric_accumulator_map[protein_eval_metric[0]](true_pos_go_ids_for_protein),
            metric_accumulator_map[protein_eval_metric[1]](true_pos_go_ids_for_protein),
        ]
        base_preds_at_points = {}
        predicted_go_ids_so_far = set()
        for i, pred in enumerate(base_preds_for_protein):
            tau = pred.prob

            predicted_go_ids_so_far.add(pred.go_id)

            metric_accumulators[0].accumulate(pred.go_id)
            metric_accumulators[1].accumulate(pred.go_id)

            base_df = base_df.append({protein_eval_metric[0]: metric_accumulators[0].get_val(),
                                      protein_eval_metric[1]: metric_accumulators[1].get_val(),
                                      'tau': tau,
                                      'preds_key': i},
                                     ignore_index=True)
            base_preds_at_points[i] = frozenset(predicted_go_ids_so_far)

        base_df = base_df.drop_duplicates('tau', 'last')  # merge proteins with the same tau
        base_df = base_df.round({protein_eval_metric[0]: 8})  # otherwise, indexes can be very slightly off
        base_df = base_df.drop_duplicates(protein_eval_metric[0],
                                          'first')  # x axis is first time we acheived that score. don't cound first good prediction and all the bad ones after it. may actually be really important!!!

        base_df = base_df.set_index(protein_eval_metric[0])

        # outer merge sorts lexicographically
        # how we sort depends on our x metric. if we want to get bigger with each added pred (mi), sort in ascending order. else (ru) descending
        sort_ascending = True  # we currently don't have any x metrics that descrease with each additional pred
        merged_df = base_df.join(hypo_df, lsuffix='.base', rsuffix='.hypo', how='outer').sort_index(
            ascending=sort_ascending)
        merged_df = merged_df.fillna(method='ffill')

        if len(eval_uniprots) == 1:
            # print hypo_df
            # print base_df
            outfile.write(str(merged_df) + '\n')

        # these give negative values if our x points go from right to left. rather than reverse the dataframe again, just take the absolute value of each calculation
        # area_simps = simps(x=x, y=y)
        hypo_area = abs(trapz(x=hypo_df.index, y=hypo_df[protein_eval_metric[1]]))
        base_area = abs(trapz(x=base_df.index, y=base_df[protein_eval_metric[1]]))
        area_diff = round(hypo_area - base_area, 8)
        result.aucs = area_diff

        for index, row in merged_df.iterrows():
            y_diff = row[protein_eval_metric[1] + '.hypo'] - row[protein_eval_metric[1] + '.base']

            result.min_y_diff = min(y_diff, result.min_y_diff)
            result.max_y_diff = max(y_diff, result.max_y_diff)

            u = hypo_preds_at_points[row['preds_key.hypo']] - base_preds_at_points[row['preds_key.base']]
            correct_u = u & true_pos_go_ids_for_protein
            v = base_preds_at_points[row['preds_key.base']] - hypo_preds_at_points[row['preds_key.hypo']]
            incorrect_v = v - true_pos_go_ids_for_protein

            if len(correct_u) > len(result.largest_correct_u):
                result.largest_correct_u = correct_u
            if len(incorrect_v) > len(result.largest_incorrect_v):
                result.largest_incorrect_v = incorrect_v

            if y_diff > 0:  # let's not consider examples where we're doing worse than baseline
                continue

            for u_i in correct_u:  # this works much better total preds on the x axis, and either recall or error on the y-axis
                for v_i in incorrect_v:
                    hypo_ui_prob = next(pred.prob for pred in hypo_preds_for_protein if pred.go_id == u_i)
                    if hypo_ui_prob < .2:
                        continue

                    base_vi_prob = next(pred.prob for pred in base_preds_for_protein if pred.go_id == v_i)
                    base_ui_prob = next(pred.prob for pred in base_preds_for_protein if pred.go_id == u_i)

                    ui_increase = base_vi_prob - base_ui_prob
                    # base pred must not have predicted other pred with nearly the same probability
                    if ui_increase < .01:  # any increase means it's a closer protein
                        continue

                    vs = [effusion.dataset.gene_ontology.vertex4id(u_i),
                          effusion.dataset.gene_ontology.vertex4id(v_i)]
                    mrca = effusion.dataset.gene_ontology.get_mrca(vs)
                    # if not mrca.catalytic:
                    #     continue

                    if mrca.marginal_ic > result.best_mrca_ic:
                        result.best_mrca_ic = mrca.marginal_ic
                        result.best_mrca_catalytic = mrca.catalytic
                        result.best_mrca_ui_vi = u_i, v_i
                        result.best_mrca_base_ui_increase = ui_increase
                        result.best_mrca_ydiff = y_diff

            # value of supplementary negatives.
            # find something in v (base w/o negatives predicted it but it was wrong and avoided by hypo)...
            # ... that has a sibling that is correct and has already been predicted
            # <strike>the more unlikely it was with siblings, the better</strike>
            if example_finding_details == 'siblings':
                if area_diff < 0 and y_diff < 0:  # the query is resource intensive. skip unless hypo is doing well
                    for v_i in incorrect_v:
                        vi_vertex = effusion.dataset.gene_ontology.vertex4id(v_i)
                        sibling_ids = effusion.dataset.gene_ontology.get_sibling_ids(vi_vertex)
                        sibling_ids_predicted_positively = sibling_ids & base_preds_at_points[row['preds_key.base']]
                        if not len(sibling_ids_predicted_positively):
                            continue

                        # don't do this. just look to see if hypo predicted it at a much lower probability
                        # p_vi_given_pos_siblings = effusion.get_single_protein_conditional(
                        #     event_state={v_i: True},
                        #     given_state={sib_id: True for sib_id in sibling_ids_predicted_positively}, weighted=True, require_mf=True, require_bp=False, add_pseudocounts=False, contingency_table=None)

                        # for the term predicted incorrectly by base, hypo predicted it this much lower than it's actual prediction
                        hypo_vi_prob = next(pred.prob for pred in hypo_preds_for_protein if pred.go_id == v_i)
                        hypo_ui_prob = row['tau.hypo']
                        hypo_vi_decrease = hypo_ui_prob - hypo_vi_prob

                        if hypo_vi_decrease > result.best_siblings_hypo_vi_decrease:
                            result.best_siblings_hypo_vi_decrease = hypo_vi_decrease
                            result.best_siblings_vi = v_i
                            result.best_siblings_vi_depth = vi_vertex['term'].depth
                            result.best_siblings_predicted_correctly = sibling_ids_predicted_positively

            # value of weighting and adding pseudocounts to params
            # go through the correct predictions in u and find big difference in prob between base and hypo
            # i don't like comparing probabilities from different methods though...
            if example_finding_details == 'pseudocounts' or example_finding_details == 'weighting':
                if area_diff < 0 and y_diff < 0:
                    for u_i in correct_u:
                        # parent_go_ids = effusion.dataset.gene_ontology.get_parent_go_ids(u_i)

                        hypo_ui_prob = next(pred.prob for pred in hypo_preds_for_protein if pred.go_id == u_i)
                        base_ui_prob = next(pred.prob for pred in base_preds_for_protein if pred.go_id == u_i)
                        prob_diff_hypo_ui_to_base_ui = hypo_ui_prob - base_ui_prob

                        p_11_no_pseudocounts = effusion.get_pairwise_protein_conditional(
                            {u_i: True},
                            given_gok_state={},  # aka gok
                            given_pnp_state={u_i: True},  # aka pnp. len() == 1
                            edge_type='blast',
                            weighted=example_finding_details != 'weighting',
                            # if we aren't looking at weighting study, use weights
                            require_mf=True,
                            require_bp=False,
                            add_pseudocounts=example_finding_details != 'pseudocounts')  # if we aren't looking at pseudocounts, use them
                        # 0.967717996289

                        p_11_pseudocounts = effusion.get_pairwise_protein_conditional(
                            {u_i: True},
                            given_gok_state={},  # aka gok
                            given_pnp_state={u_i: True},  # aka pnp. len() == 1
                            edge_type='blast',
                            weighted=True,
                            require_mf=True,
                            require_bp=False,
                            add_pseudocounts=True)
                        p11_diff_from_pseudocounts = p_11_pseudocounts - p_11_no_pseudocounts

                        if p11_diff_from_pseudocounts > result.best_params_p11_diff:
                            result.best_params_p11_diff = p11_diff_from_pseudocounts
                            result.best_params_ui = u_i
                            result.best_params_hypo_ui_prob = hypo_ui_prob
                            result.best_params_base_ui_prob = base_ui_prob
                            result.best_params_hypo_ui_base_ui_diff = hypo_ui_prob - base_ui_prob
                            result.best_params_p11_raw = p_11_no_pseudocounts
                            result.best_params_p11 = p_11_pseudocounts
                            result.best_params_p11_diff = p_11_pseudocounts - p_11_no_pseudocounts

        if True:
            outfile.write(str(result) + '\n')

        blah = 'hi'
    outfile.close()


def main():
    parser = effusion.config.get_argument_parser()

    parser.add_argument('dataset', action='store')
    parser.add_argument('phase', action='store')
    parser.add_argument('--uniprots', action='store')

    subparsers = parser.add_subparsers(help='sub-command help')

    parser_g = subparsers.add_parser('build', help='build')
    parser_g.set_defaults(command=build)
    parser_g.add_argument('build_method')
    parser_g.add_argument('--query-uniprot')
    parser_g.add_argument('--task_number', type=int)

    parser_c = subparsers.add_parser('predict', help='predict')
    parser_c.set_defaults(command=predict)
    parser_c.add_argument('prediction_class')
    parser_c.add_argument('--task_number', type=int)
    parser_c.add_argument('--query-uniprot')

    parser_d = subparsers.add_parser('evaluate-in-aggregate', help='evaluate in aggregate')
    parser_d.set_defaults(command=evaluate_in_aggregate)

    parser_e = subparsers.add_parser('evaluate-per-protein', help='evaluate per protein')
    parser_e.set_defaults(command=evaluate_per_protein)
    parser_e.add_argument('--xval', type=float)
    parser_e.add_argument('--tsv')

    args = effusion.config.parse_args(parser)
    effusion.dataset.init(args.dataset)

    # some tests
    # print effusion.get_pairwise_protein_conditional(
    #     {'GO:0003824': True},
    #     given_gok_state={},  # aka gok
    #     given_pnp_state={},  # aka pnp. len() == 1
    #     edge_type='blast',
    #     weighted=True,
    #     require_mf=True,
    #     require_bp=False)
    #
    # print effusion.get_pairwise_protein_conditional(
    #     {'GO:0003824': True},
    #     given_gok_state={'GO:0003674': True},  # aka gok
    #     given_pnp_state={'GO:0003824': True},  # aka pnp. len() == 1
    #     edge_type='blast',
    #     weighted=True,
    #     require_mf=True,
    #     require_bp=False)
    #
    # print effusion.get_pairwise_protein_conditional(
    #     {'GO:0003824': True},
    #     given_gok_state={'GO:0003674': True},  # aka gok
    #     given_pnp_state={'GO:0003824': True},  # aka pnp. len() == 1
    #     edge_type='string',
    #     weighted=True,
    #     require_mf=True,
    #     require_bp=False)

    # effusion.dataset.set_phase_and_role(effusion.args.phase, 'training')


    # print effusion.get_single_protein_conditional({'GO:0016860': True}, {'GO:0016853': True},
    #                                               weighted=False, require_mf=True, require_bp=False, add_pseudocounts=False)


    # print effusion.get_pairwise_protein_conditional(
    #     {'GO:0004769': True},
    #     given_gok_state={'GO:0016863': True},  # aka gok
    #     given_pnp_state={'GO:0004769': True},  # aka pnp. len() == 1
    #     edge_type='blast',
    #     weighted=False,
    #     require_mf=True,
    #     require_bp=False,
    #     add_pseudocounts=True)
    #
    # print effusion.get_pairwise_protein_conditional(
    #     {'GO:0016863': True},
    #     given_gok_state={'GO:0016860': True},  # aka gok
    #     given_pnp_state={'GO:0016863': True},  # aka pnp. len() == 1
    #     edge_type='blast',
    #     weighted=False,
    #     require_mf=True,
    #     require_bp=False,
    #     add_pseudocounts=False)
    #
    # print effusion.get_pairwise_protein_conditional(
    #     {'GO:0016860': True},
    #     given_gok_state={'GO:0016853': True},  # aka gok
    #     given_pnp_state={'GO:0016860': True},  # aka pnp. len() == 1
    #     edge_type='blast',
    #     weighted=False,
    #     require_mf=True,
    #     require_bp=False,
    #     add_pseudocounts=False)


    # looking at weighting for paper
    if "we are" == "getting some stats for the paper":
        effusion.dataset.set_phase_and_role('test', 'training')

        # no more negs example
        # print effusion.get_single_protein_conditional({'GO:0030276': True}, {'GO:0005515': True}, weighted=True, require_mf=True, require_bp=False, add_pseudocounts=True)
        # print effusion.get_single_protein_conditional({'GO:0030276': True}, {'GO:0019899': True}, weighted=True, require_mf=True, require_bp=False, add_pseudocounts=True)


        # params example
        go_term = 'GO:0016790'
        gok_term = 'GO:0016788'
        gok_parent_term = 'GO:0016787'
        print 'raw', pprint.pformat(effusion.get_pairwise_contingency_table(
            go_ids_for_this_protein={go_term, gok_term},
            go_ids_for_parent_protein={go_term},
            edge_type='blast',
            weighted=False,
            require_mf=True,
            require_bp=False,
            add_pseudocounts=False))
        print 'weighted', pprint.pformat(effusion.get_pairwise_contingency_table(
            go_ids_for_this_protein={go_term, gok_term},
            go_ids_for_parent_protein={go_term},
            edge_type='blast',
            weighted=True,
            require_mf=True,
            require_bp=False,
            add_pseudocounts=False))
        print 'pseudo', pprint.pformat(effusion.get_pairwise_contingency_table(
            go_ids_for_this_protein={go_term, gok_term},
            go_ids_for_parent_protein={go_term},
            edge_type='blast',
            weighted=False,
            require_mf=True,
            require_bp=False,
            add_pseudocounts=True))
        p1 = effusion.get_pairwise_protein_conditional(
            {gok_term: False},
            given_gok_state={gok_parent_term: True},  # aka gok
            given_pnp_state={gok_term: True},  # aka pnp. len() == 1
            edge_type='blast',
            weighted=True,
            require_mf=True,
            require_bp=False,
            add_pseudocounts=True)
        print 'p(%s=0 | %s=1, %s=1) = %f' % (gok_term, gok_term, gok_parent_term, p1)
        p2 = effusion.get_pairwise_protein_conditional(
            {go_term: False},
            given_gok_state={gok_parent_term: True},
            # aka gok   # <- note here we are using to_terms grandparent. see paper
            given_pnp_state={go_term: True},  # aka pnp. len() == 1
            edge_type='blast',
            weighted=True,
            require_mf=True,
            require_bp=False,
            add_pseudocounts=True)
        print 'p(%s=0 | %s=1, %s=1) = %f' % (go_term, go_term, gok_term, p2)

        sys.exit(1)

        print effusion.get_pairwise_protein_conditional(
            {'GO:0004998': True},
            given_gok_state={'GO:0038024': True},  # aka gok
            given_pnp_state={'GO:0004998': True},  # aka pnp. len() == 1
            edge_type='blast',
            weighted=False,
            require_mf=True,
            require_bp=False,
            add_pseudocounts=False)
        # 0.967717996289
        # no parents / kids p(x) = 0
        # with p(x | parent GO:0038024 = 1) = .5

        print effusion.get_pairwise_protein_conditional(
            {'GO:0004998': True},
            given_gok_state={'GO:0038024': True},  # aka gok
            given_pnp_state={'GO:0004998': True},  # aka pnp. len() == 1
            edge_type='blast',
            weighted=False,
            require_mf=True,
            require_bp=False,
            add_pseudocounts=True)
        # no parents / kids p(x) = 93%
        # with p(x | parent GO:0038024 = 1) = .9916

        print effusion.get_pairwise_protein_conditional(
            {'GO:0004998': True},
            given_gok_state={'GO:0038024': True},  # aka gok
            given_pnp_state={'GO:0004998': True},  # aka pnp. len() == 1
            edge_type='blast',
            weighted=False,
            require_mf=True,
            require_bp=False,
            add_pseudocounts=True)

        # general and specific params example
        print effusion.get_pairwise_protein_conditional(
            {'GO:0016788': False},
            given_gok_state={'GO:0016787': True},  # aka gok
            given_pnp_state={'GO:0016788': True},  # aka pnp. len() == 1
            edge_type='blast',
            weighted=True,
            require_mf=True,
            require_bp=False,
            add_pseudocounts=True)
        print effusion.get_pairwise_protein_conditional(
            {'GO:0102667': False},
            given_gok_state={'GO:0016787': True},  # aka gok
            given_pnp_state={'GO:0102667': True},  # aka pnp. len() == 1
            edge_type='blast',
            weighted=True,
            require_mf=True,
            require_bp=False,
            add_pseudocounts=True)

        # weighting example P34945
        print 'weighting example'
        print effusion.get_pairwise_protein_conditional(
            {'GO:0005488': True},
            given_gok_state={},  # aka gok
            given_pnp_state={'GO:0005488': True},  # aka pnp. len() == 1
            edge_type='blast',
            weighted=False,
            require_mf=True,
            require_bp=False,
            add_pseudocounts=True)
        print effusion.get_pairwise_protein_conditional(
            {'GO:0005488': True},
            given_gok_state={},  # aka gok
            given_pnp_state={'GO:0005488': True},  # aka pnp. len() == 1
            edge_type='blast',
            weighted=True,
            require_mf=True,
            require_bp=False,
            add_pseudocounts=True)

        # pseudocount example  GO:1901681
        print 'pseudocount example'
        print effusion.get_pairwise_protein_conditional(
            {'GO:1901681': True},
            given_gok_state={},  # aka gok
            given_pnp_state={'GO:1901681': True},  # aka pnp. len() == 1
            edge_type='blast',
            weighted=True,
            require_mf=True,
            require_bp=False,
            add_pseudocounts=False)
        print effusion.get_pairwise_protein_conditional(
            {'GO:1901681': True},
            given_gok_state={},  # aka gok
            given_pnp_state={'GO:1901681': True},  # aka pnp. len() == 1
            edge_type='blast',
            weighted=True,
            require_mf=True,
            require_bp=False,
            add_pseudocounts=True)

        sys.exit(1)

    args.command()


if __name__ == '__main__':
    main()
