from __future__ import division

__author__ = 'jeff'

import datetime
import configargparse
import gzip
import logging
import math

from sqlalchemy import create_engine, Index, func, case, intersect, distinct, and_, literal_column
from sqlalchemy.orm import scoped_session, sessionmaker

import effusion


def main():
    # assert effusion.args.goa_db.endswith('.db')
    # os.remove(effusion.args.goa_db)

    engine_string = "sqlite+pysqlite:///%s" % effusion.args.goa_db
    engine = create_engine(engine_string,
                           echo_pool=False,
                           echo=False)
    # http://docs.sqlalchemy.org/en/latest/core/engines.html#dbengine-logging
    logging.getLogger('sqlalchemy.engine').setLevel(logging.WARNING)

    session_factory = sessionmaker(bind=engine, autoflush=True, autocommit=False, expire_on_commit=False)
    Session = scoped_session(session_factory)
    session = Session()

    make_go_term_table(engine, session)
    # effusion.dataset.gene_ontology.load_from_db(session)  # needs to be called if we aren't inserting the go_term data
    make_annotation_table(engine, session)
    make_dataset_go_term_table(engine, session)
    make_dataset_uniprot_table(engine, session)


def make_go_term_table(engine, session):
    logging.info('Making GO term table.')

    effusion.GoTerm.__table__.create(bind=engine)
    effusion.dataset.gene_ontology.save_to_db(session)

    logging.info('Creating indices...')
    Index('ix_go_terms_go_id', effusion.GoTerm.go_id).create(engine)
    Index('ix_go_terms_aspect', effusion.GoTerm.aspect).create(engine)


def make_annotation_table(engine, session):
    effusion.Annotation.__table__.create(bind=engine)
    # don't bother loading "manual" IEAs; they all have the same, recent last updated date anyway
    # todo although they still might be useful for prediction...
    # also useful if my dataset is based on actually archived data files
    # but that will still pose new challenges for the algorithm, since protein network and GO structure will be much larger
    batch_size = 50000
    i = 0

    logging.info('Inserting annotations...')
    with gzip.open(effusion.args.goa_gaf_gz) as f:
        new_objects = []

        for next_uniprot, direct_annotations in get_direct_annotations_for_next_uniprot(f):
            if effusion.args.limit is not None and i >= effusion.args.limit:
                break

            new_annotations = []
            new_annotations.extend(direct_annotations)

            implied_annotations = effusion.Annotation.get_all_implied_annotations(
                effusion.dataset.gene_ontology,
                direct_annotations)
            new_annotations.extend(implied_annotations)

            # simplify annotations. one annotation per (uniprot, go_id) pair. keep the earliest, direct annotation
            # uniprot is consistent in new_annotations
            # assert len({a.uniprot for a in new_annotations}) == 1

            earliest_annotations = {}
            # implied gets higher priority, so things will always be consistent.
            #  rarely, we have a direct negative annotation as the parent of a child positive annotation. see Q703P0
            new_annotations.sort(key=lambda x: (x.date,
                                                {'direct': 1,
                                                 'implied': 0,
                                                 'moonlight': 2}[x.source]))
            for a in new_annotations:
                if a.go_id in earliest_annotations:
                    # sys.stdout.write('already saw %s\n' % a)
                    continue
                earliest_annotations[a.go_id] = a
            new_objects.extend(earliest_annotations.values())

            i += 1
            if i % 1000 == 0:
                logging.info('%d' % i)

            if len(new_objects) >= batch_size:
                session.add_all(new_objects)
                session.commit()
                new_objects = []
    session.add_all(new_objects)
    session.commit()

    logging.info('Creating indices...')

    Index('ix_annotations_id', effusion.Annotation.id).create(engine)
    Index('ix_annotations_uniprot', effusion.Annotation.uniprot).create(engine)
    Index('ix_annotations_go_id', effusion.Annotation.go_id).create(engine)
    Index('ix_annotations_source', effusion.Annotation.source).create(engine)
    Index('ix_annotations_date', effusion.Annotation.date).create(engine)
    Index('ix_annotations_value', effusion.Annotation.value).create(engine)
    Index('ix_annotations_date_value_source', effusion.Annotation.date, effusion.Annotation.value,
          effusion.Annotation.source).create(engine)
    Index('ix_annotations_go_id_date_value_source', effusion.Annotation.go_id, effusion.Annotation.date,
          effusion.Annotation.value, effusion.Annotation.source).create(engine)
    Index('ix_annotations_uniprot_date_value_source', effusion.Annotation.uniprot, effusion.Annotation.date,
          effusion.Annotation.value, effusion.Annotation.source).create(engine)


def get_direct_annotations_for_next_uniprot(f):
    curr_uniprot = None
    direct_annotations = None

    for line in f:
        line = line.strip()
        if line[0] == '!':
            continue
        split_line = line.split('\t')
        evidence_code = split_line[6]
        go_ref = split_line[5]
        if (evidence_code == 'IEA' and
                (not effusion.args.manual_iea or
                         go_ref not in ('GO_REF:0000003', 'GO_REF:0000041', 'GO_REF:0000037', 'GO_REF:0000039'))):
            continue

        if split_line[0] != 'UniProtKB':
            continue

        accession = split_line[1]

        if curr_uniprot is not None and accession != curr_uniprot:
            # if we have a new accession...
            # ... yield ...
            yield curr_uniprot, direct_annotations
            # ... and reset
            curr_uniprot, direct_annotations = None, None

        if curr_uniprot is None:
            curr_uniprot, direct_annotations = accession, []

        qualifier = split_line[3]
        if qualifier not in ('', 'NOT'):  # only look at pos or neg *unmodified* annotations
            continue

        value = not qualifier.startswith('NOT')
        go_id = split_line[4]
        try:
            go_term = effusion.dataset.gene_ontology.vs.find(name=go_id)['term']
        except ValueError:
            logging.warning('Annotation with obsolete term. TODO Was there an annotation with the replacing term?')
            continue

        date_str = split_line[13]
        date = datetime.datetime.strptime(date_str, '%Y%m%d').date()
        # o.write('\t'.join([uniprot, qualifier, go_id, aspect, '%d' % catalytic, evidence_code, date]))
        # o.write('\n')
        direct_annotations.append(effusion.Annotation(
            uniprot=curr_uniprot,
            go_id=go_id,
            value=value,
            source='direct',
            evidence_code=evidence_code,
            date=date))

    if curr_uniprot is not None:
        yield curr_uniprot, direct_annotations


def make_dataset_go_term_table(engine, session):
    effusion.DatasetGoTerm.__table__.create(bind=engine, checkfirst=True)
    for dataset_name in effusion.dataset_partition:
        logging.info('Making Dataset-GO Term table for %s' % dataset_name)

        cutoff_date = effusion.dataset.cutoff_dates[dataset_name]

        q = session.query(distinct(effusion.Annotation.uniprot)).filter(effusion.Annotation.date <= cutoff_date).filter(
            effusion.Annotation.value == True).filter(effusion.Annotation.source.in_(('direct', 'implied')))
        num_uniprots_with_evidence = q.count()

        q = session.query(effusion.Annotation.go_id, func.count(effusion.Annotation.uniprot))
        q = q.filter(effusion.Annotation.date <= cutoff_date).filter(effusion.Annotation.value == True).filter(
            effusion.Annotation.source.in_(('direct', 'implied')))
        q = q.group_by(effusion.Annotation.go_id)
        num_uniprots_per_go_id = {}
        for go_id, num_uniprots_with_this_go_id in q.all():
            num_uniprots_per_go_id[go_id] = num_uniprots_with_this_go_id

        cached_dataset_go_terms = {}

        logging.info('Inserting GO IC data...')
        i = 0
        for go_vertex_id in effusion.dataset.gene_ontology.topological_sorting():
            go_vertex = effusion.dataset.gene_ontology.vs[go_vertex_id]
            go_id = go_vertex['name']

            num_uniprots_with_this_go_id = num_uniprots_per_go_id.get(go_id, 0)

            # no uniprots with this go id. infinite IC
            if num_uniprots_with_this_go_id == 0:
                logging.info('No uniprots with %s. Infinite IC. Skipping...' % go_id)
                continue

            # -log(count how many / total uniprots with evidence)
            marginal_ic = -math.log(num_uniprots_with_this_go_id / num_uniprots_with_evidence, 2)

            parent_vertices = go_vertex.predecessors()
            # subtract cumulative joint parents IC from marginal IC to get this conditional IC

            if len(parent_vertices) == 0:
                # if there are no parent vertices...
                conditional_ic = marginal_ic
            elif len(parent_vertices) == 1:
                # if term has only one parent, use the previously calculated value
                dataset_go_term = cached_dataset_go_terms[parent_vertices[0]['name']]

                parent_cumulative_ic = dataset_go_term.marginal_ic
                conditional_ic = marginal_ic - parent_cumulative_ic
            else:
                # if the term has multiple parents, calculate cumulative joint IC of parents with an old fashioned sql query

                # using intersect
                q = session.query(func.count())
                q_template = session.query(effusion.Annotation.uniprot).filter(
                    effusion.Annotation.date <= cutoff_date).filter(effusion.Annotation.value == True).filter(
                    effusion.Annotation.source.in_(('direct', 'implied')))
                intersect_qs = []
                for parent_vertex in parent_vertices:
                    parent_go_id = parent_vertex['name']
                    intersect_q = q_template.filter(effusion.Annotation.go_id == parent_go_id)
                    intersect_qs.append(intersect_q)
                num_uniprots_with_parent_go_ids = q.select_from(intersect(*intersect_qs)).one()[0]

                logging.debug('%s %d %d' % (go_id, num_uniprots_with_parent_go_ids, num_uniprots_with_evidence))

                # there may be a few cases where a there is a negative direct annotation (with the same date)
                #  that overrides a positive annotation implied from a child term
                # if there's only 1 annotation to the child term, then num_uniprots_with_parent_go_ids could be zero
                #  which would result in a domain error here
                # therefore, I am going to change the order to make indirect annotations take priority over direct annotations
                parent_cumulative_ic = -math.log(num_uniprots_with_parent_go_ids / num_uniprots_with_evidence, 2)
                conditional_ic = marginal_ic - parent_cumulative_ic
                pass

            o = effusion.DatasetGoTerm(dataset_name=dataset_name, go_id=go_id, marginal_ic=marginal_ic,
                                       conditional_ic=conditional_ic)
            cached_dataset_go_terms[go_id] = o
            session.add(o)
            i += 1
            if i % 1000 == 0:
                logging.info("%d" % i)
        session.commit()

    logging.info('Creating indices...')
    Index('ix_dataset_dataset_name', effusion.DatasetGoTerm.dataset_name).create(engine)
    Index('ix_dataset_go_id', effusion.DatasetGoTerm.go_id).create(engine)
    Index('ix_dataset_dataset_name_go_id', effusion.DatasetGoTerm.dataset_name, effusion.DatasetGoTerm.go_id).create(
        engine)


def make_dataset_uniprot_table(engine, session):
    effusion.DatasetUniprot.__table__.create(bind=engine)
    for dataset_name in effusion.dataset_partition:
        logging.info('Making Dataset-Uniprot table for %s' % dataset_name)

        cutoff_date = effusion.dataset.cutoff_dates[dataset_name]

        q = session.query(effusion.Annotation.uniprot,
                          func.sum(effusion.DatasetGoTerm.conditional_ic).label('ic'),
                          func.sum(case(
                              [(effusion.GoTerm.aspect == 'molecular_function', effusion.DatasetGoTerm.conditional_ic)],
                              else_=literal_column('0'))).label('mf_ic'),
                          func.sum(case(
                              [(effusion.GoTerm.aspect == 'biological_process', effusion.DatasetGoTerm.conditional_ic)],
                              else_=literal_column('0'))).label('bp_ic'),
                          func.sum(case(
                              [(effusion.GoTerm.aspect == 'cellular_component', effusion.DatasetGoTerm.conditional_ic)],
                              else_=literal_column('0'))).label('cc_ic')
                          ).join(effusion.GoTerm).join(effusion.DatasetGoTerm,
                                                       and_(effusion.Annotation.go_id == effusion.DatasetGoTerm.go_id,
                                                            effusion.DatasetGoTerm.dataset_name == dataset_name))
        q = q.filter(effusion.Annotation.date <= cutoff_date).filter(effusion.Annotation.value == True).filter(
            effusion.Annotation.source.in_(('direct', 'implied')))
        q = q.group_by(effusion.Annotation.uniprot)

        for r in q.all():
            dataset_uniprot = effusion.DatasetUniprot(dataset_name, r.uniprot, r.ic, r.mf_ic, r.bp_ic, r.cc_ic)
            session.add(dataset_uniprot)
        session.commit()

    logging.info('Creating indices...')
    Index('ix_dataset_uniprot_dataset_name', effusion.DatasetUniprot.dataset_name).create(engine)
    Index('ix_dataset_uniprot_uniprot', effusion.DatasetUniprot.uniprot).create(engine)
    Index('ix_dataset_uniprot_dataset_name_uniprot', effusion.DatasetUniprot.dataset_name,
          effusion.DatasetUniprot.uniprot).create(engine)


if __name__ == '__main__':
    parser = configargparse.get_argument_parser()
    parser.add_argument('--manual_iea', action='store_true',
                        help='include \'manual\' IEA evidence codes, like ones that automatically apply a manual mapping of manual swissprot keywords. These have dates of the last automatic application, AKA most recent.')
    parser.add_argument('--limit', action='store', type=int)

    # needs withhold_date, cutoff_date, and gene ontology
    parser.add_argument('dataset', action='store')

    parser.add_argument('goa_gaf_gz', type=str,
                        help='e.g., goa_uniprot_all.gaf.gz')

    parser.add_argument('goa_db', type=str,
                        help='e.g., goa.db')

    args = parser.parse_args()
    effusion.args = args
    effusion.dataset.init(effusion.args.dataset)

    main()
