import argparse

from sqlalchemy import create_engine, Index
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()
from sqlalchemy.orm import scoped_session, sessionmaker
import gzip
import sys

import effusion


def main(args):
    engine_string = "sqlite+pysqlite:///%s" % args.out_db
    engine = create_engine(engine_string,
                           echo_pool=False,
                           echo=False)

    session_factory = sessionmaker(bind=engine, autoflush=False, expire_on_commit=False)
    Session = scoped_session(session_factory)
    session = Session()

    effusion.StringDBBase.metadata.drop_all(engine)
    effusion.StringDBBase.metadata.create_all(session.bind)

    string_to_uniprot_map = get_string_to_uniprot_map(args.aliases)

    with gzip.open(args.links, 'rb') as f:

        batch_size = 50000
        mappings = []
        for i, line in enumerate(f):
            if i == 0:
                continue

            if i % 10000 == 0:
                sys.stderr.write('%d...\n' % i)

            line = line.strip()
            string1, string2, combined_score_str = line.split(' ')
            try:
                uniprot1 = string_to_uniprot_map[string1]
                uniprot2 = string_to_uniprot_map[string2]
            except KeyError:
                continue

            combined_score = int(combined_score_str)
            if combined_score < args.min_string_score:
                continue
            mappings.append(dict(uniprot1=uniprot1, uniprot2=uniprot2, combined_score=combined_score))

            if len(mappings) >= batch_size:
                session.bulk_insert_mappings(effusion.StringLink, mappings=mappings)
                mappings = []
            pass

    session.bulk_insert_mappings(effusion.StringLink, mappings=mappings)
    session.commit()

    i = Index('ix_string_links_uniprot1', effusion.StringLink.__table__.c.uniprot1)
    i.create(engine)

    i = Index('ix_string_links_uniprot2', effusion.StringLink.__table__.c.uniprot2)
    i.create(engine)


def get_uniprot_to_string_map(string_aliases_file_path):
    retval = {}
    with gzip.open(string_aliases_file_path, 'rb') as f:
        for line in f:
            if line[0] == "#": continue

            line = line.strip()
            split_line = line.split("\t")
            id_types = split_line[2]
            if 'BLAST_UniProt_AC' in id_types:  # not sure how "Uniprot_AC" differs (besides, having so few entries, or the uniprot mapping file, for that matter)
                retval[split_line[1]] = split_line[0]

    return retval


def get_string_to_uniprot_map(string_aliases_file_path):
    uniprot_to_string_map = get_uniprot_to_string_map(string_aliases_file_path)
    string_to_uniprot_map = {v: k for k, v in uniprot_to_string_map.items()}
    return string_to_uniprot_map


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='create stringdb sqlite db')
    parser.add_argument('links', action='store',
                        help='protein.links.v10.txt.gz. Download from http://string-db.org/download/protein.links.v10.txt.gz')
    parser.add_argument('aliases', action='store',
                        help='protein.aliases.v10.txt.gz. Download from http://string-db.org/download/protein.aliases.v10.txt.gz')
    parser.add_argument('out_db', action='store')
    parser.add_argument('--min-string-score', type=int, action='store', default=0)
    args = parser.parse_args()
    main(args)
