from __future__ import division

import logging
from collections import defaultdict
from itertools import chain

from functools32 import lru_cache
from sqlalchemy import Column, Integer, String, Boolean, ForeignKey
from sqlalchemy import orm, Date, distinct, func, and_
from sqlalchemy.orm import aliased
from sqlalchemy.sql.expression import literal_column

import effusion


class QuickPred(object):
    value = True

    def __init__(self, uniprot, go_id, prob):
        self.uniprot = uniprot
        self.go_id = go_id
        self.prob = prob

    def to_json(self):
        return {
            'uniprot': self.uniprot,
            'go_id': self.go_id,
            'prob': self.prob
        }

    @classmethod
    def get_all_implied_preds(cls, preds):
        all_implied_preds = set()

        # todo jeff: could simplify here...
        for pred in preds:
            implied_preds = pred.get_implied_preds()
            all_implied_preds |= implied_preds
        return all_implied_preds

    def __str__(self):
        return '(%s, %s, %f)' % (self.uniprot, self.go_id, self.prob)

    def get_implied_preds(self):
        # we are only propagating up, since preds are always positive annotations. gene ontology doesn't matter for now. may matter if predicting on goslim
        gene_ontology = effusion.dataset.gene_ontology

        # todo do i python program with too many for comprehensions?

        implied_go_ids = {v['name'] for v in gene_ontology.get_ancestors(gene_ontology.vertex4id(self.go_id))}

        retval = {self.__class__(self.uniprot, implied_go_id, self.prob) for implied_go_id in implied_go_ids}

        return retval

    # @property
    # def value(self):
    #     return True

    # called as a last resort
    # ref: https://docs.python.org/3/reference/datamodel.html#object.__getattr__
    # only because get implied copies source, date, etc.
    def __getattr__(self, item):
        return None


class QuickAnnotation(object):
    def __init__(self, uniprot, go_id, value):
        self.uniprot = uniprot
        self.go_id = go_id
        self.value = value


class Annotation(effusion.EffusionBase):
    __tablename__ = 'annotations'

    id = Column(Integer, primary_key=True)

    uniprot = Column(String(length=10))

    go_id = Column(ForeignKey(effusion.GoTerm.go_id))
    go_term = orm.relationship(effusion.GoTerm)

    value = Column(Boolean())
    source = Column(String(length=40))
    evidence_code = Column(String(length=4))
    date = Column(Date)

    @classmethod
    def get_goas_for_uniprot(cls, uniprot, include_negative=True, include_direct=True, include_implied=True,
                             include_moonlight=False):
        assert effusion.dataset.cutoff_date

        annotations = []

        query = effusion.dataset.database_session.query(cls)
        query = query.filter(cls.uniprot == uniprot).filter(Annotation.date <= effusion.dataset.cutoff_date)

        if not include_negative:
            query = query.filter(effusion.Annotation.value == True)

        sources = []
        if include_direct:
            sources.append('direct')
        if include_implied:
            sources.append('implied')
        if include_moonlight:
            sources.append('moonlight')
        query = query.filter(effusion.Annotation.source.in_(sources))

        for annotation in query:
            annotations.append(annotation)

        # Make set values hashable to i(T) can be cached
        # evidence = {k:frozenset(my_set) for k, my_set in evidence.items()}
        return annotations

    @classmethod
    def get_uniprots_matching_evidence(cls, evidence, require_mf, require_bp, limit=None):
        """

        :param evidence: go_id -> value map
        :return:
        """

        assert effusion.dataset.cutoff_date
        session = effusion.dataset.database_session

        q = session.query(effusion.DatasetUniprot.uniprot.label('uniprot'),
                          literal_column('1').label('raw_count'))
        q = q.filter(effusion.DatasetUniprot.dataset_name == effusion.dataset.dataset_cutoff_name)

        # can use this to limit to terms with mf. can also join to 3674
        if require_mf:
            q = q.filter(effusion.DatasetUniprot.mf_ic > 0)
        if require_bp:
            q = q.filter(effusion.DatasetUniprot.bp_ic > 0)

        for go_id, go_val in evidence.items():
            a = aliased(effusion.Annotation)
            q = q.outerjoin(a, and_(a.uniprot == effusion.DatasetUniprot.uniprot,
                                    a.date <= effusion.dataset.cutoff_date,
                                    a.source.in_(('direct', 'implied')),
                                    a.go_id == go_id))
            q = q.filter(a.value == go_val)

        if limit:
            q = q.limit(limit)

        print q
        uniprots = None
        for r in q:
            print r

        return uniprots

    # cached in the DB by make_goa_db.py
    @classmethod
    def get_all_implied_annotations(cls, gene_ontology, annotations):
        """ todo xxx DOES NOT retval includes original terms """
        # todo xxx DOES NOT retval includes original terms
        # todo(jyunes) there's also the case for decoupling uniprot from the go_ids, so i wonder if there's a standard normalization / denormalization mechanism


        retval = []
        for annotation in annotations:
            if annotation.value is True:
                implied_go_ids = {v['name'] for v in
                                  gene_ontology.get_ancestors(gene_ontology.vertex4id(annotation.go_id))}
            else:
                # a negative annotation is not a 1-p positive annotation. it's a label!
                assert annotation.value is False
                implied_go_ids = {v['name'] for v in
                                  gene_ontology.get_descendants(gene_ontology.vertex4id(annotation.go_id))}

            for implied_go_id in implied_go_ids:

                if implied_go_id == annotation.go_id:
                    # don't call this annotation implied
                    continue

                # new sqlalchemy / DB object
                # todo couldn't get this to work
                # implied_annotation = copy(annotation)
                # orm.make_transient(implied_annotation)
                # implied_annotation.id = None
                implied_annotation = cls(uniprot=annotation.uniprot,
                                         go_id=implied_go_id,
                                         value=annotation.value,
                                         evidence_code=annotation.evidence_code,
                                         date=annotation.date,
                                         source='implied')

                # the penalty simply allows higher level terms to be predicted before lower level terms
                # otherwise the terms may be evaluated in random order (only affects detailed analysis)
                # todo(jyunes) formalize this better
                # todo xxx
                # level_penalty = .0001 * gene_ontology.term4id(implied_go_id).level
                # implied_annotation.prob -= level_penalty

                retval.append(implied_annotation)
        return retval

    @classmethod
    def get_all_implicitly_negative_annotations(cls, gene_ontology, uniprot, annotations):
        implied_annotations = []

        positive_go_ids = {a.go_id for a in annotations if a.value is True}
        negative_go_ids = gene_ontology.imply_negative_go_ids(positive_go_ids)

        for negative_go_id in negative_go_ids:
            # todo(jyunes) not sure what to set as implied_by, date, or prob when multiple siblings had evidence. one for each?
            # don't set for now, may not be used anyway
            implied_annotations.append(Annotation(uniprot, negative_go_id, value=False))

        return implied_annotations

    @classmethod
    def get_uif_annotations(cls, gene_ontology, annotations):
        uniprots_in_annotations = {a.uniprot for a in annotations}
        assert len(uniprots_in_annotations) == 1
        uniprot = iter(uniprots_in_annotations).next()

        implied_annotations = []

        positive_go_ids = {a.go_id for a in annotations if a.value}
        negative_go_ids = gene_ontology.imply_negative_uif_go_ids(positive_go_ids)

        # this part is pretty slow
        for negative_go_id in negative_go_ids:
            # implied_annotations.append(Annotation(uniprot, negative_go_id, value=False, source='uif', evidence_code=None, date=None))
            implied_annotations.append(QuickAnnotation(uniprot=uniprot, go_id=negative_go_id, value=False))

        return implied_annotations

    @classmethod
    def get_annotations_for_uniprots(cls, uniprots, *args, **kwargs):
        retval = {}
        for uniprot in uniprots:
            annotations = cls.get_goas_for_uniprot(uniprot, *args, **kwargs)
            retval[uniprot] = annotations
        return retval

    # @orm.reconstructor
    # def init_on_load(self):
    #     self.implied_by = None


    def __init__(self, uniprot, go_id, value=True, source=None, evidence_code=None, date=None):
        super(Annotation, self).__init__()

        self.uniprot = uniprot
        self.go_id = go_id
        self.value = value
        self.source = source
        self.evidence_code = evidence_code
        self.date = date

    def __repr__(self):
        return '(%s, %s)' % (self.uniprot, self.go_id)

    def __eq__(self, other):
        return (self.uniprot == other.uniprot and
                self.go_id == other.go_id and
                self.value == other.value)

    def __hash__(self):
        return hash(self.uniprot + self.go_id + str(self.value))

    def to_json(self):
        return {
            'go_id': self.go_id,
            'implied_by': self.implied_by,
            'prob': self.prob,
            'uniprot': self.uniprot,
            'value': self.value
        }


@lru_cache()
def query_uniprots_with_evidence():
    assert effusion.dataset.cutoff_date

    session = effusion.dataset.database_session

    # interesting how sqlite lets me return an arbitrary row, so that I can return ORM mapped objects...
    # i'll try to do this more portably though
    # q = session.query(effusion.Uniprot).join(effusion.Annotation)
    # q = q.filter(Annotation.date < cutoff_date).group_by(effusion.Uniprot.id)


    # todo how can i use orm without a session. using a session when i'm not actually querying seems redundant
    q = session.query(distinct(effusion.Annotation.uniprot).label('uniprot'))
    q = q.filter(Annotation.date <= effusion.dataset.cutoff_date).filter(effusion.Annotation.value == True).filter(
        effusion.Annotation.source.in_(('direct', 'implied')))

    uniprots_with_evidence = {r.uniprot for r in q.all()}
    assert len(uniprots_with_evidence)
    return uniprots_with_evidence


def get_single_protein_conditional(event_state, given_state, weighted, require_mf, require_bp, add_pseudocounts,
                                   contingency_table=None):
    go_ids = set(chain(event_state.keys(), given_state.keys()))

    # todo assert provided contingeny table covers all the keys
    if contingency_table is None:
        contingency_table = get_single_contingency_table(go_ids,
                                                         weighted=weighted,
                                                         require_mf=require_mf,
                                                         require_bp=require_bp,
                                                         add_pseudocounts=add_pseudocounts)

    given_count = 0
    for state, count in contingency_table.iteritems():
        state_matches_given = all(state[k] == v for k, v in given_state.iteritems())
        if state_matches_given:
            given_count += count

    # todo make sure given count is significant for calculating statistics?

    joint_state = given_state.copy()
    joint_state.update(event_state)
    joint_count = 0
    for state, count in contingency_table.iteritems():
        state_matches_joint = all(state[k] == v for k, v in joint_state.iteritems())
        if state_matches_joint:
            joint_count += count

    # possible, e.g., p(8150 | 3674 = 0)
    # let's not give that a prob = 1 or 0.
    # with pseudocounts, 1/1 = 1, and that made \sum p(8150) and \sum p(3674) not always = 1!
    if given_count == 0:
        conditional = .5
    else:
        conditional = joint_count / given_count

    return conditional


def get_single_contingency_table(go_ids, weighted, require_mf, require_bp, add_pseudocounts):
    go_ids = tuple(sorted(go_ids))
    retval = get_single_contingency_table_impl(go_ids, weighted, require_mf, require_bp, add_pseudocounts=None)

    if add_pseudocounts:
        # get similar contingency table.
        # can we get similar go_ids?
        # strongly depends on relationship of go_ids. are they siblings? are they kin?
        # here's an idea: a child terms counts are a subset of parent terms counts. use parent table as pseudocount
        # would not work at all for p(term | siblings), since they all have the same parent.
        # that would be p(parent | parent, parent, parent) = 1
        # What to do if a term doesn't have any parents? don't add pseudocounts, i guess
        # what to do if a term has multiple parents? get mrca of all parents, i guess

        parent_id_to_child_ids_map = defaultdict(list)
        all_go_ids_have_a_parent = True
        for go_id in go_ids:
            go_vertex = effusion.dataset.gene_ontology.vertex4id(go_id)
            parent_go_ids = effusion.dataset.gene_ontology.get_parent_go_ids(go_vertex)

            if len(parent_go_ids) == 0:
                all_go_ids_have_a_parent = False
                break

            mrca_id = effusion.dataset.gene_ontology.get_mrca(parent_go_ids).go_id
            parent_id_to_child_ids_map[mrca_id].append(go_id)

        if all_go_ids_have_a_parent:
            # note this is a re-use of variable name
            parent_go_ids = parent_id_to_child_ids_map.keys()

            # this also uses pseudocounts.
            # otherwise, any leaf (which must have an annotation to be modeled) that has a parent with only 1 pairwise evidence would force a 00 or 11 for the root (and all its ancestors).
            # happened a lot
            parents_contingency_table = get_single_contingency_table(parent_go_ids, weighted, require_mf, require_bp,
                                                                     add_pseudocounts=True)

            # we don't want to modify the cached contingency table
            retval = retval.copy()
            for state_in_parents_contingency_table, pseudocount in parents_contingency_table.iteritems():
                state_in_contingency_table = {}
                for parent_go_id, val_of_parent_go_id in state_in_parents_contingency_table.iteritems():
                    for child_go_id in parent_id_to_child_ids_map[parent_go_id]:
                        state_in_contingency_table[child_go_id] = val_of_parent_go_id
                state_in_contingency_table = effusion.ImmutableDict(state_in_contingency_table)
                retval[state_in_contingency_table] = retval.get(state_in_contingency_table,
                                                                0) + effusion.args.pseudocount_weight * pseudocount

    return retval


@lru_cache(maxsize=None)
def get_single_contingency_table_impl(go_ids, weighted, require_mf, require_bp, add_pseudocounts):
    if add_pseudocounts:
        # this parameter is a trap!
        # This function is for caching the query. Access through get_single_contingency_table to add pseudocounts
        raise ValueError

    # go_ids = ('GO:0016787', 'GO:0016740')

    logging.debug('Getting single protein contingency table for %s' % str(go_ids))

    annotation_aliases = []
    ifnull_columns = []
    for go_id in go_ids:
        a = aliased(effusion.Annotation)
        annotation_aliases.append(a)
        ifnull_columns.append(func.ifnull(a.value, 0).label(go_id))

    q = effusion.dataset.database_session.query(effusion.DatasetUniprot.uniprot.label('uniprot'),
                                                literal_column('1').label('raw_count'),
                                                effusion.DatasetUniprot.ic.label('weighted_count'),
                                                *ifnull_columns)

    for go_id, a in zip(go_ids, annotation_aliases):
        q = q.outerjoin(a, and_(a.uniprot == effusion.DatasetUniprot.uniprot,
                                a.date <= effusion.dataset.cutoff_date,
                                a.source.in_(('direct', 'implied')),
                                a.go_id == go_id))
    q = q.filter(effusion.DatasetUniprot.dataset_name == effusion.dataset.dataset_cutoff_name)

    # can use this  to limit to terms with mf. can also join to 3674
    if require_mf:
        q = q.filter(effusion.DatasetUniprot.mf_ic > 0)
    if require_bp:
        q = q.filter(effusion.DatasetUniprot.bp_ic > 0)

    s = q.subquery()

    subquery_columns = []
    for go_id in go_ids:
        subquery_columns.append(s.c[go_id])

    q = effusion.dataset.database_session.query(func.sum(s.c.raw_count).label('raw_count'),
                                                func.sum(s.c.weighted_count).label('weighted_count'),
                                                *subquery_columns).select_from(s).group_by(*subquery_columns)

    # print q
    retval = {}
    for row in q:
        if weighted:
            count = row[1]
        else:
            count = row[0]
        state = row[2:]
        go_state = effusion.ImmutableDict({go_id: bool(s) for go_id, s in zip(go_ids, state)})
        retval[go_state] = count

    return retval


def get_num_uniprots_with_evidence(require_mf=False, require_bp=False):
    assert effusion.dataset.cutoff_date

    q = effusion.dataset.database_session.query(distinct(effusion.Annotation.uniprot)).filter(
        effusion.Annotation.date <= effusion.dataset.cutoff_date).filter(
        effusion.Annotation.value == True).filter(effusion.Annotation.source.in_(('direct', 'implied')))

    # todo this is a lot faster:
    # select count(distinct(uniprot)) from dataset_uniprots du where dataset_name = 'training';

    if require_mf or require_bp:
        raise NotImplementedError

    return q.count()
