from __future__ import division

from itertools import chain, product

import effusion


class AnnotationVariable(object):
    def __init__(self, uniprot, go_id, go_term):
        self.uniprot = uniprot
        self.go_id = go_id
        self.go_term = go_term

    @property
    def name(self):
        return 'AV:%s.%s' % (self.uniprot, self.go_id)

    def __str__(self):
        return str((self.uniprot, self.go_id))

    def __repr__(self):
        return self.__str__()

    def scope(self):
        return (True, False)


class AnnotationFactor(object):
    def __init__(self, this_go_variable, label, go_term_parents, protein_network_parents, edge_type,
                 relevant_or_interont):
        self.this_go_variable = this_go_variable
        self.label = label
        self.go_term_parents = go_term_parents
        self.protein_network_parents = protein_network_parents
        self.edge_type = edge_type
        self.relevant_or_interont = relevant_or_interont

    @property
    def variables(self):
        return list(chain(
            self.protein_network_parents if self.protein_network_parents else [],
            self.go_term_parents if self.go_term_parents else [],
            [self.this_go_variable]))

    def scope(self, lsb='last'):
        retval = []
        variables = self.variables
        scopes_of_variables = [v.scope() for v in variables]
        # from itertools.product doc: The nested loops cycle like an odometer with the rightmost element advancing on every iteration.
        # aka, the last is the least significant bit. this is the default
        if lsb == 'first':
            scopes_of_variables.reverse()
        for state in product(*scopes_of_variables):
            assert len(variables) == len(state)
            state_dict = {v: s for v, s in zip(variables, state)}
            retval.append(state_dict)
        return retval

    def psi(self, ignore_evidence, state):
        retval = 0
        zero = 0
        # zero = 1e-12

        val_of_v = state[self.this_go_variable]

        if effusion.args.top_down_or_bottom_up == 'top_down':
            all_kin_vals = all(state[x] for x in self.go_term_parents)
            invalid_state = val_of_v is True and all_kin_vals is False
            definite_state = val_of_v is False and all_kin_vals is False
        else:
            # actually, it's any_child_vals here
            any_kin_val = any(state[x] for x in self.go_term_parents)
            invalid_state = val_of_v is False and any_kin_val is True
            definite_state = val_of_v is True and any_kin_val is True

        # todo: maybe move this into contingency table
        if invalid_state:
            # retval = math.log(zero)
            retval = zero
        elif definite_state:
            retval = 1 - zero
        else:
            if self.relevant_or_interont == 'relevant' and self.edge_type is not None:  # we need pnp contingeny tables
                # could get rid of a few queries by doing marginalization myself
                # pnp contingency tables
                event_state = {}
                given_state_for_this_protein = {}
                given_state_for_parent_protein = {}
                for v in self.variables:
                    if v == self.this_go_variable:
                        event_state[v.go_id] = state[v]
                    else:
                        if v.uniprot == self.this_go_variable.uniprot:
                            given_state_for_this_protein[v.go_id] = state[v]
                        else:
                            given_state_for_parent_protein[v.go_id] = state[v]

                # todo we are supposed to limit counting to proteins that have some this_var.ontology annotation, right?
                conditional = effusion.get_pairwise_protein_conditional(
                    event_state,
                    given_state_for_this_protein,
                    given_state_for_parent_protein,
                    edge_type=self.edge_type,
                    weighted=not effusion.args.no_weighted_counts,
                    require_mf=(self.this_go_variable.go_term.aspect == effusion.GoAspectEnum.molecular_function),
                    require_bp=(self.this_go_variable.go_term.aspect == effusion.GoAspectEnum.biological_process),
                    add_pseudocounts=True
                )

            elif self.relevant_or_interont == 'interont' or self.edge_type is None:
                event_state = {}
                given_state = {}
                for v in self.variables:
                    if v == self.this_go_variable:
                        event_state[v.go_id] = state[v]
                    else:
                        given_state[v.go_id] = state[v]

                # todo we are supposed to limit counting to proteins that are co-annotated, right?
                conditional = effusion.get_single_protein_conditional(event_state,
                                                                      given_state,
                                                                      weighted=not effusion.args.no_weighted_counts,
                                                                      require_mf=True,
                                                                      require_bp=True,
                                                                      add_pseudocounts=True)

            else:
                raise

            # todo xxx
            # possible, e.g., p(8150 | 3674 = 0)
            # let's not give that a prob = 1 or 0.
            # with pseudocounts, 1/1 = 1, and that made \sum p(8150) and \sum p(3674) not always = 1!
            # if given_count == 0:
            #     conditional = .5
            # else:
            #     conditional = joint_count / given_count

            retval = conditional

        # note that probs that don't match will be driven to zero, but probs that do match will remain the same / not driven to one
        if not ignore_evidence:
            if self.label is not None and self.label != val_of_v:
                # retval = math.log(zero)
                retval = zero
            pass

        return retval
