from __future__ import division

import logging
from itertools import chain

import igraph

import effusion
from annotation_variable import AnnotationVariable, AnnotationFactor


class BayesNet(igraph.Graph):

    @classmethod
    @effusion.log_time()
    def learn_bayes_net(cls,
                        dataset,
                        protein_network,
                        annotations_dict,
                        model_bp,
                        query_info,
                        build_pgm=True):  # dict w/ inference_class

        # step n: calculate the label space / gene ontology subgraph. optimization that avoids using all go terms
        candidate_go_ontology = cls.get_candidate_go_ontology(annotations_dict, dataset, model_bp, protein_network)

        # collecting info
        candidate_mf_ontology = dataset.gene_ontology.subgraph_for_aspect(effusion.GoAspectEnum.molecular_function)
        query_info['candidate_ontology_mf_terms'] = len(candidate_mf_ontology.vs)
        query_info['candidate_ontology_mf_leaves'] = len(candidate_mf_ontology.vs.select(_outdegree=0))
        query_info['candidate_ontology_mf_ic'] = candidate_mf_ontology.get_ic()
        query_info['candidate_ontology_mf_gop_degree_max'] = max(candidate_mf_ontology.indegree())
        query_info['candidate_ontology_mf_goc_degree_max'] = max(candidate_mf_ontology.outdegree())

        if not effusion.args.no_write_visual_files:
            go_name_components = filter(None, [effusion.args.experiment_name, 'template.go.eps'])
            go_name = '.'.join(go_name_components)
            candidate_go_ontology.plot(go_name)

        # step n: convert annotations to labels
        labels_dict = {}

        query_info['num_proteins_with_pos_nonroot_evidence'] = 0
        query_info['num_proteins_with_pos_evidence'] = 0
        query_info['num_proteins_with_evidence'] = 0

        for uniprot, annotations in annotations_dict.items():
            if effusion.args.add_inferred_negative_annotations == 'moonlight':
                if effusion.args.top_down_or_bottom_up == 'bottom_up':
                    raise NotImplementedError
                assert effusion.args.top_down_or_bottom_up == 'top_down'

                implicitly_negative_annotations = effusion.Annotation.get_all_implicitly_negative_annotations(
                    gene_ontology=candidate_go_ontology,
                    uniprot=uniprot,
                    annotations=annotations)
                logging.debug('implicitly moon negative annotations: %s' % implicitly_negative_annotations)
            elif effusion.args.add_inferred_negative_annotations == 'uif':
                implicitly_negative_annotations = effusion.Annotation.get_uif_annotations(
                    gene_ontology=dataset.gene_ontology,
                    annotations=annotations)
            elif effusion.args.add_inferred_negative_annotations == 'none':
                implicitly_negative_annotations = []
            else:
                raise ValueError

            label = effusion.Label(gene_ontology=candidate_go_ontology,
                                   all_annotations=chain(annotations, implicitly_negative_annotations))
            labels_dict[uniprot] = label

            num_terms_with_positive_evidence = sum(label.values())
            query_info['num_proteins_with_pos_nonroot_evidence'] += int(
                num_terms_with_positive_evidence > 1)  # root + at least one more
            query_info['num_proteins_with_pos_evidence'] += int(num_terms_with_positive_evidence > 0)
            query_info['num_proteins_with_evidence'] += int(len(label) > 0)

            # add label to network
            try:
                protein_network.vs.find(uniprot)['label'] = ['%s:%s' % (k, '+' if v else '-') for k, v in label.items()]
                # print uniprot, str(label.get('GO:0016740')), str(label.get('GO:0016853'))
            except ValueError as e:
                # Evidence for vertex not in networks. Probably a disconnected component.
                pass

        # step n: decide which correlations to model

        if not effusion.args.no_write_visual_files:
            candidate_go_ontology.plot('go.svg')

        # step n: calculate MST and prune of protein network. optimization that avoids cycles and useless proteins
        # todo(jyunes) maybe storing queries as a list would be more consistent with how I store evidence?
        important_node_names = {v['name'] for v in protein_network.vs.select(query=True)}
        important_node_names |= {uniprot for uniprot, label in labels_dict.items() if len(label) > 0}
        protein_network.make_tractable(important_node_names=important_node_names)

        logging.info('Num vertices in reduced_network:%d' % len(protein_network.vs))

        # step n: learn free parameters. this inits a factor tree with no free params + runs inference
        # learning_engine = effusion.ScipyLearningEngine(**learning_params)
        factor_tree_params = {'reduced_network': protein_network,
                              'labels_dict': labels_dict,
                              'gene_ontology_subgraph': candidate_go_ontology,
                              'model_bp': model_bp,
                              }

        # todo(jyunes) separate this out do don't need conditional. used to explore network but don't need pgm
        factor_tree = None
        if build_pgm:
            factor_tree = BayesNet(**factor_tree_params)

        return factor_tree, protein_network

    @classmethod
    def get_candidate_go_ontology(cls, annotations_dict, dataset, model_bp, protein_network):
        # todo possibly exclude negative annots when computing candidates
        candidate_go_terms = set()
        proteins_in_network = set(protein_network.vs['name'])
        for uniprot, annotations in annotations_dict.items():
            if uniprot not in proteins_in_network:
                continue

            for a in annotations:
                go_term = dataset.gene_ontology.term4id(a.go_id)
                if go_term.aspect == effusion.GoAspectEnum.molecular_function:
                    candidate_go_terms.add(go_term)
                elif model_bp and go_term.aspect == effusion.GoAspectEnum.biological_process:
                    raise NotImplementedError
                    # if a.go_id not in metabolic_descendant_go_ids:
                    #     continue
                    candidate_go_terms.add(go_term)
        candidate_go_ontology = dataset.gene_ontology.subgraph(frozenset({t.go_id for t in candidate_go_terms}),
                                                               keep_ancestors=True)

        # in order to avoid high max out degree:
        # find a tree with no nodes of out degree > k, with minimum nodes removed
        # either add nodes unless they violate, or remove violating nodes
        # constraint: removing an edge to a target node removes all descendant nodes
        # todo
        # if 'limit_bp' == 'limit_bp':
        #     if model_bp:  # otherwise max(degrees) is max of empty sequence
        #         max_degree = 4
        #         bp_subgraph = self.subgraph(frozenset(bp_leaves), keep_ancestors=True)
        #         get_descendants = bp_subgraph.get_descendants.__wrapped__
        #
        #         degrees = bp_subgraph.degree(mode="OUT")
        #         while max(degrees) > max_degree:
        #             # find the first node that violates the condition
        #             # degrees = bp_subgraph.degree(mode="OUT")
        #             bfs = bp_subgraph.bfs(vid=0, mode="OUT")[0]
        #             vid_of_first_node_that_violates = None
        #             for node in bfs:
        #                 curr_degree = degrees[node]
        #                 if curr_degree > max_degree:
        #                     vid_of_first_node_that_violates = node
        #                     break
        #
        #             # determine for which edge, the target node has the fewest descendants
        #             # get descendants per edge
        #             out_edges = bp_subgraph.vs[vid_of_first_node_that_violates].out_edges()
        #             descendants_per_edge = {e: len(bp_subgraph.get_descendants(e.target_vertex)) for e in out_edges}
        #             out_edges.sort(key=lambda e: descendants_per_edge[e], reverse=True)
        #             edges_to_delete = out_edges[max_degree:]  # keep the first `max_degree` edges, delete the rest
        #             vids_to_delete = set()
        #             for e in edges_to_delete:
        #                 vids_to_delete |= {v.index for v in get_descendants(bp_subgraph, e.target_vertex)}
        #             bp_subgraph.delete_vertices(vids_to_delete)
        #
        #             degrees = bp_subgraph.degree(mode="OUT")
        #
        #         bp_leaves = set(bp_subgraph.vs.select(_outdegree=0)['name'])  # because leaves are our interface

        # todo(jyunes) research: could continue to add correlated terms...

        return candidate_go_ontology

    def __init__(self,
                 reduced_network,
                 labels_dict,
                 gene_ontology_subgraph,
                 model_bp,
                 *args,
                 **kwargs):

        # saved for general interest
        self.reduced_network = reduced_network.copy()
        self.labels_dict = labels_dict
        self.gene_ontology_subgraph = gene_ontology_subgraph
        self.model_bp = model_bp
        # self.params = effusion.ImmutableDict(params)  # shared potential functions -> can't just change the vals of params

        # for testing: deletes all evidence after network was pruned
        # labels_dict = {}

        super(BayesNet, self).__init__(*args, **kwargs)

        self.reduced_network.es.select(MST=False).delete()
        root_protein_index = 0
        # root_protein_index = reduced_network.vs.find('Q3AEF6').index
        self.reduced_network = effusion.as_directed(self.reduced_network, root_protein_index)
        # todo this is all terrible
        # let's copy directed edge info to original undirected network so we can display arrow heads
        for e_in_our_network in self.reduced_network.es:
            # search by protein names rather than vertices. don't know if indices are the same because at one point they had different vertices (param reduced network pruned)
            source_vertex_in_input_network = reduced_network.vs.find(name=e_in_our_network.source_vertex['name'])
            target_vertex_in_input_network = reduced_network.vs.find(name=e_in_our_network.target_vertex['name'])

            # we can't find edge by source_vertex, so get the index
            source_vertex_index_in_input_network = source_vertex_in_input_network.index
            target_vertex_index_in_input_network = target_vertex_in_input_network.index

            # even though the input network is undirected, we have to get the correct order of source and target
            e_in_input_network = reduced_network.es.find(
                _within=[source_vertex_index_in_input_network, target_vertex_index_in_input_network])

            # now just mark the undirected edge with its direction
            if e_in_input_network.source_vertex['name'] == e_in_our_network.source_vertex['name']:
                e_in_input_network['direction_in_mst'] = 'forward'
            else:
                e_in_input_network['direction_in_mst'] = 'reverse'

        sorted_mf_vertices = sorted(
            self.gene_ontology_subgraph.vs.select(
                lambda x: x['term'].aspect == effusion.GoAspectEnum.molecular_function),
            key=lambda x: x['term'].depth, reverse=(effusion.args.top_down_or_bottom_up == 'bottom_up'))
        sorted_bp_vertices = sorted(
            self.gene_ontology_subgraph.vs.select(
                lambda x: x['term'].aspect == effusion.GoAspectEnum.biological_process),
            key=lambda x: x['term'].depth, reverse=(effusion.args.top_down_or_bottom_up == 'bottom_up'))

        for protein_vid in self.reduced_network.bfs(root_protein_index)[0]:
            protein_vertex = self.reduced_network.vs[protein_vid]
            # for ontology in protein_template?:


            protein_network_parents = protein_vertex.predecessors()
            assert len(protein_network_parents) in {0, 1}
            edge_type = None
            sorted_relevant_vertices = sorted_mf_vertices

            if len(protein_vertex.in_edges()) == 1:
                edge_type = protein_vertex.in_edges()[0]['type']

                if edge_type == 'blast' or not model_bp:
                    pass  # sorted_mf_vertices is the default
                elif edge_type == 'string':
                    sorted_relevant_vertices = sorted_bp_vertices
                else:
                    raise

            # create the vertices even if there is no protein network parent
            self.link_relevant_vertices_to_protein_parent(edge_type, labels_dict, protein_network_parents,
                                                          protein_vertex, sorted_relevant_vertices)

            if model_bp:
                if edge_type is None or edge_type == 'blast':
                    sorted_interont_vertices = sorted_bp_vertices
                elif edge_type == 'string':
                    sorted_interont_vertices = sorted_mf_vertices
                else:
                    raise

                self.link_interont_vertices(edge_type, labels_dict, protein_vertex,
                                            sorted_interont_vertices)

    def link_relevant_vertices_to_protein_parent(self, edge_type, labels_dict, protein_network_parents,
                                                 protein_vertex, sorted_relevant_vertices):
        for go_term_vertex in sorted_relevant_vertices:
            go_term = go_term_vertex['term']
            go_id = go_term.go_id

            if effusion.args.top_down_or_bottom_up == 'top_down':
                go_kin = go_term_vertex.predecessors()
            elif effusion.args.top_down_or_bottom_up == 'bottom_up':
                go_kin = go_term_vertex.successors()
            else:
                raise

            # num_parents = len(go_parents) + len(protein_network_parents)
            # print name, num_parents

            # I'm still using a factor tree inference algorithm, so create variables and factors
            # add a potential for each (TERM | parents)

            label = labels_dict.get(protein_vertex['name'])
            if label is not None:
                label = label.get(go_id)
            # label = None

            annotation_variable = AnnotationVariable(protein_vertex['name'], go_id, go_term)
            self.add_vertex(name='AV:%s.%s' % (protein_vertex['name'], go_id),
                            type='variable', variable_type='annotation', protein=protein_vertex['name'],
                            go_id=go_id, aspect=go_term.aspect, variable=annotation_variable)
            # one factor for each node and its parents

            # name is optimization. it's indexed
            vertices_for_go_parents = [
                self.vs.find(
                    name='AV:%s.%s' % (protein_vertex['name'], v['name']))
                for v in go_kin]
            vertices_for_go_parents.sort(key=lambda x: x.index)  # sorted for opengm
            variables_of_go_parents = [v['variable'] for v in vertices_for_go_parents]

            # # experiment: reduce double counting of duplicate edges between hierarchies. only link children? but they might have counts that are too small...
            # vertices_of_go_children = go_term_vertex.successors()
            # if len(vertices_of_go_children) == 0:  # experiment: only link children!
            # check independence assumptions. removing edges just adds more assumptions. maybe the only problem before wasn't double counting, but too many poor counts?
            variables_of_protein_network_parent = []
            if len(protein_network_parents):
                variables_of_protein_network_parent.append(
                    self.vs.find(name='AV:%s.%s' % (protein_network_parents[0]['name'], go_id))['variable'])

            annotation_factor = AnnotationFactor(annotation_variable,
                                                          label,
                                                          variables_of_go_parents,
                                                          variables_of_protein_network_parent,
                                                          edge_type=edge_type,
                                                          relevant_or_interont='relevant')
            self.add_vertex(name='AF:%s.%s' % (protein_vertex['name'], go_id),
                            type='factor',
                            factor=annotation_factor)

    def link_interont_vertices(self, edge_type, labels_dict, protein_vertex,
                               sorted_interont_vertices):
        if not sorted_interont_vertices:
            return

        for go_term_vertex in sorted_interont_vertices:
            go_term = go_term_vertex['term']
            go_id = go_term.go_id

            # opengm requires they are sorted. by the order of self.variables.
            if effusion.args.top_down_or_bottom_up == 'top_down':
                go_parents = go_term_vertex.predecessors()
            elif effusion.args.top_down_or_bottom_up == 'bottom_up':
                go_parents = go_term_vertex.successors()
            else:
                raise

            interontological_parents = fly_interont_parents.get(go_id)
            variables_of_interontological_parent = []
            for p in interontological_parents:
                assert self.gene_ontology_subgraph.hasId(p)
                variables_of_interontological_parent.append(self.vs.find(
                    name='AV:%s.%s' % (protein_vertex['name'], p))['variable'])

            # num_parents = len(go_parents) + len(protein_network_parents)
            # print name, num_parents

            # I'm still using a factor tree inference algorithm, so create variables and factors
            # add a potential for each (TERM | parents)

            label = labels_dict.get(protein_vertex['name'])
            if label is not None:
                label = label.get(go_id)

            # TODO XXX just to confirm that BP adds nothing to reduced SS networks with no BP evidence
            # if protein_vertex['name'] != 'O01916':
            #     label = None
            # if go_id == 'GO:0008444':
            #     label = None
            # label = None

            annotation_variable = AnnotationVariable(protein_vertex['name'], go_id, go_term)
            self.add_vertex(name='AV:%s.%s' % (protein_vertex['name'], go_id),
                            type='variable',
                            variable_type='annotation',
                            protein=protein_vertex['name'],
                            go_id=go_id,
                            aspect=go_term.aspect,
                            variable=annotation_variable)
            # one factor for each node and its parents


            # using name kwargs is a shortcut, because it's indexed
            vertices_for_go_parents = [
                self.vs.find(name='AV:%s.%s' % (protein_vertex['name'], v['name']))
                for v in go_parents]
            vertices_for_go_parents.sort(key=lambda x: x.index)  # sorted for opengm
            variables_of_go_parents = [v['variable'] for v in vertices_for_go_parents]

            annotation_factor = AnnotationFactor(annotation_variable,
                                                          label,
                                                          variables_of_go_parents,
                                                          variables_of_interontological_parent,
                                                          edge_type=edge_type,
                                                          relevant_or_interont='interont')

            self.add_vertex(name='AF:%s.%s' % (protein_vertex['name'], go_id),
                            type='factor',
                            factor=annotation_factor)

    @property
    def variables(self):
        retval = self.vs.select(type='variable')['variable']

        return retval

    @property
    def factors(self):
        return self.vs.select(type='factor')['factor']


    def draw_factor_tree(self):
        layout = self.layout("kk")
        self.vs["label"] = self.vs["name"]
        shape_map = {'variable': 'circle', 'factor': 'rectangle'}
        self.vs["shape"] = [shape_map[v['type']] for v in self.vs]
        igraph.plot(self, layout=layout)
