__author__ = 'jeff'

import decimal
import json

import numpy

import effusion


# this is to format the number of decimal places without converting it to a string
# cite: http://stackoverflow.com/questions/1960516/python-json-serialize-a-decimal-object
class MyEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, decimal.Decimal):
            return float(o)
        elif isinstance(o, numpy.int32):
            return int(o)
        elif isinstance(o, set):
            return list(o)
        elif isinstance(o, tuple):
            return list(o)
        try:
            return super(MyEncoder, self).default(o)
        except:
            return -1


# TODO in the future, get display function with most information
def make_cytoscape_json_struct(name, attributes):
    retval = {'data': {'id': name}}

    for attribute_name, attribute_val in attributes.items():
        # todo(jyunes) report bug to cytoscape
        if attribute_val is None:  # cytoscape can't handle nulls
            continue
        if 'items' in dir(attribute_val):
            attribute_dict = attribute_val
            for attribute_key, attribute_val in attribute_dict.items():
                retval['data']['%s: %s' % (attribute_name, attribute_key)] = attribute_val
        else:
            retval['data'][str(attribute_name)] = attribute_val

    return retval


def make_cyjs_edge(name, source, target, attributes):
    retval = {'data': {'id': name,
                       'source': str(source),
                       'target': str(target)
                       }}

    for attribute_name, attribute_val in attributes.items():
        # todo(jyunes) report bug to cytoscape
        if attribute_val is None:  # cytoscape can't handle nulls
            continue

        retval['data'][attribute_name] = attribute_val

    return retval


def export_graph_as_cyjs(g, path):
    assert g.is_simple()

    g = g.copy()

    # # TODO(jyunes) edges may be frankensteins. but the algorithm uses the 1 correct mst edge i think maybe
    # g.to_undirected(mode='each')
    # g.simplify(multiple=True,
    #            loops=True,
    #            combine_edges={'MST': max,
    #                           'type': lambda x: str.join('+', set(x)),
    #                           'score': max,
    #                           'evalue': min,
    #                           'round_added': min,
    #                           'direction_in_mst': max})  # min will return None

    add_interesting_attribute(g)

    elements_dict = {'data': {},
                     'elements': {
                         'nodes': [],
                         'edges': []
                     }}

    for node in g.vs:
        attributes = node.attributes()
        attributes = filter_attributes_for_cyjs(attributes)
        node_dict = make_cytoscape_json_struct(node['name'], attributes)
        elements_dict['elements']['nodes'].append(node_dict)

    for edge in sorted(g.es, key=lambda x: x['score'], reverse=True):
        attributes = edge.attributes()
        source = g.vs[edge.source]
        target = g.vs[edge.target]
        name = None
        cyjs_edge = make_cyjs_edge(name, source['name'], target['name'], attributes)
        elements_dict['elements']['edges'].append(cyjs_edge)

        if len(elements_dict['elements']['edges']) >= effusion.config.max_display_edges:
            break

    elements_json = json.dumps(elements_dict, cls=MyEncoder, indent=4)
    f = open(path, 'w+')
    f.write(elements_json)
    f.close()


def export_factor_graph_as_cyjs(factor_graph, path):
    elements_dict = {'data': {},
                     'elements': {
                         'nodes': [],
                         'edges': []
                     }}

    for node in factor_graph.vs.select(type='variable'):
        attributes = node.attributes()
        attributes = filter_attributes_for_cyjs(attributes)
        node_dict = make_cytoscape_json_struct(node['name'], attributes)
        elements_dict['elements']['nodes'].append(node_dict)

    for factor_vertex in factor_graph.vs.select(type='factor'):
        attributes = factor_vertex.attributes()
        attributes = filter_attributes_for_cyjs(attributes)
        node_dict = make_cytoscape_json_struct(factor_vertex['name'], attributes)
        elements_dict['elements']['nodes'].append(node_dict)

        factor = factor_vertex['factor']
        edge_list = []
        for factor_variable in factor.variables:
            edge = make_cyjs_edge('', factor_vertex['name'], factor_variable.name, {'type': 'SS'})
            edge_list.append(edge)

        elements_dict['elements']['edges'].extend(edge_list)

    elements_json = json.dumps(elements_dict, cls=MyEncoder, indent=4)
    f = open(path, 'w+')
    f.write(elements_json)
    f.close()


def filter_attributes_for_cyjs(attributes):
    my_filter = lambda k: k in ['protein'] or 'unnormalized' in k
    retval = {k: v for k, v in attributes.items() if not my_filter(k)}

    return retval


def add_interesting_attribute(protein_network):
    interestingness_map = {'query': 2, 'evidence': 1, 'example': 2}
    for node in protein_network.vs:
        interestingness = 0
        for interesting_attribute, interestingness in interestingness_map.items():
            if node.attributes().get(interesting_attribute):
                interestingness = max(interestingness, node.attributes().get(interesting_attribute))
        node['interesting'] = interestingness


@effusion.log_time()
def add_marginals_as_attribute(inf_engine, protein_network):
    factor_tree = inf_engine.factor_tree

    for node in protein_network.vs:
        for go_vertex in factor_tree.gene_ontology_subgraph.vs:
            matching_variable_in_factor_graph = \
                factor_tree.vs.find(protein=node['name'], go_id=go_vertex['name'], type='variable',
                                    variable_type=effusion.config.display_p_annotation_or_function)['variable']
            node[go_vertex['name']] = inf_engine.get_marginal(matching_variable_in_factor_graph, True)


def copy_node_attrs(g1, g2):
    for node in g1.vs:
        matching_nodes_in_g2 = g2.vs.select(name=node['name'])
        if len(matching_nodes_in_g2) == 0:
            continue
        assert len(matching_nodes_in_g2) == 1

        corresponding_node_in_g2 = matching_nodes_in_g2[0]
        corresponding_node_in_g2.update_attributes(node.attributes())
