from __future__ import division

import datetime
import json
import logging
import sys

import os
from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker

from effusion.gene_ontology import GeneOntology

# todo maybe rename phase? role seems more like a phase
phases = ['discovery', 'test', 'prediction']
role_names = ['training', 'evaluation']
dataset_partition = ['training', 'validation', 'test']

# which partitions do we use for each role during a given phase?
# still seems a bit complicated, given we just have training date < validation date
# data may include multiple partitions, but since we only represent (and have params for) our datasets by cutoffs
# only use the latest dataset, and we'll get the cutoff date based on that. AKA think of the values as dates
roles = {
    'discovery': {
        'training': 'training',
        'evaluation': 'validation'
    },
    'test': {
        'training': 'validation',
        'evaluation': 'test'
    },
    'prediction': {
        'training': 'test',
        'evaluation': 'test'  # evaluation data starts at test and ends at test (aka all the data)
    },
}

# Sick and tired of PyCharm warning!
path = None
database_session = None
cutoff_dates = None
cutoff_date = None
dataset_cutoff_name = None
gene_ontology = None
networks_dir = None
preds_dir = None
go_hierarchy_params = None
evaluation_proteins = None
goa_diamond_db = None
goa_blast_db = None
num_uniprots_with_evidence = None
phase = None
role = None


def init(path):
    """

    :param path:
    :param is_toy: whether it should use a fake toy in memory database
    :return:
    """
    if path is None or not os.path.isdir(path):
        print path
        raise ValueError

    data_metadata = json.load(open(os.path.join(path, 'data.json')))

    cutoff_dates = data_metadata['cutoff_dates']
    for phase, date_str in cutoff_dates.iteritems():
        if date_str is None:
            date = datetime.date.today()
        else:
            date = datetime.datetime.strptime(date_str, '%Y-%m-%d').date()
        cutoff_dates[phase] = date

    go_obo = os.path.join(path, 'go.obo')
    gene_ontology = GeneOntology(go_obo, catalytic_id=data_metadata['catalytic_id'])

    engine = create_engine("sqlite://")
    engine.execute("pragma temp_store = 2")
    engine.execute("attach database '%s' as goa_db" % os.path.join(path, 'goa.db'))
    engine.execute("attach database '%s' as string_db" % os.path.join(path, 'string.db'))
    engine.execute("attach database '%s' as pnp_db" % os.path.join(path, 'protein_network_parents.db'))
    session_factory = sessionmaker(bind=engine)
    database_session = scoped_session(session_factory)

    networks_dir = os.path.join(path, 'networks')
    preds_dir = os.path.join(path, 'preds')

    # todo(jyunes) Haven't yet decided if I want an object or just the module
    class Dataset(object):
        def __init__(self, **kwargs):
            self.__dict__.update(kwargs)

    dataset = Dataset(path=path,
                      database_session=database_session,
                      cutoff_dates=cutoff_dates,
                      cutoff_date=cutoff_date,
                      gene_ontology=gene_ontology,
                      networks_dir=networks_dir,
                      preds_dir=preds_dir,
                      )

    this_module = sys.modules[__name__]
    this_module.__dict__.update(dataset.__dict__)

    # if dataset_name is not None:
    #     set_phase(dataset_name)


def set_phase_and_role(phase, role):
    assert phase

    this_module = sys.modules[__name__]

    this_module.phase = phase
    this_module.role = role

    evaluation_proteins = []
    try:
        # todo this should be a function shared by the script
        eval_proteins_path = os.path.join(effusion.args.data_dir, 'evaluation_proteins.%s.txt' % phase)
        with open(os.path.join(path, eval_proteins_path)) as f:
            for line in f:
                line = line.strip()
                if line == '':
                    continue
                evaluation_proteins.append(line)
    except IOError:
        logging.warning('evaluation_proteins not setup. Run choose_evaluation_proteins.py')
    this_module.evaluation_proteins = evaluation_proteins

    # todo don't know what to call this...
    this_module.dataset_cutoff_name = roles[phase][role]
    this_module.cutoff_date = cutoff_dates[this_module.dataset_cutoff_name]
    # todo these should have been / sort of were made based on phases rather than partitions
    # i.e., these dbs named by partition are really all data with a cutoff of each partition (aka, training + validation)
    # (still a thought in progress)
    this_module.goa_blast_db = os.path.join(effusion.args.data_dir, 'goa.%s' % this_module.dataset_cutoff_name,
                                            'goa.%s' % this_module.dataset_cutoff_name)
    this_module.goa_diamond_db = os.path.join(effusion.args.data_dir,
                                              'diamond_goa.%s.dmnd' % this_module.dataset_cutoff_name)

    # make sure the GoTerms are attached to the session. that way, when we do annotation.go_term, it will return this object, with these additional attributes.
    # todo not sure if sqlalchemy will ever return new instances without these attributes...
    gene_ontology.load_from_db(database_session)
    this_module.num_uniprots_with_evidence = effusion.get_num_uniprots_with_evidence()

    # set uniform prior, or there will be predictions with terms with infinite IC
    for v in effusion.dataset.gene_ontology.vs:
        v['term'].marginal_ic = 1 / this_module.num_uniprots_with_evidence
        v['term'].conditional_ic = 1 / this_module.num_uniprots_with_evidence

    q = database_session.query(effusion.DatasetGoTerm).filter(
        effusion.DatasetGoTerm.dataset_name == this_module.dataset_cutoff_name)
    for dataset_go_term in q:
        go_term = gene_ontology.vs.find(dataset_go_term.go_id)['term']
        go_term.marginal_ic = dataset_go_term.marginal_ic
        go_term.conditional_ic = dataset_go_term.conditional_ic
