from __future__ import division

import logging
import math

from functools32 import lru_cache

import effusion


# todo move to gene ontology
def get_ic_for_subset(subset_type):
    if subset_type == 'catalytic':
        subset_terms = {v['term'] for v in effusion.dataset.gene_ontology.vs if v['term'].catalytic}
        root_id = effusion.dataset.gene_ontology.catalytic_id

    elif subset_type == 'molecular_function':
        subset_terms = {v['term'] for v in effusion.dataset.gene_ontology.vs if
                        v['term'].aspect == effusion.GoAspectEnum.molecular_function}
        root_id = effusion.dataset.gene_ontology.root_mf_id

    elif subset_type == 'biological_process':
        subset_terms = {v['term'] for v in effusion.dataset.gene_ontology.vs if
                        v['term'].aspect == effusion.GoAspectEnum.biological_process}
        root_id = effusion.dataset.gene_ontology.root_bp_id

    else:
        raise ValueError

    # we need to add the marginal IC of the root term, rather than the conditional ic
    # (alternatively, add all of the root terms parents)
    # this only really makes a difference for catalytic subset
    root_v = effusion.dataset.gene_ontology.vertex4id(root_id)
    subset_terms |= {v['term'] for v in effusion.dataset.gene_ontology.get_ancestors(root_v)}

    retval = sum(go_term.conditional_ic for go_term in subset_terms)

    return retval


def wru(T, P):
    numerator = 0
    denominator = 0

    for protein_name, T_i in T.items():
        P_i = P[protein_name]

        numerator += i(T_i) * ru(T_i, P_i)
        denominator += i(T_i)

    return numerator / denominator


def wmi(T, P):
    numerator = 0
    denominator = 0

    for protein_name, T_i in T.items():
        P_i = P[protein_name]

        numerator += i(T_i) * mi(T_i, P_i)
        denominator += i(T_i)

    return numerator / denominator


@lru_cache()
def i_old(T):
    retval = -math.log(Pr(T), 2)
    return retval


# @lru_cache()
def i(T):
    retval = 0
    for v in T:
        # todo(jyunes) this is a keyerror, so i'm using get, because i downloaded the new go data
        # todo(jyunes) assuming all everything label is true?
        try:
            go_term = effusion.dataset.gene_ontology.term4id(v)
            retval += go_term.conditional_ic
        except ValueError as e:
            logging.error('%s on %s' % (str(v), str(e)))

    return retval


@lru_cache(maxsize=None)
def p_term(go_id):
    go_annotation = effusion.Annotation(effusion.dataset.gene_ontology, uniprot='', go_id=go_id)
    upward_annotations = effusion.Annotation.get_all_implied_annotations(effusion.dataset.gene_ontology,
                                                                         {go_annotation})
    upward_go_ids = {a.go_id for a in upward_annotations}

    return Pr(upward_go_ids)


def Pr(T):
    retval = 1
    for v in T:
        # todo(jyunes) this is a keyerror, so i'm using get, because i downloaded the new go data
        # todo(jyunes) assuming all everything label is true?
        try:
            retval *= get_raw_ghp(v)
        except ValueError as e:
            logging.error('%s on %s' % (str(v), str(e)))

    return retval


def ia(v):
    retval = 0
    # todo(jyunes) this is a keyerror, so i'm using get, because i downloaded the new go data
    # todo xxx when accumulating correct or incorrect terms, do we want to add the conditional IC or the marginal IC? conditional IC?
    try:
        go_term = effusion.dataset.gene_ontology.term4id(v)
        retval = go_term.conditional_ic
    except ValueError as e:
        logging.error('%s on %s' % (str(v), str(e)))
    return retval


def ru(T, P):
    retval = 0
    for v in (T - P):
        retval += ia(v)

    return retval


def mi(T, P):
    retval = 0

    for v in (P - T):
        retval += ia(v)

    return retval


# These two accumulators are for the entire prediction set (opposed to per protein).
# They are intended to keep track of the numerator and denominator for calculating the weighted averages.
# These accumulators share code, but doesn't seem useful to generalize

class MethodWruAccumulator(object):
    def __init__(self, all_Ts):
        self.info_in_ground_truth = {protein_name: i(T) for protein_name, T in all_Ts.items()}  # optimization
        self.protein_accumulators = {protein_name: ProteinRuAccumulator(T) for protein_name, T in all_Ts.items()}

        self.numerator = 0
        self.denominator = 0
        for protein_name, T_i in all_Ts.items():
            ru_acc = self.protein_accumulators[protein_name]
            info_in_ground_truth = self.info_in_ground_truth[protein_name]
            self.numerator += info_in_ground_truth * ru_acc.ru
            self.denominator += info_in_ground_truth

    def accumulate(self, mop_item):
        info_in_ground_truth = self.info_in_ground_truth[mop_item.uniprot]
        ru_acc = self.protein_accumulators[mop_item.uniprot]

        # Subtract old value
        self.numerator -= info_in_ground_truth * ru_acc.ru

        # Update
        self.protein_accumulators[mop_item.uniprot].accumulate(mop_item.go_id)

        # Add new value
        self.numerator += info_in_ground_truth * ru_acc.ru

    def get_val(self):
        return self.numerator / self.denominator


class MethodWmiAccumulator(object):
    def __init__(self, all_Ts):
        self.info_in_ground_truth = {protein_name: i(T) for protein_name, T in all_Ts.items()}  # optimization
        self.protein_accumulators = {protein_name: ProteinMiAccumulator(T) for protein_name, T in all_Ts.items()}

        self.numerator = 0
        self.denominator = 0
        for protein_name, T_i in all_Ts.items():
            mi_acc = self.protein_accumulators[
                protein_name]  # umm... these are all zero i hope. but code is parallel to Wru
            info_in_ground_truth = self.info_in_ground_truth[protein_name]
            self.numerator += info_in_ground_truth * mi_acc.mi
            self.denominator += info_in_ground_truth

    def accumulate(self, mop_item):
        info_in_ground_truth = self.info_in_ground_truth[mop_item.uniprot]
        mi_acc = self.protein_accumulators[mop_item.uniprot]

        # Subtract old value
        self.numerator -= info_in_ground_truth * mi_acc.mi

        # Update
        self.protein_accumulators[mop_item.uniprot].accumulate(mop_item.go_id)

        # Add new value
        self.numerator += info_in_ground_truth * mi_acc.mi

    def get_val(self):
        return self.numerator / self.denominator


# These two accumulators are just counters that add (or subtract) the terms information content if it's in the protein.
# One accumulator is maintained per protein, so that when we calculated the weighted average over all proteins,
# we can update the accumulator for a protein after adding a new term to the prediction set.

class ProteinRuAccumulator(object):
    order = 'descending'

    def __init__(self, T):
        self.T = T
        self.ru = i(T)

    def accumulate(self, P_i):
        if P_i in self.T:
            self.ru -= ia(P_i)

    def get_val(self):
        return self.ru


class ProteinMiAccumulator(object):
    order = 'ascending'

    def __init__(self, T):
        self.T = T
        self.mi = 0

    def accumulate(self, P_i):
        if P_i not in self.T:
            self.mi += ia(P_i)

    def get_val(self):
        return self.mi


class MethodNruAccumulator(object):
    def __init__(self, all_Ts):
        self.info_in_ground_truth = {protein_name: i(T) for protein_name, T in all_Ts.items()}  # optimization
        self.protein_accumulators = {protein_name: ProteinNruAccumulator(T) for protein_name, T in all_Ts.items()}

    def accumulate(self, mop_item):
        self.protein_accumulators[mop_item.uniprot].accumulate(mop_item.go_id)

    def get_val(self):
        nru = 0
        for protein, protein_accumulator in self.protein_accumulators.items():
            nru += protein_accumulator.get_val()

        return nru / len(self.protein_accumulators)


class ProteinNruAccumulator(object):
    def __init__(self, T):
        self.T = T

        self.numerator = i(T)
        self.denominator = i(T)

    def accumulate(self, P_i):
        if P_i in self.T:
            self.numerator -= ia(P_i)

    def get_val(self):
        # for evaluation proteins that don't have any IC.
        # happens if we are evaluating catalytic function terms but looking at proteins that might not be catalytic
        # todo do this better? maybe filter out the non cat proteins? not really sure how normalization is affected...
        # this is probably correct. the remaining uncertainty is 0 for protein with no IC.
        if self.denominator == 0:
            return 0

        return self.numerator / self.denominator


class MethodNmiAccumulator(object):
    def __init__(self, all_Ts, total_ic):
        # self.protein_accumulators = {protein_name: ProteinNmiAccumulator(T, all_Ps[protein_name]) for protein_name, T in all_Ts.items()}

        # we want to normalize each protein's MI by the total IC for the terms we are evaluating (mf, catalytic, etc.)
        self.protein_accumulators = {protein_name: ProteinNmiAccumulator(T, total_ic) for protein_name, T in
                                     all_Ts.items()}

    def accumulate(self, mop_item):
        self.protein_accumulators[mop_item.uniprot].accumulate(mop_item.go_id)

    def get_val(self):
        nru = 0
        for protein, protein_accumulator in self.protein_accumulators.items():
            nru += protein_accumulator.get_val()

        return nru / len(self.protein_accumulators)


class ProteinNmiAccumulator(object):
    def __init__(self, T, total_ic):
        self.T = T

        self.numerator = 0

        # this demoninator gives "what percent of all terms ic did i predict incorrectly"
        # self.denominator = total_ic
        # this denominator gives closer to paper and 1-precision "what percent of *predicted* terms ic did i predict incorrectly"
        self.denominator = 0

    def accumulate(self, P_i):
        if P_i not in self.T:
            self.numerator += ia(P_i)

        # add this for second definition of this metric above. comment out for first (total_ic) metric
        self.denominator += ia(P_i)

    def get_val(self):
        # also correct. if we aren't predicting anything for this protein, weighted misinformation is 0.
        if self.denominator == 0:
            return 0
        return self.numerator / self.denominator


class MethodMiAccumulator(object):
    def __init__(self, all_Ts):
        self.protein_accumulators = {protein_name: ProteinMiAccumulator(T) for protein_name, T in all_Ts.items()}

    def accumulate(self, mop_item):
        self.protein_accumulators[mop_item.uniprot].accumulate(mop_item.go_id)

    def get_val(self):
        total_mi = 0
        for protein, protein_accumulator in self.protein_accumulators.items():
            total_mi += protein_accumulator.mi

        return total_mi / len(self.protein_accumulators)


class MethodRuAccumulator(object):
    def __init__(self, all_Ts):
        self.protein_accumulators = {protein_name: ProteinRuAccumulator(T) for protein_name, T in all_Ts.items()}

    def accumulate(self, mop_item):
        self.protein_accumulators[mop_item.uniprot].accumulate(mop_item.go_id)

    def get_val(self):
        total_ru = 0
        for protein, protein_accumulator in self.protein_accumulators.items():
            total_ru += protein_accumulator.ru

        return total_ru / len(self.protein_accumulators)


class ProteinPrecisionAccumulator(object):
    order = 'descending'

    def __init__(self, T):
        self.T = T

        self.numerator = 0
        self.denominator = 0

    def accumulate(self, P_i):
        if P_i in self.T:
            self.numerator += 1

        self.denominator += 1

    def get_val(self):
        # if we aren't predicting anything for this protein, precision is 1
        if self.denominator == 0:
            return 1
        return self.numerator / self.denominator


class ProteinFnAccumulator(object):
    order = 'descending'

    def __init__(self, T):
        self.T = T

        self.numerator = len(self.T)

    def accumulate(self, P_i):
        if P_i in self.T:
            self.numerator -= 1

    def get_val(self):
        return self.numerator


class ProteinRecallAccumulator(object):
    order = 'ascending'

    def __init__(self, T):
        self.T = T

        self.numerator = 0
        self.denominator = len(T)

    def accumulate(self, P_i):
        if P_i in self.T:
            self.numerator += 1

    def get_val(self):
        if self.denominator == 0:
            return 0

        return self.numerator / self.denominator


class ProteinFpAccumulator(object):
    order = 'ascending'

    def __init__(self, T):
        self.T = T

        self.numerator = 0

    def accumulate(self, P_i):
        if P_i not in self.T:
            self.numerator += 1

    def get_val(self):
        return self.numerator


class ProteinPAccumulator(object):
    order = 'ascending'

    def __init__(self, T):
        self.T = T

        self.numerator = 0

    def accumulate(self, P_i):
        self.numerator += 1

    def get_val(self):
        return self.numerator


class ProteinPICAccumulator(object):
    order = 'ascending'

    def __init__(self, T):
        self.T = T

        self.numerator = 0

    def accumulate(self, P_i):
        self.numerator += ia(P_i)

    def get_val(self):
        return self.numerator


class MethodPrecisionAccumulator(object):
    # ALTERNATIVELY, we could average the recalls for the individual protein classifiers. different metric

    def __init__(self, all_Ts):
        self.protein_accumulators = {protein_name: ProteinPrecisionAccumulator(T) for protein_name, T in all_Ts.items()}

    def accumulate(self, mop_item):
        self.protein_accumulators[mop_item.uniprot].accumulate(mop_item.go_id)

    def get_val(self):
        sum_tp = 0
        sum_predicted = 0

        for protein, protein_accumulator in self.protein_accumulators.items():
            sum_tp += protein_accumulator.numerator
            sum_predicted += protein_accumulator.denominator

        return sum_tp / sum_predicted


class MethodRecallAccumulator(object):
    # ALTERNATIVELY, we could average the recalls for the individual protein classifiers. different metric


    def __init__(self, all_Ts):
        self.protein_accumulators = {protein_name: ProteinRecallAccumulator(T) for protein_name, T in all_Ts.items()}

    def accumulate(self, mop_item):
        self.protein_accumulators[mop_item.uniprot].accumulate(mop_item.go_id)

    def get_val(self):
        sum_tp = 0
        sum_t = 0

        for protein, protein_accumulator in self.protein_accumulators.items():
            sum_tp += protein_accumulator.numerator
            sum_t += protein_accumulator.denominator

        return sum_tp / sum_t


class Measure(object):
    def __call__(self, *args, **kwargs):
        raise NotImplementedError

    def initial_value(self):
        return 0


class Constant(Measure):
    accumulate_over = None

    def initial_value(self):
        return 1


class TruePositive(Measure):
    accumulate_over = 'P'

    def __call__(self, P_i, T):
        return 1 if P_i in T else 0


class FalsePositive(Measure):
    accumulate_over = 'P'

    def __call__(self, P_i, T):
        return 1 if P_i not in T else 0


class PredictedConditionPositive(Measure):
    accumulate_over = 'P'

    def __call__(self, P_i, T):
        return 1


class ConditionPositive(Measure):
    accumulate_over = 'T'

    def __call__(self, T_i):
        return 1


def make_protein_accumulator_class(numerator_measure, denominator_measure, term_weighted, default_value):
    class ProteinAccumulator(object):
        def __init__(self, T):
            self.T = T

            self.numerator = numerator_measure.initial_value()
            if numerator_measure.accumulate_over == 'T':
                raise NotImplementedError

            self.denominator = denominator_measure.initial_value()
            if denominator_measure.accumulate_over == 'T':
                for T_i in T:
                    if term_weighted:
                        weight = ia(T_i)
                    else:
                        weight = 1
                    self.denominator += weight * denominator_measure(T_i)

        def accumulate(self, P_i):
            if term_weighted:
                weight = ia(P_i)
            else:
                weight = 1
            if numerator_measure.accumulate_over == 'P':
                self.numerator += weight * numerator_measure(P_i, self.T)
            if denominator_measure.accumulate_over == 'P':
                self.denominator += weight * denominator_measure(P_i, self.T)

        def get_val(self):
            if self.denominator == 0:  # precision is 1 if nothing predicted, but recall is 0
                return default_value

            return self.numerator / self.denominator

    return ProteinAccumulator


def make_method_accumulator_class(protein_accumulator_class,
                                  sample_weighted):
    class MethodAccumulator(object):
        def __init__(self, all_Ts):
            self.protein_accumulators = {protein_name: protein_accumulator_class(T) for protein_name, T in
                                         all_Ts.items()}
            if sample_weighted:
                self.weights = {protein_name: i(T) for protein_name, T in all_Ts.items()}
            else:
                self.weights = {protein_name: 1 for protein_name in all_Ts.keys()}

        def accumulate(self, mop_item):
            self.protein_accumulators[mop_item.uniprot].accumulate(mop_item.go_id)

        def get_val(self):
            numerator = 0
            denominator = 0

            for protein, protein_accumulator in self.protein_accumulators.items():
                numerator += self.weights[protein] * protein_accumulator.get_val()
                denominator += self.weights[protein]

            return numerator / denominator

        def debug_val(self):
            numerator = 0
            denominator = 0

            for protein, protein_accumulator in self.protein_accumulators.items():
                numerator += self.weights[protein] * protein_accumulator.get_val()
                denominator += self.weights[protein]

            print numerator, denominator, numerator / denominator

            return numerator / denominator

    return MethodAccumulator
