# cython: boundscheck=False

from __future__ import division

import logging
import resource
import tempfile
import time
from collections import defaultdict

import dai
import os
import subprocess32 as subprocess

import effusion


class InferenceEngine(object):
    def __init__(self, factor_tree):
        self.factor_tree = factor_tree

    def infer(self):
        raise NotImplementedError

    def get_marginal(self, i, xi):
        raise NotImplementedError

    def get_go_term_marginals(self, variable):
        go_term_marginals = defaultdict(int)

        for xi in variable.scope():
            marginal = self.get_marginal(variable, xi)

            for go_id, value in xi.items():
                if value:
                    go_term_marginals[go_id] += marginal

        return go_term_marginals

    def get_energy(self):
        raise NotImplementedError


class BayesNetInferenceEngine(InferenceEngine):
    @effusion.log_time(name='BayesNetInferenceEngine.__init__')
    def __init__(self, factor_tree):
        super(BayesNetInferenceEngine, self).__init__(factor_tree)

        (self.gm,
         self.protein_to_opengm_index_map) = self.get_libdai(factor_tree)
        # accumulator = 'logsumexp'
        # accumulator = 'integrator'
        # param = opengm.InfParam(steps=100000, useNormalization=self.core_normalization, damping=.1, convergenceBound=1e-9)
        # param = opengm.InfParam(steps=100000, damping=.01, convergenceBound=1e-9)

        # self.solver = opengm.inference.BeliefPropagation(gm=self.gm,
        #                                                  accumulator=accumulator,
        #                                                  parameter=param)
        # self.solver = opengm.inference.TreeReweightedBp(gm=self.gm,
        #                                                  accumulator=accumulator,
        #                                                  parameter=param)
        # self.solver = opengm.inference.Bruteforce(gm=self.gm, accumulator='logsumexp')

        # brute_param = opengm.InfParam()
        # self.solver = opengm.inference.Bruteforce(gm=self.gm, accumulator=accumulator, parameter=brute_param)

    def get_libdai(self, factor_tree):
        logging.debug('Making libdai...')

        ignore_evidence = False
        myvar_to_daivar_map = {v: dai.Var(vi, len(v.scope())) for vi, v in enumerate(factor_tree.variables)}

        # todo could use calcState() and related functions of libdai to do this

        dai_factors = dai.VecFactor()
        for factor in factor_tree.factors:
            if isinstance(factor, effusion.AnnotationFactor):
                factor_variables = factor.variables

                # with this commented, 0 is False, 1 is true.
                # scopes_of_variables = [list(v.scope()) for v in factor_variables]

                varset = dai.VarSet()
                for myvar in factor_variables:
                    daivar = myvar_to_daivar_map[myvar]
                    varset.append(daivar)

                daifactor = dai.Factor(varset)

                state = dai.State(varset)
                while state.valid():
                    # with this bool, 0 is False, 1 is true.
                    label = {myvar: bool(state(myvar_to_daivar_map[myvar])) for myvar in factor_variables}
                    # daifactor[state.toInt()] = math.exp(factor.psi(ignore_evidence, label))
                    daifactor[state.toInt()] = factor.psi(ignore_evidence, label)

                    state.next()

                dai_factors.append(daifactor)

        dai_network = dai.FactorGraph(dai_factors)
        dai_network.WriteToFile('effusion.fg')

        logging.debug('Done making opengm')

        return dai_network, myvar_to_daivar_map

    @effusion.log_time()
    def infer(self):
        maxiter = 10000
        tol = 1e-9
        verbosity = 1
        logdomain = 1

        # name_opts = 'JTREE[inference=SUMPROD,heuristic=MINFILL,updates=HUGIN,verbose=3]'
        maxiter = 100000000
        restart = maxiter / 10
        burnin = restart * .2

        # name_opts = 'GIBBS[maxtime=%d,maxiter=%d,restart=%d,burnin=%d,verbose=3]' % (agca.config.max_time_for_inference, maxiter, restart, burnin)

        name_opts = 'BP[inference=SUMPROD,updates=SEQMAX,logdomain=0,tol=1e-9,maxiter=10000,damping=.1,verbose=3]'
        # name_opts = 'TRWBP[inference=SUMPROD,updates=SEQRND,logdomain=1,tol=1e-9,maxiter=10000,damping=.1,nrtrees=0,verbose=3]'

        # name_opts = 'LC[verbose=3,maxtime=60,cavity=UNIFORM,updates=SEQFIX,maxiter=10000,tol=1e-9,cavaiopts=[],cavainame=NONE]'
        # name_opts = 'LC[verbose=3,maxtime=1800,cavity=PAIR,reinit=1,updates=SEQRND,maxtime=300,maxiter=10000,cavainame=BP,cavaiopts=[updates=SEQMAX,tol=1e-9,maxiter=10000,logdomain=1],tol=1e-9]'
        # name_opts = 'LC[verbose=3,maxtime=60,cavity=PAIR2,reinit=1,updates=SEQFIX,maxiter=10000,cavainame=BP,cavaiopts=[updates=SEQMAX,tol=1e-9,maxiter=10000,logdomain=0],tol=1e-9]'
        # name_opts = 'LC[verbose=3,maxtime=60,cavity=FULL,reinit=1,updates=SEQFIX,maxiter=10000,cavainame=BP,cavaiopts=[updates=SEQMAX,tol=1e-9,maxiter=10000,logdomain=0],tol=1e-9]'
        # name_opts = 'GLC[verbose=3,rgntype=SINGLE,cavity=FULL,updates=SEQRND,maxiter=10000,tol=1e-9,cavainame=BP,cavaiopts=[updates=SEQMAX,tol=1e-9,maxiter=10000,logdomain=1],inainame=EXACT,inaiopts=[],tol=1e-9]'
        # name_opts = 'GLC[verbose=3,rgntype=SINGLE,cavity=PAIR,updates=SEQRND,maxiter=10000,tol=1e-9,cavainame=BP,cavaiopts=[updates=SEQMAX,tol=1e-9,maxiter=10000,logdomain=1],inainame=EXACT,inaiopts=[],tol=1e-9]'
        # name_opts = 'GLC[verbose=3,rgntype=OVLOOP,loopdepth=3,cavity=PAIR,updates=SEQRND,maxiter=10000,tol=1e-9,cavainame=BP,cavaiopts=[updates=SEQMAX,tol=1e-9,maxiter=10000,logdomain=1],inainame=JTREE,inaiopts=[inference=SUMPROD,updates=HUGIN],tol=1e-9]'
        # name_opts = 'GLC[verbose=3,rgntype=OVLOOP,loopdepth=4,cavity=PAIR,updates=SEQRND,maxiter=10000,tol=1e-9,cavainame=BP,cavaiopts=[updates=SEQMAX,tol=1e-9,maxiter=10000,logdomain=1],inainame=JTREE,inaiopts=[inference=SUMPROD,updates=HUGIN],tol=1e-9]'

        # name_opts = 'TREEEP[type=ORG,tol=1e-9,maxiter=10000,verbose=2]'


        # name_opts = 'HAK[tol=1e-09,maxiter=10000,maxtime=inf,verbose=3,doubleloop=0,clusters=MIN,init=UNIFORM,loopdepth=0,damping=0]'
        # name_opts = 'HAK[tol=1e-09,maxiter=10000,maxtime=inf,verbose=3,doubleloop=0,clusters=BETHE,init=UNIFORM,loopdepth=0,damping=0]'
        # name_opts = 'HAK[tol=1e-09,maxiter=10000,maxtime=inf,verbose=3,doubleloop=0,clusters=LOOP,init=UNIFORM,loopdepth=3,damping=0]'
        # name_opts = 'HAK[doubleloop=1,clusters=MIN,init=UNIFORM,tol=1e-9,maxiter=10000,verbose=3]'
        # name_opts = 'HAK[doubleloop=1,clusters=BETHE,init=UNIFORM,tol=1e-9,maxiter=10000,verbose=3]'
        # name_opts = 'HAK[doubleloop=1,clusters=DELTA,init=UNIFORM,tol=1e-9,maxiter=10000,maxtime=300,verbose=3]'
        # name_opts = 'HAK[doubleloop=1,clusters=LOOP,init=UNIFORM,loopdepth=3,tol=1e-9,maxiter=10000,maxtime=1800,verbose=3]'
        # name_opts = 'CBP[verbose=3,max_levels=12,updates=SEQMAX,tol=1e-9,rec_tol=1e-9,maxiter=500,choose=CHOOSE_RANDOM,recursion=REC_FIXED,clamp=CLAMP_VAR,min_max_adj=1.0e-9,bbp_cfn=CFN_FACTOR_ENT,rand_seed=0,bbp_props=[tol=1.0e-9,maxiter=10000,damping=0,updates=SEQ_BP_REV],clamp_outfile=]'

        self.solver = dai.newInfAlgFromString(name_opts, self.gm)

        self.solver.init()
        self.solver.run()

    def get_marginal(self, variable, xi):

        vi = self.protein_to_opengm_index_map[variable]
        # li = list(variable.scope()).index(xi)
        # marginal = self.solver.belief(vi)
        marginal = self.solver.belief(dai.VarSet(vi))
        # with this int, 0 is False, 1 is True. 2 other places in code here
        marginal_val = marginal[int(xi)]

        return marginal_val


class UaiInferenceEngine(InferenceEngine):
    @effusion.log_time(name='BayesNetInferenceEngine.__init__')
    def __init__(self, factor_tree):
        super(UaiInferenceEngine, self).__init__(factor_tree)

        self.hard_coded_uai_path = None
        self.hard_coded_mar_path = None
        # self.hard_coded_uai_path = '/mnt/fast/effusion/rundir/top_down_copy3.uai'
        # self.hard_coded_mar_path = None


        self.write_uai()

    def write_uai(self):
        # ref: http://www.cs.huji.ac.il/project/UAI10/fileFormat.php

        ignore_evidence = False

        self.myvar_to_uaivar_map = {}
        self.myvar_array_in_uai_order = [None] * len(self.factor_tree.variables)
        for vi, v in enumerate(self.factor_tree.variables):
            self.myvar_to_uaivar_map[v] = vi
            self.myvar_array_in_uai_order[vi] = v

        logging.debug('Making uai...')

        if self.hard_coded_uai_path:
            self.uai_file_path = self.hard_coded_uai_path
            return

        else:
            with tempfile.NamedTemporaryFile(dir=effusion.args.tmp_dir, delete=False) as of:
                self.uai_file_path = of.name
                logging.info('Writing uai to %s' % self.uai_file_path)

                of.write('MARKOV\n')
                of.write('%d\n' % len(self.factor_tree.variables))
                of.write('%s\n' % ' '.join(str(len(myvar.scope())) for myvar in self.myvar_array_in_uai_order))
                of.write('%d\n' % len(self.factor_tree.factors))

                for factor in self.factor_tree.factors:
                    if isinstance(factor, effusion.AnnotationFactor):
                        factor_variables = factor.variables
                        of.write('%d ' % len(factor_variables))

                        # remember, the scope is ordered (True, False), so scope[0] = True and scope[1] = False

                        of.write('%s\n' % ' '.join(str(self.myvar_to_uaivar_map[myvar]) for myvar in factor_variables))

                of.write('\n')

                for factor in self.factor_tree.factors:
                    if isinstance(factor, effusion.AnnotationFactor):

                        states = factor.scope(lsb='last')
                        of.write('%d\n ' % len(states))

                        # sum_psi = 0
                        for state in states:
                            val = factor.psi(ignore_evidence, state)
                            of.write('%f ' % val)
                            # sum_psi += val
                        # assert sum_psi > 0

                        of.write('\n\n')

        logging.debug('Done making uai')

    @effusion.log_time()
    def infer(self):

        with tempfile.NamedTemporaryFile(dir=effusion.args.tmp_dir, delete=False) as evidence_file:
            evidence_file.write('0\n')
            evidence_file_path = evidence_file.name

        if self.hard_coded_mar_path:
            mar_path = self.hard_coded_mar_path

        else:
            # cmd = [effusion.args.uai_solver, '-i', '14', '-n', '20', self.uai_file_path, evidence_file_path, str(effusion.args.max_time), 'MAR']
            cmd = str.format(effusion.args.uai_solver, uai_file_path=self.uai_file_path,
                             evid_file_path=evidence_file_path, max_seconds=effusion.args.max_time)
            logging.info(cmd)
            # cmd = [effusion.args.uai_solver, '-i', '10', '-n', '10', self.uai_file_path, evidence_file_path, '60', 'MAR']
            # cmd = [effusion.args.uai_solver, '-i', '10', '-n', '10', self.uai_file_path, evidence_file_path, '60', 'MAR']

            uai_filename = os.path.basename(self.uai_file_path)  # e.g., tmpFpsUfu
            mar_path = uai_filename + '.MAR'

            fix_uai14_single_variable_bug = False

            # not a check call. let it get killed by ulimit. check retcode if you want
            with subprocess.Popen(cmd,
                                  shell=True,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.STDOUT,
                                  preexec_fn=setlimits,
                                  ) as proc:
                for line in proc.stdout:
                    line = line.strip()

                    if "Assertion `size()>0' failed." in line:
                        # todo assert only variable / factor
                        fix_uai14_single_variable_bug = True

                    logging.info(line)

            logging.info('Resource errors are normal for some anytime algorithms')

            if fix_uai14_single_variable_bug:
                with open(mar_path, 'w') as uai_file:
                    uai_file.write('MAR\n1 2 1 0')

        time.sleep(4)  # give some time for flushing / filesystem to catch up maybe?
        with open(mar_path) as uai_file:  # ijgp-samplesearch puts it in the cwd
            line = uai_file.readline()
            assert line == 'MAR\n'

            for line in uai_file:
                line = line.strip()
                if line == '':
                    continue
                last_solution_line = line

            # format of line = num_vars, x0_card, x0_0, x0_1, x1_card, x1_0, x1_1, ...
            last_solution_line = last_solution_line.strip()
            split_line = last_solution_line.split()
            num_vars, marginals_list = int(split_line[0]), split_line[1:]
            del line, last_solution_line, split_line
            assert num_vars == len(self.factor_tree.variables)

            self.marginals = {}
            current_token_index = 0
            current_uai_var = 0
            while current_uai_var < len(self.myvar_array_in_uai_order):
                current_myvar = self.myvar_array_in_uai_order[current_uai_var]

                token = marginals_list[current_token_index]
                current_token_index += 1
                cardinality = int(token)
                current_scope = current_myvar.scope()
                assert len(current_scope) == cardinality
                marginal = {}
                for state in current_scope:
                    token = marginals_list[current_token_index]
                    current_token_index += 1
                    prob = float(token)
                    marginal[state] = prob
                self.marginals[current_myvar] = marginal

                current_uai_var += 1
            pass

    def get_marginal(self, variable, xi):
        marginal_val = self.marginals[variable][xi]
        return marginal_val


def setlimits():
    resource.setrlimit(resource.RLIMIT_CPU, (
        effusion.args.max_time * 1.125, effusion.args.max_time * 1.125))  # seconds. give a 40 minute run an extra 5 minutes
    resource.setrlimit(resource.RLIMIT_AS, (int(8e9), int(8e9)))  # 8 GB, expressed in bytes
    resource.setrlimit(resource.RLIMIT_CORE, (0, 0))
