from __future__ import division

import cStringIO
import logging
import multiprocessing
import random
import sys
import traceback
from collections import defaultdict
from itertools import chain

import os
import re
from Bio.Application import ApplicationError
from Bio.Blast import NCBIXML
from Bio.Blast.Applications import NcbiblastpCommandline

import effusion


def postprocess(dataset, uniprot, raw_preds, subset_type, do_filter_purified, do_break_ties, prop_down_type,
                max_predictions):
    # relevant_go_ids = set(dataset.gene_ontology.vs['name'])
    # topological_sort_indices = dataset.gene_ontology.topological_sorting()
    # topological_sort_vertices = dataset.gene_ontology.vs[topological_sort_indices]


    # todo(jyunes) why are the predictions not for the query again? answer: line 260.
    for raw_pred in raw_preds:
        raw_pred.uniprot = uniprot

    preds = simplify_annotations(raw_preds)  # more of an optimization
    preds = effusion.QuickPred.get_all_implied_preds(preds)

    # todo options. don't allow multiple. enum variable prop_down_type
    assert prop_down_type is None
    # if prop_down_type == 'prop_down':
    #     preds = propagate_downward(dataset, preds, uniprot, relevant_go_ids)
    #
    # elif prop_down_type == 'smart_prop_down':
    #     preds = smart_propagate_downward(dataset, preds, uniprot, relevant_go_ids, topological_sort_vertices=topological_sort_vertices)
    #
    # elif prop_down_type == 'prop_down_leaves':
    #     propagate_down_leaves(dataset, preds, uniprot)

    preds = simplify_annotations(preds)  # propagating up probably gives duplicates
    # todo could simplify in functions that might generate duplicates
    # todo(jyunes) this is really just set intersection with the evaluation terms.
    # todo(jyunes) but don't do it before propagating downward. order matters! (make more robust)
    if subset_type == 'catalytic':
        preds = just_catalytic_annotations(preds)

    elif subset_type == 'molecular_function':
        preds = just_ontology(preds, effusion.GoAspectEnum.molecular_function)

    elif subset_type == 'biological_process':
        preds = just_ontology(preds, effusion.GoAspectEnum.biological_process)

    else:
        # evaluating CC?
        raise NotImplementedError

    if do_filter_purified:
        preds = filter_purified(preds)

    if do_break_ties:
        preds = break_ties(preds)

    # todo go_id to give deterministic results. probably should be in break ties!!
    preds = sorted(preds, key=lambda _: [_.prob, _.go_id], reverse=True)
    preds = preds[0:max_predictions]

    return preds


# def merge_protein_preds(postprocessed_preds_dict, rank_norm, round_probs):
#     merged_ordered_preds = []
#
#     for uniprot, postprocessed_preds in postprocessed_preds_dict:
#         merged_ordered_preds.extend(preds)
#
#     merged_ordered_preds.sort(key=lambda annotation: annotation.prob, reverse=True)
#
#     # in cafa, only 2 decimals allowed. therefore, better to space out probs, rather than have lots near 1 and 0, and sparse in the middle
#     if rank_norm:
#         for i, a in enumerate(merged_ordered_preds):
#             a.prob = 1 - (i / len(merged_ordered_preds))
#
#     if round_probs:
#         for a in merged_ordered_preds:
#             a.prob = round(a.prob, 2)
#
#     return merged_ordered_preds



def simplify_annotations(annotations):
    best_annotations = {}

    for annotation in annotations:
        best_annotation = best_annotations.get(annotation.go_id)
        if best_annotation is None or annotation.prob > best_annotation.prob:
            best_annotations[annotation.go_id] = annotation

    return best_annotations.values()


def just_catalytic_annotations(annotations):
    retval = filter(lambda a: effusion.dataset.gene_ontology.term4id(a.go_id).catalytic, annotations)
    return retval


def just_ontology(annotations, aspect):
    retval = filter(lambda a: effusion.dataset.gene_ontology.term4id(a.go_id).aspect == aspect, annotations)
    return retval


def filter_purified(annotations):
    retval = []
    for a in annotations:
        # filter out stuff below the background. which includes some of the stuff added by negative evidence
        term_prior = 2 ** (-1 * effusion.dataset.gene_ontology.term4id(a.go_id).marginal_ic)
        if a.prob < term_prior:
            continue
        retval.append(a)

    return retval


def break_ties(preds):
    # the penalty simply allows higher level terms to be predicted before lower level terms
    # this helps blast a lot, because every direct and implied term for a hit would otherwise be evaluated at once

    # p(term, depth) = p(term) * p(depth). p(depth) = (1 - .0001 * go_term.depth)
    # todo maybe even calculate p(depth) from data rather than just making an arbitrary penalty
    # todo at least change the distribution!?

    # todo
    # doesn't break ties for siblings...

    retval = set()

    for pred in preds:
        go_term = effusion.dataset.gene_ontology.term4id(pred.go_id)
        pred_prob = pred.prob
        if pred_prob == 0:
            pred_prob = 1e-8  # for methods that have a lot of 0's (like top-down), still break ties
        depth_prob = 1 - .0001 * go_term.depth
        new_pred = effusion.QuickPred(pred.uniprot, pred.go_id, pred_prob * depth_prob)
        retval.add(new_pred)

    return retval


# todo(jyunes) need the whole dataset with the whole GO, but would be nice to have a catalytic subgo?
# in some places, this could be a problem if we are saying that more general terms can be irrelevent (mf root), but more specific terms are relevant (enzyme)
def propagate_downward(dataset, preds, uniprot, relevant_go_ids):
    preds = simplify_annotations(preds)  # also makes a copy
    go_id_probs = {a.go_id: a.prob for a in preds}

    topological_sort_indices = dataset.gene_ontology.topological_sorting()

    topological_sort_vertices = dataset.gene_ontology.vs[topological_sort_indices]

    for go_vertex in topological_sort_vertices:
        go_id = go_vertex['name']
        if go_id not in relevant_go_ids:  # must include children *and* parents of key (e.g., catalytic) terms!
            continue

        if go_id in go_id_probs:  # we already have an explicit prediction for this term
            continue

        else:
            parents_prob = 1

            for parent_vertex in go_vertex.predecessors():
                parent_term = parent_vertex['term']
                parents_prob *= go_id_probs.get(parent_term.id, 0)
            if parents_prob == 0:
                continue

            prob = parents_prob * dataset.go_hierarchy_params[go_id]
            new_pred = effusion.Annotation(dataset.gene_ontology, uniprot, go_id, prob=prob)
            preds.append(new_pred)
            go_id_probs[go_id] = prob

    return preds


# todo(jyunes) need the whole dataset with the whole GO, but would be nice to have a catalytic subgo?
# in some places, this could be a problem if we are saying that more general terms can be irrelevent (mf root), but more specific terms are relevant (enzyme),
#  because we might short circuit after seeing irrelevent parent term
def smart_propagate_downward(dataset, preds, uniprot, relevant_go_ids, topological_sort_vertices=None):
    preds = simplify_annotations(preds)  # also makes a copy
    positive_go_ids = {a.go_id for a in preds}
    go_id_probs = {a.go_id: a.prob for a in preds}

    if topological_sort_vertices is None:
        topological_sort_indices = dataset.gene_ontology.topological_sorting()
        topological_sort_vertices = dataset.gene_ontology.vs[topological_sort_indices]

    for go_vertex in topological_sort_vertices:
        go_id = go_vertex['name']
        if go_id not in relevant_go_ids:  # must include children *and* parents of key (e.g., catalytic) terms!
            continue

        if go_id in go_id_probs:  # we already have an explicit prediction for this term
            continue

        # optimization. don't try vertices that are impossible...
        any_parent_impossible = False
        for parent_go_id in dataset.gene_ontology.get_parent_go_ids(go_vertex):
            # todo could be more optimized...
            parents_prob = go_id_probs.get(parent_go_id, 0)
            if parents_prob <= 0:
                any_parent_impossible = True
                break
        if any_parent_impossible:
            continue

        # moonlight smart: if function is unlikely to moonlight, don't prop down
        # just try a simple version here: don't predict if there are any predictions for siblings
        any_sibling_is_positive = len(dataset.gene_ontology.get_sibling_ids(go_vertex) & positive_go_ids) > 0
        if any_sibling_is_positive:
            continue
        #### end moonlight smart

        parents_prob = 1

        for parent_go_id in dataset.gene_ontology.get_parent_go_ids(go_vertex):
            parents_prob *= go_id_probs.get(parent_go_id, 0)
        if parents_prob == 0:
            continue

        prob = parents_prob * dataset.go_hierarchy_params[go_id]
        new_pred = effusion.QuickPred(uniprot, go_id, prob)
        preds.append(new_pred)
        go_id_probs[go_id] = prob

    return preds


def propagate_down_leaves(dataset, preds, uniprot):
    preds = simplify_annotations(preds)  # also makes a copy
    positive_go_ids = {a.go_id for a in preds}
    go_id_probs = {a.go_id: a.prob for a in preds}

    for positive_go_id in positive_go_ids:
        go_vertex = dataset.gene_ontology.vs.find(positive_go_id)
        child_vs = go_vertex.successors()
        child_ids = [v['name'] for v in child_vs]

        # this hypothesis only predicts down one level from leaves
        # if the method made a prediction for any child, don't infer.
        if any((c_id in positive_go_ids) for c_id in child_ids):
            continue

        for child_id in child_ids:
            prob = go_id_probs[positive_go_id] * dataset.go_hierarchy_params[child_id]
            new_pred = effusion.Annotation(dataset.gene_ontology, uniprot, child_id, prob=prob)
            preds.append(new_pred)
            go_id_probs[child_id] = prob

    return preds


class PFPMethod(object):
    def __init__(self, dataset, *args, **kwargs):
        super(PFPMethod, self).__init__()
        self.dataset = dataset

        self.query_infos = {}

    def predict_batch(self, uniprots):
        raise NotImplementedError

    def predict_one(self, uniprot):
        return self.predict_batch([uniprot])[uniprot]


class NetworkMethod(PFPMethod):
    def __init__(self, reduce_to_model_args, model_bp,
                 stop_on_error=False, build_network=False, *args, **kwargs):
        super(NetworkMethod, self).__init__(*args, **kwargs)

        self.reduce_to_model_args = reduce_to_model_args
        self.model_bp = model_bp
        self.stop_on_error = stop_on_error
        self.build_network = build_network

    def predict_batch(self, uniprots):
        retval = {}

        for uniprot in uniprots:
            retval[uniprot] = self.predict_one(uniprot)

        return retval

    def predict_one(self, uniprot):
        self.current_query_info = {}
        self.current_query_info['_uniprot'] = uniprot

        logging.info('Predicting for %s' % uniprot)
        # agca.settings.alchemy_session.close()  # should cleanup session

        preds = None

        try:
            protein_network, training_annotations = self.get_network(uniprot)
            pgm, reduced_network = self.make_pgm(protein_network, training_annotations)
            preds = self.infer(pgm, uniprot, reduced_network)

        except None:  # Exception:
            stack_trace = traceback.format_exc()
            logging.error('Exception predicting %s: %s' % (uniprot, stack_trace))
            if self.stop_on_error:
                raise

        self.query_infos[uniprot] = self.current_query_info
        return preds

    def get_network(self, uniprot):
        # todo(jyunes): store the way the network was built so you know which evidence to collect for each node (don't use uniref90 if network was built from uniref100)
        # in other words, get evidence for the nodes in the network. and the nodes are really UniRef90_AAD3_YEAST...
        net_filename = os.path.join(effusion.dataset.networks_dir, uniprot + '.ssfa.json.gz')

        try:
            protein_network = effusion.protein_network.ProteinNetwork.from_json_gz(net_filename)

        except IOError:
            if not self.build_network:
                raise

            protein_network = effusion.ProteinNetwork.build_direct_net(
                uniprot,
                net_cache=net_filename,
                do_string=True)

            if not effusion.args.no_write_visual_files:
                # todo(jyunes) for display only (except for pruning to supervised method). could look even more like 'add interesting attribute'
                # change to evaluation role. we're putting the answers in the network
                old_phase, old_role = effusion.dataset.phase, effusion.dataset.role
                effusion.dataset.set_phase_and_role(old_phase, 'evaluation')
                annotations_dict = effusion.Annotation.get_annotations_for_uniprots(protein_network.vs['name'],
                                                                                    include_implied=False)
                protein_network.set_evidence(annotations_dict)
                protein_network.export_cyjs(uniprot + '.cyjs')
                effusion.dataset.set_phase_and_role(old_phase, old_role)

        # todo xxx needs a configuration!! otherwise won't work for string+bp method obviously
        protein_network.vs.select(
            round_added=2).delete()  # why is this better than reduce_to_model(use_stringdb=False) ?

        self.current_query_info['network_size'] = len(protein_network.vs)

        training_annotations = effusion.Annotation.get_annotations_for_uniprots(
            uniprots=protein_network.vs['name'],
            include_negative=True,
            include_direct=True,
            include_implied=True,
            include_moonlight=False)

        # for visualization
        protein_network.set_evidence(training_annotations)

        if uniprot == "Q4ZIL6":  # for the main example in the paper
            go_terms_to_color = ['GO:0008395', 'GO:0008404']
            for v in protein_network.vs:
                v['which_func'] = 0
                for color, go_term_to_color in enumerate(go_terms_to_color, start=1):
                    if go_term_to_color in v['goas']:
                        v['which_func'] = color

            example_proteins = ['Q4ZIL6', 'O54750', 'P13108']
            for example_protein in example_proteins:
                protein_network.vs.find(name=example_protein)['example'] = True

        if 'we are' == 'exporting raw networks and quitting':
            # make SSN
            cyjs_name_components = filter(None, [uniprot, effusion.args.experiment_name, 'diamond', 'cyjs'])
            cyjs_name = '.'.join(cyjs_name_components)
            protein_network.export_cyjs(cyjs_name)

            # make blast network
            query = protein_network.vs.find(query=True)
            es_not_from_query = protein_network.es.select(_within=[v.index for v in protein_network.vs if v != query])
            es_not_from_query.delete()
            for v in protein_network.vs:
                if query == v:
                    continue
                if len(protein_network.es.select(_within=[query.index, v.index])) == 0:
                    protein_network.add_edge(query, v, **{'MST': False, 'type': 'special', 'score': 0, 'evalue': 1,
                                                          'round_added': 0})
            cyjs_name_components = filter(None, [uniprot, effusion.args.experiment_name, 'blast', 'cyjs'])
            cyjs_name = '.'.join(cyjs_name_components)
            protein_network.export_cyjs(cyjs_name)
            sys.exit(1)

        protein_network.reduce_to_model(annotations_dict=training_annotations, **self.reduce_to_model_args)
        # protein_network.set_evidence(training_annotations)
        # protein_network.export_cyjs('my-export.cyjs')
        # sys.exit(1)

        return protein_network, training_annotations

    def make_pgm(self, protein_network, training_annotations):
        func_name = 'NetworkMethod.make_pgm'

        uniprot = protein_network.vs.find(query=True)['name']

        if len(list(chain.from_iterable(training_annotations.values()))) == 0:
            logging.warning('No training annotations in network')

        pgm, reduced_network = effusion.BayesNet.learn_bayes_net(
            dataset=self.dataset,
            protein_network=protein_network,
            annotations_dict=training_annotations,
            query_info=self.current_query_info,
            model_bp=self.model_bp
        )

        if not effusion.args.no_write_visual_files:
            go_name_components = filter(None, [effusion.args.experiment_name, uniprot, 'template.go.eps'])
            go_name = '.'.join(go_name_components)
            pgm.gene_ontology_subgraph.plot(go_name)

        self.current_query_info['network_reduced_size'] = len(reduced_network.vs)
        self.candidate_ontology = pgm.gene_ontology_subgraph

        return pgm, reduced_network

    # todo(jyunes) all this reduced network stuff
    def infer(self, pgm, uniprot, reduced_network):
        func_name = 'NetworkMethod.infer'

        preds = []

        if effusion.args.use_libdai:
            inf_engine = effusion.BayesNetInferenceEngine(pgm)
        else:
            inf_engine = effusion.UaiInferenceEngine(pgm)
        inf_engine.infer()

        if not effusion.args.no_write_visual_files:
            effusion.cyjs_utils.add_marginals_as_attribute(inf_engine, reduced_network)
            cyjs_name_components = filter(None, [uniprot, effusion.args.experiment_name, 'reduced', 'cyjs'])
            cyjs_name = '.'.join(cyjs_name_components)
            reduced_network.export_cyjs(cyjs_name)

            # plot go of each query
            for v in reduced_network.vs.select(query=True):
                print 'found query:', v['name']
                go_name_components = filter(None, [effusion.args.experiment_name, v['name'], 'go.eps'])
                go_name = '.'.join(go_name_components)
                pgm.gene_ontology_subgraph.plot(go_name, protein_node=v, vertex_color=(0.0, 0.0, 0.0))

        go_term_marginals = {}  # a map from go_id -> prob
        matching_nodes_in_factor_graph = pgm.vs.select(protein=uniprot, type='variable',
                                                       variable_type=effusion.config.display_p_annotation_or_function)
        for matching_node_in_factor_graph in matching_nodes_in_factor_graph:
            go_term_marginals[matching_node_in_factor_graph['go_id']] = inf_engine.get_marginal(
                matching_node_in_factor_graph['variable'], True)

        for go_id, prob in go_term_marginals.items():
            preds.append(
                effusion.QuickPred(uniprot=uniprot, go_id=go_id, prob=prob))

        # todo these are just globals so I can peek at them later for cafa
        self.last_net = reduced_network

        return preds


# todo(jyunes) Maybe it should predict negatives too?
# todo(jyunes) Maybe it should add all labels and Label should filter?
class NullMethod(NetworkMethod):
    def __init__(self, *args, **kwargs):
        # todo(jyunes) real nice
        super(NullMethod, self).__init__(None, None, None, None, None, *args, **kwargs)

    def predict_batch(self, uniprots):
        return {uniprot: self.predict_one(uniprot) for uniprot in uniprots}

    def predict_one(self, uniprot):
        '''
        The naive prediction, viewed as an empty network without even the query.
        The version that predicts the background distribution of terms is an application of the propagate_down param
        :param uniprot:
        :param dataset:
        :return: A certain prediction of root terms of MF and BP.
        '''

        # todo(jyunes) Probabilities for root terms, rather than default 1
        predictions = []
        predictions.append(effusion.Annotation(
            gene_ontology=self.dataset.gene_ontology,
            go_id=self.dataset.gene_ontology.root_mf_id,
            uniprot=uniprot))
        predictions.append(effusion.Annotation(
            gene_ontology=self.dataset.gene_ontology,
            go_id=self.dataset.gene_ontology.root_bp_id,
            uniprot=uniprot))

        return predictions


class QueryMethod(NetworkMethod):
    def __init__(self, reduce_to_model_args, model_bp, *args, **kwargs):
        super(QueryMethod, self).__init__(reduce_to_model_args, model_bp,
                                          *args, **kwargs)
        raise NotImplementedError


# todo(jyunes) just consider
class ExclusiveBranchesMethod(PFPMethod):
    def __init__(self, dataset, *args, **kwargs):
        super(ExclusiveBranchesMethod, self).__init__(dataset, *args, **kwargs)
        raise NotImplementedError


# todo(jyunes) just consider
# would have to make a classifier for each candidate go term.
# that's kind of a medium amount of work....
class DontUseGOMethod(PFPMethod):
    def __init__(self, dataset, *args, **kwargs):
        super(DontUseGOMethod, self).__init__(dataset, *args, **kwargs)
        raise NotImplementedError


class BlastMethod(PFPMethod):
    def __init__(self, dataset, *args, **kwargs):
        self.evalue = effusion.args.evalue
        self.max_target_seqs = effusion.args.max_target_seqs
        self.min_query_cover = effusion.args.min_query_cover

        super(BlastMethod, self).__init__(dataset=dataset, *args, **kwargs)

    def predict_batch(self, uniprots):
        retval = defaultdict(list)

        # for normalization of scores, so that they can be used to augment other predictions
        # calculated from validation dataset
        # todo: normalizing by bitscore of self hit would probably be better
        max_bitscore_per_residue = 2.13191
        min_bitscore_per_residue = 0.139002

        query_fasta_file = effusion.write_fasta_for_accessions(effusion.args.uniprot_db, uniprots)
        if os.stat(query_fasta_file).st_size == 0:
            logging.warning('Nothing to BLAST, given %d query accessions' % len(uniprots))
            return {}

        # todo convert min_bit_score_per_residue to evalues OR switch to diamond
        # for now, just use a more lenient evalue and filter to min_bit_score_per_residue

        blast_cline = NcbiblastpCommandline(cmd=effusion.args.blastp_bin,
                                            task="blastp",  # could also use blastp-fast, but was missed some targets
                                            query=query_fasta_file,
                                            db=effusion.dataset.goa_blast_db,
                                            max_hsps=1,
                                            num_threads=multiprocessing.cpu_count(),
                                            max_target_seqs=self.max_target_seqs,
                                            outfmt=5,
                                            evalue=self.evalue,
                                            qcov_hsp_perc=self.min_query_cover,
                                            parse_deflines=True)  # returns accession rather than "Query 1"
        print blast_cline
        try:
            stdout, stderr = blast_cline()
        except ApplicationError as e:
            logging.error("Couldn't run blast:\n+"
                          "cmd: %s\n" % str(e.cmd) +
                          "retcode: %d\n" % e.returncode +
                          "msg: %s\n" % str(e.message) +
                          "stderr: %s\n" % str(e.stderr))
            return {}

        stdout_sio = cStringIO.StringIO(stdout)
        blast_records = NCBIXML.parse(stdout_sio)

        fasta_header_pattern = re.compile(r'\w+\|(\w+)\|\w+')

        for blast_record in blast_records:
            query_uniprot = blast_record.query_id
            # suddenly not needed?
            # match = fasta_header_pattern.match(query_uniprot)
            # query_uniprot = match.group(1)
            predictions = []

            # todo(jyunes) We probably don't need to worry too much about propagating the same annotation multiple times
            # Evaluation framework should handle that
            map_for_go_id_to_prob_for_this_uniprot = {}  # save only the best prob for each go_id (per protein)
            for alignment in blast_record.alignments:
                target_uniprot = alignment.accession

                bit_score_per_residue = alignment.hsps[0].bits / max(blast_record.query_length,
                                                                     len(alignment.hsps[0].sbjct))
                if bit_score_per_residue < effusion.args.min_bit_score_per_residue:
                    continue
                prob = (bit_score_per_residue - min_bitscore_per_residue) / (
                    max_bitscore_per_residue - min_bitscore_per_residue)

                # negative annoations aren't a positive prediction
                # todo research what about evaluating negative predictions?
                annotations = effusion.Annotation.get_goas_for_uniprot(uniprot=target_uniprot,
                                                                       include_negative=False,
                                                                       include_direct=True,
                                                                       include_implied=True,
                                                                       include_moonlight=False)

                for annotation in annotations:
                    # todo(jyunes): does the prob depend on the quality of the original evidence?!
                    if map_for_go_id_to_prob_for_this_uniprot.get(annotation.go_id, None) is None or prob > \
                            map_for_go_id_to_prob_for_this_uniprot[annotation.go_id]:
                        predictions.append(
                            effusion.QuickPred(query_uniprot, annotation.go_id, prob=prob))
                        map_for_go_id_to_prob_for_this_uniprot[
                            annotation.go_id] = prob  # save only the best prob for each go_id (per protein)
            retval[query_uniprot] = predictions

        return retval


class RandomBlastMethod(PFPMethod):
    def __init__(self, dataset, *args, **kwargs):
        self.evalue = effusion.args.evalue
        self.max_target_seqs = effusion.args.max_target_seqs
        self.min_query_cover = effusion.args.min_query_cover

        super(RandomBlastMethod, self).__init__(dataset=dataset, *args, **kwargs)

    def predict_batch(self, uniprots):
        retval = defaultdict(list)

        query_fasta_file = effusion.write_fasta_for_accessions(effusion.args.uniprot_db, uniprots)
        if os.stat(query_fasta_file).st_size == 0:
            logging.warning('Nothing to BLAST, given %d query accessions' % len(uniprots))
            return {}

        blast_cline = NcbiblastpCommandline(cmd=effusion.args.blastp_bin,
                                            task="blastp",  # could also use blastp-fast, but was missed some targets
                                            query=query_fasta_file,
                                            db=effusion.dataset.goa_blast_db,
                                            max_hsps=1,
                                            num_threads=multiprocessing.cpu_count(),
                                            max_target_seqs=self.max_target_seqs,
                                            outfmt=5,
                                            evalue=self.evalue,
                                            qcov_hsp_perc=self.min_query_cover,
                                            parse_deflines=True)  # returns accession rather than "Query 1"
        print blast_cline
        try:
            stdout, stderr = blast_cline()
        except ApplicationError as e:
            logging.error("Couldn't run blast:\n+"
                          "cmd: %s\n" % str(e.cmd) +
                          "retcode: %d\n" % e.returncode +
                          "msg: %s\n" % str(e.message) +
                          "stderr: %s\n" % str(e.stderr))
            return {}

        stdout_sio = cStringIO.StringIO(stdout)
        blast_records = NCBIXML.parse(stdout_sio)

        fasta_header_pattern = re.compile(r'\w+\|(\w+)\|\w+')

        for blast_record in blast_records:
            query_uniprot = blast_record.query_id
            # suddenly not needed?
            # match = fasta_header_pattern.match(query_uniprot)
            # query_uniprot = match.group(1)
            predictions = []

            # todo(jyunes) We probably don't need to worry too much about propagating the same annotation multiple times
            # Evaluation framework should handle that
            map_for_go_id_to_prob_for_this_uniprot = {}  # save only the best prob for each go_id (per protein)
            for alignment in blast_record.alignments:
                target_uniprot = alignment.accession

                bit_score_per_residue = alignment.hsps[0].bits / max(blast_record.query_length,
                                                                     len(alignment.hsps[0].sbjct))
                if bit_score_per_residue < effusion.args.min_bit_score_per_residue:
                    continue

                prob = random.random()

                # negative annoations aren't a positive prediction
                # todo research what about evaluating negative predictions?
                annotations = effusion.Annotation.get_goas_for_uniprot(uniprot=target_uniprot,
                                                                       include_negative=False,
                                                                       include_direct=True,
                                                                       include_implied=True,
                                                                       include_moonlight=False)

                for annotation in annotations:
                    # todo(jyunes): does the prob depend on the quality of the original evidence?!
                    if map_for_go_id_to_prob_for_this_uniprot.get(annotation.go_id, None) is None or prob > \
                            map_for_go_id_to_prob_for_this_uniprot[annotation.go_id]:
                        predictions.append(
                            effusion.QuickPred(query_uniprot, annotation.go_id, prob=prob))
                        map_for_go_id_to_prob_for_this_uniprot[
                            annotation.go_id] = prob  # save only the best prob for each go_id (per protein)
            retval[query_uniprot] = predictions

        return retval


class BlastLikeMethod(NetworkMethod):
    def __init__(self, dataset, *args, **kwargs):
        reduce_to_model_args = {'use_network': False, 'use_unlabeled': False, 'use_stringdb': False}
        model_bp = False

        super(BlastLikeMethod, self).__init__(reduce_to_model_args, model_bp,
                                              dataset=dataset, *args, **kwargs)


class SupervisedNetworkMethod(NetworkMethod):
    def __init__(self, dataset, *args, **kwargs):
        reduce_to_model_args = {'use_network': True, 'use_unlabeled': False, 'use_stringdb': False}
        model_bp = False

        super(SupervisedNetworkMethod, self).__init__(reduce_to_model_args, model_bp,
                                                      dataset=dataset, *args, **kwargs)


class SSMethod(NetworkMethod):
    def __init__(self, dataset, *args, **kwargs):
        reduce_to_model_args = {'use_stringdb': False}
        model_bp = False

        super(SSMethod, self).__init__(reduce_to_model_args, model_bp,
                                       dataset=dataset, *args, **kwargs)


class MitrofanovaMethod(NetworkMethod):
    def __init__(self, dataset, *args, **kwargs):
        reduce_to_model_args = {'use_stringdb': False}
        model_bp = False

        super(MitrofanovaMethod, self).__init__(reduce_to_model_args, model_bp,
                                                dataset=dataset, *args, **kwargs)

