from __future__ import division

import math
from itertools import chain

import enum
import igraph
import pandas as pd
from functools32 import lru_cache
from goatools.obo_parser import GODag
from patsy import dmatrices
from sqlalchemy import Column, Integer, String, Boolean, ForeignKey
from sqlalchemy import Enum, Float, PrimaryKeyConstraint
from sqlalchemy import orm

import effusion

get_go_dag = (lru_cache(maxsize=100))(GODag)


def truncate_name(name):
    MAX_WORDS = 3
    split_name = name.split()
    retval = ' '.join(split_name[0:MAX_WORDS])
    if len(split_name) > MAX_WORDS:
        retval += '...'

    return retval


class GoAspectEnum(enum.Enum):
    # C = 'C'
    # F = 'F'
    # P = 'P'
    cellular_component = 'C'
    molecular_function = 'F'
    biological_process = 'P'


class GoTerm(effusion.EffusionBase):
    __tablename__ = 'go_terms'

    go_id = Column(String(length=40), primary_key=True)
    name = Column(String(length=100))
    aspect = Column(Enum(GoAspectEnum))
    level = Column(Integer)
    depth = Column(Integer)  # max level
    catalytic = Column(Boolean)

    def __init__(self, go_id, name, aspect, level, depth, catalytic):
        self.go_id = go_id
        self.name = name
        self.aspect = aspect
        self.level = level
        self.depth = depth
        self.catalytic = catalytic


class DatasetGoTerm(effusion.EffusionBase):
    __tablename__ = 'dataset_go_terms'
    __table_args__ = (PrimaryKeyConstraint('dataset_name', 'go_id'),)

    # training, validation, or test
    dataset_name = Column(String(length=40))

    go_id = Column(ForeignKey(GoTerm.go_id))
    go_term = orm.relationship(GoTerm)

    conditional_ic = Column(Float())

    # sum of IC for all ancestor terms
    marginal_ic = Column(Float())

    def __init__(self, dataset_name, go_id, conditional_ic, marginal_ic):
        self.dataset_name = dataset_name
        self.go_id = go_id
        self.conditional_ic = conditional_ic
        self.marginal_ic = marginal_ic


class GeneOntology(igraph.Graph):
    def __init__(self, obo_file, catalytic_id, *args, **kwds):
        # Warning. Not called by class methods. Bug filed.
        """

        :param obo_file:
        :param catalytic_id:
        :param args:
        :param kwds:
        """

        obo_dag = get_go_dag(obo_file)
        catalytic_namespace = obo_dag[catalytic_id].namespace

        go_terms = set()
        edges = set()

        goatools_terms = set(obo_dag.values())  # ontology terms are replicated for alt_ids
        for goatools_term in goatools_terms:
            if goatools_term.is_obsolete:
                continue
            # nodes

            #  alternatively, attributes could be stored as igraph vertex attributes
            catalytic = (goatools_term.namespace == catalytic_namespace and
                         (goatools_term.id == catalytic_id or catalytic_id in goatools_term.get_all_parents()))

            go_term = GoTerm(goatools_term.id,
                             goatools_term.name,
                             goatools_term.namespace,
                             goatools_term.level,
                             goatools_term.depth,
                             catalytic)
            # do not add these to the database / session. we may want to load them from the DB instead

            go_terms.add(go_term)

            # edges
            if not goatools_term.parents:  # None or [] -> root term
                edges_as_go_ids = goatools_term.get_all_child_edges()
                edges_as_go_ids = [(e[1], e[0]) for e in edges_as_go_ids]
                edges.update(edges_as_go_ids)

        go_terms = sorted(go_terms, key=lambda t: t.level)
        go_ids = [t.go_id for t in go_terms]

        vertex_attrs = {'name': go_ids, 'term': go_terms}
        super(GeneOntology, self).__init__(n=len(go_ids),
                                           directed=True,
                                           vertex_attrs=vertex_attrs,
                                           *args,
                                           **kwds)
        self.add_edges(edges)

        self.catalytic_id = catalytic_id
        self.obo_dag = obo_dag

    def copy(self):
        retval = super(GeneOntology, self).copy()

        # addressing a bug in init / copy
        # retval.vs['state'] = False
        # retval.valid_states_cache = None
        # retval.get_valid_go_states = retval.get_valid_go_states_cached

        return retval

    # todo should these take a session, or use the global session?
    # todo can use the new_obj = session.merge(detached_) function too
    def save_to_db(self, session):
        go_terms = self.vs['term']
        session.add_all(go_terms)
        session.commit()

    def load_from_db(self, session):
        q = session.query(GoTerm)
        for go_term in q.all():
            go_vertex = self.vs.find(name=go_term.go_id)
            go_vertex['term'] = go_term

    def get_ic(self):
        retval = 0

        for v in self.vs:
            retval += v['term'].conditional_ic

        return retval

    def __hash__(self):
        return id(self)

    @property
    def root_mf_id(self):
        return self.vs.select(_indegree_eq=0).find(lambda v: v['term'].aspect == GoAspectEnum.molecular_function)[
            'name']

    @property
    def root_bp_id(self):
        return self.vs.select(_indegree_eq=0).find(lambda v: v['term'].aspect == GoAspectEnum.biological_process)[
            'name']

    @property
    @lru_cache()
    def root_ids(self):
        return set(self.vs.select(_indegree_eq=0)['name'])

    # colors: http://igraph.org/python/doc/igraph.drawing.colors-pysrc.html
    # lemon_chiffon_rgb = (1., .98, .8)
    # lavender_rgb = (.9, .9, .98)
    # red_rgb = (1.0, 0.0, 0.0)
    # blue_rgb = (0.0, 0.0, 1.0)
    # black_rgb = (0.0, 0.0, 0.0)
    # green_rgb = (0.0, 1.0, 0.0)
    # grey_rgb = (0.8, 0.8, 0.8)
    def plot(self, target, protein_node=None, vertex_color=(0.8, 0.8, 0.8)):
        # todo(jyunes) also plot correlations

        layout = self.layout("tree")
        self.vs["label"] = self.vs["name"]
        # shape_map = {'variable': 'circle', 'factor': 'rectangle'}
        # self.vs["shape"] = [shape_map[v['type']] for v in self.vs]
        # todo(jyunes) curved edges and lots of other fun options here:
        # http://igraph.org/python/doc/igraph.Graph-class.html#__plot__
        # put the first two words of the go term in parens

        vertex_labels = ['%s (%s)' % (v['term'].go_id, truncate_name(v['term'].name)) for v in self.vs]
        # vertex_label_dists = [(1 + random.randint(0, 3)) for _ in vertex_labels]
        vertex_label_dists = [2 for _ in vertex_labels]

        # vertex_colors = [green_rgb if v['term'].ontology == 'molecular_function' else red_rgb for v in self.vs]
        vertex_colors = [vertex_color for v in self.vs]

        if protein_node:
            for i, v in enumerate(self.vs):
                prob = protein_node[v['name']]
                old_color = vertex_colors[i]

                # shade
                shade_factor = 1 - prob
                vertex_colors[i] = map(lambda x: x * shade_factor, old_color)

                # tint
                vertex_colors[i] = map(lambda x: x + ((1 - prob) * (1 - x)), old_color)

        igraph.plot(self,
                    target=target,
                    layout=layout,
                    margin=(120, 20, 120, 60),
                    edge_curved=.05,
                    vertex_shape='circle',
                    vertex_size=40,
                    vertex_label=vertex_labels,
                    vertex_label_dist=vertex_label_dists,
                    vertex_color=vertex_colors,
                    vertex_label_size=7,
                    vertex_label_angle=-math.pi / 2,
                    bbox=(4096, 768),
                    # bbox=(8000, 768),
                    background=None)

        # "shape 3 is a rectangle that automatically sizes to the inner text."
        # shapes = [3] * len(self.vs)
        # self.write_svg(fname, layout="tree", shapes=shapes)

    def get_ancestors(self, v):
        assert self.is_directed()
        # this includes v.index
        ai = self.neighborhood(v, order=9999, mode=igraph.IN)
        a = self.vs[ai]
        return a

    @lru_cache(maxsize=None)
    def get_descendants(self, v):
        assert self.is_directed()
        # this includes v.index
        ai = self.neighborhood(v, order=9999, mode=igraph.OUT)
        a = self.vs[ai]
        a = set(a)
        return a

    def get_sibling_vertices(self, go_vertex):
        retval = set()
        for parent_vertex in go_vertex.predecessors():
            for sibling_vertex in parent_vertex.successors():
                if sibling_vertex == go_vertex:
                    continue
                retval.add(sibling_vertex)

        return retval

    # todo pretty specific functions. but needed it.
    @lru_cache(maxsize=None)
    def get_sibling_ids(self, v):
        retval = set()

        for parent_vertex in v.predecessors():
            # includes siblings by any parent and does NOT include v.

            retval |= {sibling_vertex['name'] for sibling_vertex in parent_vertex.successors() if
                       sibling_vertex != v}

        return retval

    def get_siblings_relative_to_parent(self, v):
        """

        :param v:
        :return: yields (parent_id, sibling_id) pairs. does not include v as a sibling
        """
        for parent_vertex in v.predecessors():
            # includes siblings by any parent and does NOT include v.

            for sibling_vertex in parent_vertex.successors():
                if sibling_vertex == v:
                    continue

                else:
                    yield (parent_vertex['name'], sibling_vertex['name'])

    @lru_cache(maxsize=None)
    def get_parent_go_ids(self, v):
        return {parent_v['name'] for parent_v in v.predecessors()}

    def get_mrca(self, vs):
        all_ancestors = []

        for v in vs:
            ancestors = set(self.get_ancestors(v))
            all_ancestors.append(ancestors)

        ancestors_in_common = set.intersection(*all_ancestors)
        mrca_term = None
        for a in ancestors_in_common:
            if mrca_term is None or a['term'].depth > mrca_term.depth:
                mrca_term = a['term']

        return mrca_term

    def hasId(self, go_id):
        try:
            self.vertex4id(go_id)
        except ValueError:
            return False
        return True

    def vertex4id(self, go_id):
        return self.vs.find(go_id)

    def term4id(self, go_id):
        return self.vs.find(go_id)['term']

    @lru_cache()  # return the same gene_ontology for the same terms.
    # this is nice, unless we're trying lots of subgraphs and using the GO once
    def subgraph(self, go_ids, keep_ancestors):
        print 'computing subgraph for %s' % repr(self)
        retval = self.copy()

        retval.vs['prune'] = True
        for go_id in go_ids:
            v = retval.vs.find(go_id)
            v['prune'] = False

            if keep_ancestors:
                for ancestor in retval.get_ancestors(v):
                    ancestor['prune'] = False

        retval.vs.select(prune=True).delete()

        return retval

    def subgraph_for_aspect(self, aspect):
        retval = self.copy()

        retval.vs.select(lambda x: x['term'].aspect != aspect).delete()

        return retval

    @lru_cache()
    def get_vertices_for_ontology(self, ontology):
        retval = set()

        for v in self.vs:
            term = v['term']
            if term.aspect == ontology:
                retval.add(v)

        return frozenset(retval)

    @lru_cache()
    def get_ids_for_ontology(self, ontology):
        retval = {v['name'] for v in self.get_vertices_for_ontology()}
        return frozenset(retval)

    def imply_negative_go_ids(self, positive_go_ids):
        """ If the protein is annnotated to a term that's unlikely to moonlight with a sibling, mark those siblings as negative"""

        negative_go_ids = set()

        for aspect in GoAspectEnum:
            vertices_for_ontology = self.get_vertices_for_ontology(aspect)
            ids_for_ontology = {v['name'] for v in vertices_for_ontology}
            if not positive_go_ids & ids_for_ontology:
                continue

            for go_vertex in vertices_for_ontology:
                go_term = go_vertex['term']
                go_id = go_vertex['name']

                if go_id in positive_go_ids:  # todo or if it's negatively annotated, but shouldn't have any effect
                    continue

                if go_id in negative_go_ids:  # it was already a descendant of a negative go id, then we don't need to check it again
                    continue

                # if this term is unlikely given the other annotations, mark as negative
                # however, there's not a lot of counts for a specific set of annotations...
                # limit it to positive annotations of any modeled siblings

                sibling_ids = self.get_sibling_ids(go_vertex)
                sibling_ids_with_positive_evidence = sibling_ids & positive_go_ids
                if not len(sibling_ids_with_positive_evidence):
                    continue

                # method 1: get p(self | siblings) = p(self, siblings) / p(siblings).
                # requires at least two queries, but possibly as many as #(siblings), since it has to run once for each self in siblings

                # method 2. calculate one query per term with siblings, and marginalize ourselves
                # extract get_marginalized_contingency_table
                siblings_plus_self = sibling_ids | {go_id}

                contingency_table = effusion.get_single_contingency_table(siblings_plus_self,
                                                                          weighted=True,
                                                                          require_mf=(
                                                                              go_term.aspect == GoAspectEnum.molecular_function),
                                                                          require_bp=(
                                                                              go_term.aspect == GoAspectEnum.biological_process),
                                                                          add_pseudocounts=False)
                event_state = {go_id: True}
                given_state = {sibling_id: True for sibling_id in sibling_ids_with_positive_evidence}

                conditional = effusion.get_single_protein_conditional(event_state, given_state, weighted=None,
                                                                      require_mf=None, require_bp=None,
                                                                      add_pseudocounts=False,
                                                                      contingency_table=contingency_table)

                if conditional >= .5:  # todo optimize parameter
                    continue
                descendants_ids = {d['name'] for d in self.get_descendants(go_vertex)}
                negative_go_ids |= descendants_ids

        return negative_go_ids

    def imply_negative_uif_go_ids(self, positive_go_ids):
        negative_go_ids = set()

        for root_id in self.root_ids:
            if root_id not in positive_go_ids:
                continue

            root_vertex = self.vs.find(root_id)
            ontology = root_vertex['term'].aspect

            ontology_ids = self.get_ids_for_ontology(ontology)
            negative_go_ids |= ontology_ids - positive_go_ids  # ... that don't have an annotation

        return negative_go_ids

    def load_fly_contingency_tables(self):
        mf_vertices = None
        bp_vertices = None

        for mf_vertex in mf_vertices:
            self.load_fly_contingency_table_for_go_id(mf_vertex, bp_vertices)

    def load_fly_contingency_table_for_go_id(self, go_vertex, interont_vertices):

        # part 1: what are the best correlations?
        # step 1: take a bunch of samples
        # let's take interesting samples:
        # * for top down, this means anything that has the parent annotations
        # * for bottom up, this means anything that doesn't have a child annotation (which is a much larger set)
        # ^ take 50% that have self annotation = True, and 50% that don't have self annotation

        if effusion.args.top_down_or_bottom_up == 'top_down':
            gok_ids = self.predecessors(go_vertex)
            needed_gok_value = True
        elif effusion.args.top_down_or_bottom_up == 'bottom_up':
            gok_ids = self.successors(go_vertex)
            needed_gok_value = False
        else:
            raise

        evidence = {}
        for gok_id in gok_ids:
            evidence[gok_id] = needed_gok_value

        evidence[go_vertex['name']] = True
        uniprots_with_go_val_one = effusion.Annotation.get_uniprots_matching_evidence(evidence, require_mf=True,
                                                                                      require_bp=True, limit=1000)

        evidence[go_vertex['name']] = False
        uniprots_with_go_val_zero = effusion.Annotation.get_uniprots_matching_evidence(evidence, require_mf=True,
                                                                                       require_bp=True, limit=1000)

        # step 2: compute their labels
        labels = {}
        for uniprot in chain(uniprots_with_go_val_one, uniprots_with_go_val_zero):
            annotations = effusion.Annotation.get_goas_for_uniprot(uniprot, include_negative=True, include_direct=True,
                                                                   include_implied=True, include_moonlight=False)
            # but we've already limited to uniprots with mf and bp, so we can probably do this more quickly...
            implicitly_negative_annotations = effusion.Annotation.get_uif_annotations(
                gene_ontology=self,
                annotations=annotations)
            label = effusion.Label(gene_ontology=self,
                                   all_annotations=chain(annotations, implicitly_negative_annotations))
            labels[uniprot] = label

        df = pd.DataFrame.from_dict(labels, orient='index')
        y, X = dmatrices('affair ~ rate_marriage + age + yrs_married + children + \
                          religious + educ + C(occupation) + C(occupation_husb)',
                         df, return_type="dataframe")
        print X.columns

        # step 3: run logistic regression with elastic net and interaction terms and pick out the best correlations


        # part 2: calculate the contingency table for the best features

        return
