from __future__ import division

__author__ = 'jeff'

from itertools import chain
import logging
# import opengm
from collections import defaultdict

import igraph
import os

import cyjs_utils

import effusion
import gzip
import json


class ProteinNetwork(igraph.Graph):
    def __init__(self, *args, **kwargs):
        super(ProteinNetwork, self).__init__(*args, **kwargs)

        # Warning. Not called by class methods. Bug filed.

    def __str__(self):
        """Returns a string representation of the graph.
        Behind the scenes, this method constructs a L{GraphSummary}
        instance and invokes its C{__str__} method with a verbosity of 1
        and attribute printing turned off.
        See the documentation of L{GraphSummary} for more details about the
        output.
        """
        params = dict(
            verbosity=0,
            width=78,
            print_graph_attributes=False,
            print_vertex_attributes=False,
            print_edge_attributes=False
        )

        return self.summary(**params)

    @classmethod
    def make_toy_network(cls, num_proteins, children, annotations_dict):

        # todo(jyunes) warning. factor graph must be tree. why again? see coin toss example
        # protein_network = cls(n=num_proteins)
        protein_network = cls.Tree(num_proteins, children=children, type=igraph.TREE_OUT)
        protein_network.vs['name'] = ['Protein %d' % i for i in range(len(protein_network.vs))]

        # protein_template = ToyProteinTemplateGenerator(num_terms, children=2).get_protein_template()
        # make_protein = protein_factory(protein_template)
        # protein_network.imbue_network_with_template(make_protein)

        protein_network.es['score'] = 1
        protein_network.es['type'] = 'blast'
        protein_network.vs['query'] = True
        protein_network.vs['round_added'] = 0

        if annotations_dict:
            protein_network.set_evidence(annotations_dict)

        return protein_network

    @classmethod
    @effusion.log_time()
    def build_direct_net(cls, query_uniprot, net_cache=None, force=False, do_string=True):
        # For practical purposes (xval), the network topology must not change depending on the evidence.
        # So it is not a parameter. Evidence is added here at the end just for display purposes.

        logging.debug('building direct net for %s' % query_uniprot)

        if net_cache and os.path.isfile(net_cache) and not force:
            retval = cls.from_json_gz(net_cache)
            logging.debug('returning prebuilt network')
            return retval

        all_uniprots = set()
        name_idx_map = {}
        node_attrs = defaultdict(list)

        # round 0: queries
        query_uniprots = {query_uniprot}
        for query_uniprot in query_uniprots:
            node_attrs['name'].append(query_uniprot)
            node_attrs['query'].append(True)
            node_attrs['round_added'].append(0)
            name_idx_map[query_uniprot] = len(name_idx_map)
        all_uniprots |= query_uniprots

        # round 1: blast
        blast_uniprots = effusion.blast(query_accessions=query_uniprots)
        new_blast_uniprots = blast_uniprots - all_uniprots
        logging.info('Added %d new proteins from blast 1' % len(new_blast_uniprots))
        for new_uniprot in new_blast_uniprots:
            node_attrs['name'].append(new_uniprot)
            node_attrs['query'].append(False)
            node_attrs['round_added'].append(1)
            name_idx_map[new_uniprot] = len(name_idx_map)
        all_uniprots |= new_blast_uniprots

        # round 2: string
        string_uniprots, string_edges = set(), set()
        if do_string:
            string_uniprots, string_edges = effusion.query_string(query_uniprots | new_blast_uniprots,
                                                                  min_string_score=effusion.args.min_string_score)

        new_string_uniprots = string_uniprots - all_uniprots
        logging.info('Added %d new proteins from string' % len(new_string_uniprots))
        for new_uniprot in new_string_uniprots:
            node_attrs['name'].append(new_uniprot)
            node_attrs['query'].append(False)
            node_attrs['round_added'].append(2)
            name_idx_map[new_uniprot] = len(name_idx_map)
        all_uniprots |= new_string_uniprots

        edges_with_indices = []
        edge_attrs = {'type': [],
                      'score': [],
                      'evalue': []}

        blast_edges = effusion.all_by_all(uniprots=all_uniprots)
        for edge in blast_edges:
            edges_with_indices.append((name_idx_map[edge[0]], name_idx_map[edge[1]]))
            edge_attrs['type'].append('blast')
            edge_attrs['score'].append(edge[2])
            edge_attrs['evalue'].append(edge[3])

        for edge in string_edges:
            edges_with_indices.append((name_idx_map[edge[0]], name_idx_map[edge[1]]))
            edge_attrs['type'].append('string')
            edge_attrs['score'].append(edge[2])
            edge_attrs['evalue'].append(None)  # length of lists in edge_attrs must be the same...

        # each values of node/edge attr dictionary is a list. len(each list) == # nodes/edges
        node_attr_lengths = [len(node_attr_list) for node_attr_list in node_attrs.values()]
        assert all(x == len(all_uniprots) for x in node_attr_lengths)
        edge_attr_lengths = [len(edge_attr_list) for edge_attr_list in edge_attrs.values()]
        assert all(x == len(edges_with_indices) for x in edge_attr_lengths)
        g = cls(n=len(all_uniprots), edges=edges_with_indices, directed=True, vertex_attrs=node_attrs,
                edge_attrs=edge_attrs)

        if not g.is_connected(mode=igraph.WEAK):
            components = g.decompose(mode=igraph.WEAK)
            logging.warning('There are %d components' % len(components))

            num_disconnected_nodes = g.vcount() - len(g.subcomponent(g.vs.find(name=query_uniprot).index))
            logging.warning('%d vertices are disconnected from the query' % num_disconnected_nodes)

        if net_cache:
            g.to_json_gz(net_cache)

        return g

    def to_json_gz(self, jgz_file_path):
        r = self.__reduce__()
        r = list(r)
        r[0] = r[0].__name__

        with gzip.open(jgz_file_path, 'wb') as f:
            json.dump(r, f)

    @classmethod
    def from_json_gz(cls, jgz_file_path):
        """ exactly the same as the pickle process, except constructor is hardcoded """

        with gzip.open(jgz_file_path, 'rb') as f:
            r = json.load(f)
            classname, constructor_args, state_dict = r
            # todo(jyunes) assert classname matches?
            retval = cls(*constructor_args)
            retval.__dict__.update(state_dict)

        return retval

    def export_cyjs(self, path):
        cyjs_utils.export_graph_as_cyjs(self, path)

    # todo(jyunes) cls.Tree doesn't call ProteinNetwork.__init__
    def set_evidence(self, annotations_dict):
        """
        Just for display...
        :param annotations_dict:
        :return:
        """
        self.vs['goas'] = [set() for _ in range(len(self.vs))]
        self.vs['- goas'] = [set() for _ in range(len(self.vs))]
        self.vs['evidence'] = False

        for uniprot, annotations in annotations_dict.items():
            try:
                vertex = self.vs.find(uniprot)

                for annotation in annotations:
                    if annotation.value:
                        vertex['goas'].add(annotation.go_id)
                    else:
                        vertex['- goas'].add(annotation.go_id)
                    vertex['evidence'] = True
            except ValueError as e:
                logging.warning('Evidence for %s but not in network' % uniprot)

    def reduce_to_model(self, use_query=True, use_similarity=True, use_network=True, use_unlabeled=True,
                        use_stringdb=True, annotations_dict=None):
        # todo(jyunes) maybe weird coupling
        if annotations_dict is not None:
            self.set_evidence(annotations_dict)

        if not use_query:
            self.vs.delete()
            return

        elif not use_similarity:
            self.vs.select(query=False).delete()
            self.es.delete()
            return

        if not use_network:
            # source or target must be a query
            query_vids = [v.index for v in self.vs.select(query=True)]
            self.es['first_round_edge'] = False
            self.es.select(_source_in=query_vids)['first_round_edge'] = True
            self.es.select(_target_in=query_vids)['first_round_edge'] = True
            self.es.select(first_round_edge=False).delete()

        if not use_unlabeled:
            assert annotations_dict is not None

            # of the nodes that aren't queries, delete the ones without evidence
            self.vs.select(query=False).select(evidence=False).delete()

        if not use_stringdb:
            if 'type' not in self.es.attribute_names():
                assert len(self.es) == 0
            else:
                self.es.select(type='string').delete()

        # keep only the vertices reachable from the query/ies
        vids = [v.index for v in self.vs.select(query=True)]
        list_of_neighbors_vids = self.neighborhood(vids, order=9999)
        neighbor_vis = list(chain.from_iterable(list_of_neighbors_vids))
        neighbor_names = self.vs.select(neighbor_vis)['name']
        self.vs.select(name_notin=neighbor_names).delete()

        return

    def make_tractable(self, important_node_names):
        """
        :param important_node_names:
        :return: A more tractable network.
        """

        # calculate edge scores
        types = self.es['type']
        score_multiplier_map = {'blast': effusion.args.blast_score_multiplier,
                                'string': effusion.args.string_score_multiplier}
        scores = self.es['score']  # should probably be renamed raw_score
        adjusted_scores = [score * score_multiplier_map[type] for score, type in effusion.zip_equal(scores, types)]
        self.es['adjusted_score'] = adjusted_scores

        # keep the best edge between two vertices
        best_edge = {}
        for e in self.es:
            endpoints = min(e.source, e.target), max(e.source, e.target)
            if endpoints not in best_edge or e['adjusted_score'] > best_edge[endpoints]['adjusted_score']:
                best_edge[endpoints] = e
        for e in best_edge.values():
            e['best_edge'] = True
        self.es.select(best_edge_ne=True).delete()

        # remove self loops
        self.simplify(multiple=False, loops=True)

        assert self.is_simple()

        t = self.mst('adjusted_score', max_spanning_tree=True)  # returns mst, but self will be annotated
        pruned_node_names = t.prune(important_node_names)
        self.vs.select(name_in=pruned_node_names).delete()

        return self

    def mst(self, weight_column, max_spanning_tree):
        # todo(jyunes) don't love modification on self

        self.es['MST'] = False

        weights = self.es[weight_column]
        if max_spanning_tree:
            weights = [-w for w in weights]

        i = 0
        logging.debug('Sample edges input for MST:')

        t = self.spanning_tree(weights, return_tree=True)
        # assert the vertex ids match
        assert t.vs['name'] == self.vs['name']

        pairs = [(e.source, e.target) for e in t.es]
        mst_edge_ids = self.get_eids(pairs=pairs, directed=False)
        self.es[mst_edge_ids]['MST'] = True

        return t

    def prune(self, important_node_names):
        important_nodes = self.vs.select(name_in=important_node_names)

        self.vs['pruned'] = True

        query_node = self.vs.find(query=True)
        keepers = set(important_nodes)
        for important_node in important_nodes:
            path_ids = self.get_shortest_paths(important_node, query_node, mode=igraph.ALL)[0]
            keepers |= set(self.vs[path_ids])

        self.vs[[n.index for n in keepers]]['pruned'] = False

        pruned_node_names = {v['name'] for v in self.vs.select(pruned=True)}
        self.vs.select(pruned=True).delete()

        return pruned_node_names
