from __future__ import division

import logging
from collections import defaultdict
from itertools import chain

import enum
from functools32 import lru_cache
from sqlalchemy import Column, Integer, String
from sqlalchemy import Enum
from sqlalchemy import func, and_
from sqlalchemy.orm import aliased
from sqlalchemy.sql.expression import literal_column

import effusion


class EdgeTypeEnum(enum.Enum):
    blast = 'blast'
    string = 'string'


class ProteinNetworkParent(effusion.EffusionBase):
    __tablename__ = 'protein_network_parents'

    # __table_args__ = (PrimaryKeyConstraint('dataset_name', 'uniprot'), )
    id = Column(Integer, primary_key=True)

    # training, validation, or test
    dataset_name = Column(String(length=40))

    # similarity_type
    # todo make an enum?
    edge_type = Column(Enum(EdgeTypeEnum))

    uniprot = Column(String(length=10))
    parent_uniprot = Column(String(length=10))

    def __init__(self, dataset_name, edge_type, uniprot, parent_uniprot):
        self.dataset_name = dataset_name
        self.edge_type = edge_type
        self.uniprot = uniprot
        self.parent_uniprot = parent_uniprot

    def __str__(self):
        return '(%s, %s, %s)' % (self.dataset_name, self.uniprot, self.parent_uniprot)

    def __repr__(self):
        return str(self)


def get_pairwise_protein_conditional(
        event_state,
        given_gok_state,  # aka gok
        given_pnp_state,  # aka pnp. len() == 1
        edge_type,
        weighted,
        require_mf,
        require_bp,
        add_pseudocounts):
    go_ids_for_this_protein = set(chain(event_state.keys(), given_gok_state.keys()))
    go_ids_for_parent_protein = given_pnp_state.keys()

    contingency_table = get_pairwise_contingency_table(go_ids_for_this_protein,
                                                       go_ids_for_parent_protein,
                                                       edge_type,
                                                       weighted=weighted,
                                                       require_mf=require_mf,
                                                       require_bp=require_bp,
                                                       add_pseudocounts=add_pseudocounts)
    # pprint(contingency_table)

    # pnp keys are prepended with pnp_ to avoid clashes
    given_pnp_state_with_prefix = {'pnp_%s' % k: v for k, v in given_pnp_state.iteritems()}

    given_count = 0
    for state, count in contingency_table.iteritems():
        state_matches_given_for_this = all(state[k] == v for k, v in given_gok_state.iteritems())
        state_matches_given_for_parent = all(state[k] == v for k, v in given_pnp_state_with_prefix.iteritems())

        if state_matches_given_for_this and state_matches_given_for_parent:
            given_count += count

    # todo make sure given count is significant for calculating statistics?

    joint_state = given_gok_state.copy()
    joint_state.update(given_pnp_state_with_prefix)
    joint_state.update(event_state)
    joint_count = 0
    for state, count in contingency_table.iteritems():
        state_matches_joint = all(state[k] == v for k, v in joint_state.iteritems())
        if state_matches_joint:
            joint_count += count

    # possible, e.g., p(8150 | 3674 = 0)
    # let's not give that a prob = 1 or 0.
    # with pseudocounts, 1/1 = 1, and that made \sum p(8150) and \sum p(3674) not always = 1!

    # print 'conditional = joint_count / given'
    # print 'joint_count:', joint_count, 'given_count:', given_count

    if given_count == 0:
        conditional = .5
    else:
        conditional = joint_count / given_count

    return conditional


def get_pairwise_contingency_table(go_ids_for_this_protein,
                                   go_ids_for_parent_protein,
                                   edge_type,
                                   weighted,
                                   require_mf,
                                   require_bp,
                                   add_pseudocounts):
    go_ids_for_this_protein = tuple(sorted(go_ids_for_this_protein))
    go_ids_for_parent_protein = tuple(sorted(go_ids_for_parent_protein))

    retval = get_pairwise_contingency_table_impl(go_ids_for_this_protein, go_ids_for_parent_protein, edge_type,
                                                 weighted, require_mf, require_bp)

    if add_pseudocounts:
        # see comments in annotation.get_pairwise_contingency_table

        all_go_ids_have_a_parent = True
        parent_id_to_child_ids_map = defaultdict(set)
        parent_go_ids_for_this_protein = set()
        for go_id in go_ids_for_this_protein:
            go_vertex = effusion.dataset.gene_ontology.vertex4id(go_id)
            parent_go_ids = effusion.dataset.gene_ontology.get_parent_go_ids(go_vertex)

            if len(parent_go_ids) == 0:
                all_go_ids_have_a_parent = False
                break

            mrca_id = effusion.dataset.gene_ontology.get_mrca(parent_go_ids).go_id
            parent_id_to_child_ids_map[mrca_id].add(go_id)
            parent_go_ids_for_this_protein.add(mrca_id)

        parent_go_ids_for_parent_protein = set()
        for go_id in go_ids_for_parent_protein:
            go_vertex = effusion.dataset.gene_ontology.vertex4id(go_id)
            parent_go_ids = effusion.dataset.gene_ontology.get_parent_go_ids(go_vertex)

            if len(parent_go_ids) == 0:
                all_go_ids_have_a_parent = False
                break

            mrca_id = effusion.dataset.gene_ontology.get_mrca(parent_go_ids).go_id
            parent_id_to_child_ids_map['pnp_' + mrca_id].add('pnp_' + go_id)
            parent_go_ids_for_parent_protein.add(mrca_id)

        if all_go_ids_have_a_parent:
            # this also uses pseudocounts.
            # otherwise, any leaf (which must have an annotation to be modeled) that has a parent with only 1 pairwise evidence would force a 00 or 11 for the root (and all its ancestors).
            # happened a lot
            parents_contingency_table = get_pairwise_contingency_table(parent_go_ids_for_this_protein,
                                                                       parent_go_ids_for_parent_protein,
                                                                       edge_type,
                                                                       weighted,
                                                                       require_mf,
                                                                       require_bp,
                                                                       add_pseudocounts=True)

            # we don't want to modify the cached contingency table
            retval = retval.copy()

            for state_in_parents_contingency_table, pseudocount in parents_contingency_table.iteritems():
                state_in_contingency_table = {}
                for parent_go_id, val_of_parent_go_id in state_in_parents_contingency_table.iteritems():
                    for child_go_id in parent_id_to_child_ids_map[parent_go_id]:
                        state_in_contingency_table[child_go_id] = val_of_parent_go_id
                state_in_contingency_table = effusion.ImmutableDict(state_in_contingency_table)
                retval[state_in_contingency_table] = retval.get(state_in_contingency_table,
                                                                0) + effusion.args.pseudocount_weight * pseudocount

    return retval


"""
attach database 'protein_network_parents.db' as pnp;
attach database 'goa.db' as goa_db;

select sum(raw_count) as raw_count, sum(weighted_count) as weighted_count, val, pnp, gok1, gok2 from (
select pnp.uniprot, pnp.parent_uniprot, 1 as raw_count, du.ic as weighted_count, a1.value, a2.value, ifnull(a3.value, 0) as val, ifnull(a4.value, 0) as pnp, ifnull(a5.value, 0) as gok1, ifnull(a6.value, 0) as gok2
from protein_network_parents pnp
join goa_db.dataset_uniprots du on pnp.uniprot = du.uniprot
join goa_db.annotations a1 on a1.uniprot = pnp.uniprot and a1.go_id = 'GO:0003674'
join goa_db.annotations a2 on a2.uniprot = pnp.parent_uniprot and a2.go_id = 'GO:0003674'
left outer join goa_db.annotations a3 on a3.uniprot = pnp.uniprot and a3.go_id = 'GO:0003824'
left outer join goa_db.annotations a4 on a4.uniprot = pnp.parent_uniprot and a4.go_id = 'GO:0003824'
left outer join goa_db.annotations a5 on a5.uniprot = pnp.uniprot and a5.go_id = 'GO:0016787'
left outer join goa_db.annotations a6 on a6.uniprot = pnp.uniprot and a6.go_id = 'GO:0016740'
where pnp.dataset_name = 'training'
and pnp.edge_type = 'string'
and du.dataset_name = 'training'
) group by val, pnp, gok1, gok2;

returns:

raw_count|weighted_count|val|pnp|gok1|gok2
19543|2452544.2462385|0|0|0|0
5895|800879.633019382|0|1|0|0
708|99964.4247973069|1|0|0|0
1947|375816.70078233|1|0|0|1
2212|357440.99992628|1|0|1|0
38|10081.8987743797|1|0|1|1
4654|625515.168288077|1|1|0|0
7754|1041603.18900976|1|1|0|1
5190|742457.929827927|1|1|1|0
209|40293.5154826046|1|1|1|1

TODO XXX make sure to apply standard filter to annotations
"""


@lru_cache(maxsize=None)
def get_pairwise_contingency_table_impl(go_ids_for_this_protein,
                                        go_ids_for_parent_protein,
                                        edge_type,
                                        weighted,
                                        require_mf,
                                        require_bp):
    assert effusion.dataset.cutoff_date

    logging.debug('Getting pairwise protein contingency table for %s %s' % (
        str(go_ids_for_this_protein), str(go_ids_for_parent_protein)))

    annotation_aliases_for_this_protein = []
    ifnull_columns = []
    for go_id in go_ids_for_this_protein:
        a = aliased(effusion.Annotation)
        annotation_aliases_for_this_protein.append(a)
        ifnull_columns.append(func.ifnull(a.value, literal_column('0')).label(go_id))

    annotation_aliases_for_parent_protein = []
    for go_id in go_ids_for_parent_protein:
        a = aliased(effusion.Annotation)
        annotation_aliases_for_parent_protein.append(a)
        ifnull_columns.append(func.ifnull(a.value, literal_column('0')).label('pnp_%s' % go_id))

    q = effusion.dataset.database_session.query(ProteinNetworkParent.uniprot,
                                                ProteinNetworkParent.parent_uniprot,
                                                literal_column('1').label('raw_count'),
                                                effusion.DatasetUniprot.ic.label('weighted_count'),
                                                *ifnull_columns)

    q = q.join(effusion.DatasetUniprot, and_(ProteinNetworkParent.uniprot == effusion.DatasetUniprot.uniprot,
                                             effusion.DatasetUniprot.dataset_name == effusion.dataset.dataset_cutoff_name))
    du_alias_for_parent_protein = aliased(effusion.DatasetUniprot)
    q = q.join(du_alias_for_parent_protein,
               and_(ProteinNetworkParent.parent_uniprot == du_alias_for_parent_protein.uniprot,
                    du_alias_for_parent_protein.dataset_name == effusion.dataset.dataset_cutoff_name))

    for go_id, a in zip(go_ids_for_this_protein, annotation_aliases_for_this_protein):
        q = q.outerjoin(a, and_(a.uniprot == ProteinNetworkParent.uniprot,
                                a.date <= effusion.dataset.cutoff_date,
                                a.source.in_(('direct', 'implied')),
                                a.go_id == go_id))

    for go_id, a in zip(go_ids_for_parent_protein, annotation_aliases_for_parent_protein):
        q = q.outerjoin(a, and_(a.uniprot == ProteinNetworkParent.parent_uniprot,
                                a.date <= effusion.dataset.cutoff_date,
                                a.source.in_(('direct', 'implied')),
                                a.go_id == go_id))

    q = q.filter(ProteinNetworkParent.edge_type == edge_type)
    q = q.filter(ProteinNetworkParent.dataset_name == effusion.dataset.dataset_cutoff_name)

    # can use this  to limit to terms with mf. can also join to 3674
    if require_mf:
        q = q.filter(effusion.DatasetUniprot.mf_ic > 0)
        q = q.filter(du_alias_for_parent_protein.mf_ic > 0)
    if require_bp:
        q = q.filter(effusion.DatasetUniprot.bp_ic > 0)
        q = q.filter(du_alias_for_parent_protein.bp_ic > 0)

    s = q.subquery()

    subquery_columns = []
    for ifnull_column in ifnull_columns:
        subquery_columns.append(s.c[ifnull_column.key])
        # subquery_columns.append(ifnull_column)  # doing this adds another join to annotation for some reason...

    q = effusion.dataset.database_session.query(func.sum(s.c.raw_count).label('raw_count'),
                                                func.sum(s.c.weighted_count).label('weighted_count'),
                                                *subquery_columns).select_from(s).group_by(*subquery_columns)

    # print q
    retval = {}
    for row in q:
        if weighted:
            count = row[1]
        else:
            count = row[0]
        state = row[2:]
        state_keys = row.keys()[2:]  # lots of options for where to get this
        go_state = effusion.ImmutableDict({k: bool(v) for k, v in zip(state_keys, state)})
        retval[go_state] = count

    return retval
