__author__ = 'jeff'

import cStringIO
import logging
import multiprocessing
import subprocess
import sys
import tempfile
import uuid

import os
import re
from Bio import SeqIO
from Bio.Blast import NCBIXML
from Bio.Blast.Applications import NcbiblastpCommandline
from sqlalchemy import func, and_, distinct

import effusion


class ProteinSearcher(object):
    def __init__(self):
        super(ProteinSearcher, self).__init__()

    def filter_results_to_proteins_within(self, results, only_proteins_within):
        assert isinstance(only_proteins_within, set)

        for query_uniprot, hits in results.items():
            results[query_uniprot] = [hit for hit in hits if hit['uniprot'] in only_proteins_within]
            # Sometimes, after this filtering, hits will be empty.

        return results


def write_fasta_for_accessions(blast_db, accessions):
    query_fasta_file = tempfile.NamedTemporaryFile(delete=False, dir=effusion.args.tmp_dir)
    seq_records = get_seq_records_for_accessions_batch(blast_db, accessions)
    SeqIO.write(seq_records, query_fasta_file, "fasta")
    query_fasta_file.close()
    return query_fasta_file.name


def get_seq_records_for_accessions_batch(blast_db, accessions):
    """
    Gets the sequence of the sequence of the representative
    :param accessions:
    :return:
    """
    with tempfile.NamedTemporaryFile(delete=False, dir=effusion.args.tmp_dir) as accesssions_file:
        for accession in accessions:
            accesssions_file.write("%s\n" % accession)
    cmd = "%s -db %s -entry_batch %s -outfmt %%f" % (
        effusion.args.blastdbcmd_bin, blast_db, accesssions_file.name)
    logging.debug('get_seq_records_for_accessions_batch cmd: %s' % cmd)
    my_popen = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    stdout, stderr = my_popen.communicate()

    if stderr: logging.error(stderr)
    # diamond doesn't parse out the database. blast will remove it, but doesn't care...
    stdout = re.sub(r'>(?:sp|tr|lcl)\|(?P<accession>\w+)\|.*\n', r'>\g<accession>\n', stdout)
    # stdout = stdout.replace('lcl|', '')
    records = SeqIO.parse(cStringIO.StringIO(stdout), "fasta")

    if effusion.args.remove_tmp_files:
        os.remove(accesssions_file.name)

    return list(records)


def convert_blast_results_to_edge_tuples(blast_results):
    for query_uniprot, hits in blast_results.items():
        for blast_result in blast_results:
            yield (query_uniprot, blast_result['uniprot'], blast_result['bit_score_per_residue'])


def blast(query_accessions, evalue=None, max_target_seqs=None):
    """
    :param query_accessions:
    :return:
    """
    blast_db = effusion.args.uniprot_db
    evalue = evalue if evalue else effusion.args.evalue
    max_target_seqs = max_target_seqs if max_target_seqs else effusion.args.max_target_seqs
    min_query_cover = effusion.args.min_query_cover

    query_accessions = list(query_accessions)  # Ensuring an ordered collection
    query_fasta_file_path = write_fasta_for_accessions(blast_db, query_accessions)
    if os.stat(query_fasta_file_path).st_size == 0:
        logging.warning('Nothing to BLAST, given %d query accessions' % len(query_accessions))
        return set()

    # todo convert min_bit_score_per_residue to evalues OR switch to diamond
    # for now, just use a more lenient evalue and filter to min_bit_score_per_residue
    blast_cline = NcbiblastpCommandline(cmd=effusion.args.blastp_bin,
                                        task="blastp",  # could also use blastp-fast, but was missed some targets
                                        query=query_fasta_file_path,
                                        db=blast_db,
                                        max_hsps=1,
                                        num_threads=multiprocessing.cpu_count(),
                                        max_target_seqs=max_target_seqs,
                                        outfmt=5,
                                        evalue=evalue,
                                        qcov_hsp_perc=min_query_cover,
                                        parse_deflines=True)  # returns accession rather than "Query 1"
    print blast_cline
    stdout, stderr = blast_cline()
    if stderr: logging.error(stderr)
    stdout_sio = cStringIO.StringIO(stdout)
    blast_records = NCBIXML.parse(stdout_sio)
    os.remove(query_fasta_file_path)

    retval = set()
    for blast_record in blast_records:
        for alignment in blast_record.alignments:
            if blast_record.query_id == alignment.accession:
                continue

            bit_score_per_residue = alignment.hsps[0].bits / max(blast_record.query_length,
                                                                 len(alignment.hsps[0].sbjct))
            if bit_score_per_residue < effusion.args.min_bit_score_per_residue:
                continue

            retval.add(alignment.accession)
    return retval


def get_annotated_diamond_parents(query_accessions):
    min_query_cover = effusion.args.min_query_cover

    query_accessions = list(query_accessions)  # Ensuring an ordered collection
    blast_db = effusion.args.uniprot_db
    query_fasta_file_path = write_fasta_for_accessions(blast_db, query_accessions)
    # query_fasta_file_path = '/Volumes/SSD/tmp/tmpM6ct6L'
    if os.stat(query_fasta_file_path).st_size == 0:
        logging.warning('Nothing to BLAST, given %d query accessions' % len(query_accessions))
        return {}

    # 2. diamond to get (bit)score (per reside)
    matches_file_path_wo_ext = os.path.join(effusion.args.tmp_dir, uuid.uuid4().hex)
    # matches_file_path_wo_ext = '/Volumes/SSD/tmp/722b50343c944b8b8dc932e789959d60'

    # matches_file_path = matches_file_path_wo_ext + '.gz'
    matches_file_path = matches_file_path_wo_ext

    cmd = (effusion.args.diamond_bin +
           " blastp" +
           " --query %s" +
           " --max-target-seqs 1" +
           " --no-self-hits" +
           " --db %s" +
           " --evalue %e" +
           " --query-cover %d" +
           " --subject-cover %d" +
           " --out %s" +
           " --outfmt 6 qseqid sseqid evalue bitscore qlen slen" +
           " --tmpdir %s"
           # " --compress 1"
           ) % (
              query_fasta_file_path,
              effusion.dataset.goa_diamond_db,
              # remember, this is for calculating parameters of adjacent unannoted proteins,
              # by looking at adjacent annotated proteins
              #  it will only be taking the best one. use the same params as blast search for blastp and for filtering (below)
              effusion.args.evalue,
              min_query_cover,
              min_query_cover,
              matches_file_path_wo_ext,
              effusion.args.tmp_dir)
    if effusion.args.num_cpus is not None:
        cmd += ' --threads %d' % effusion.args.num_cpus
    logging.debug(cmd)
    # logging.debug('not actually running')
    subprocess.check_call(cmd, shell=True)
    # retval = defaultdict(set)
    retval = {}

    with open(matches_file_path, 'r') as f:
        # with gzip.open(matches_file_path, 'rb') as f:
        for line in f:
            split_line = line.strip().split("\t")
            query_uniprot, target_uniprot, evalue, bit_score, qlen, slen = split_line[0], split_line[1], float(
                split_line[2]), float(split_line[3]), int(split_line[4]), int(split_line[5])

            bit_score_per_residue = bit_score / max(qlen, slen)
            if bit_score_per_residue < effusion.args.min_bit_score_per_residue:
                continue

                # if 'diamond' == 'diamond':
                # tr|M9N715|M9N715_ASHG1
                # 17.04: just 'M9N715' for some reason...
                # target_uniprot = target_uniprot.split('|')[1]

            # no self hits anymore
            # if query_uniprot == target_uniprot:
            #     continue

            # retval[query_uniprot].add(target_uniprot)
            retval[query_uniprot] = target_uniprot

    if effusion.args.remove_tmp_files:
        os.remove(query_fasta_file_path)
        os.remove(matches_file_path)

    return retval


def get_annotated_string_parents(query_uniprots):
    min_string_score = effusion.args.min_string_score
    assert effusion.dataset.cutoff_date

    retval = {}

    # max_score_query_1/2 determines the maximum string score for each query
    # query_1/2 picks out the (query,) target, and string score matching this (query, maximum string score) pair
    # *_2 does the query with reversed uniprot1 and uniprot2 to ensure the results from stringdb are symmetrical
    # the query is ordered in ascending order by the string score, so that we can go through all the results, and override the map to store the best target for each query

    # max_score_query_1 = effusion.dataset.database_session.query(effusion.StringLink.uniprot1, func.max(effusion.StringLink.combined_score).label('max_score'))
    # max_score_query_1 = max_score_query_1.filter(effusion.StringLink.combined_score >= min_string_score)
    # max_score_query_1 = max_score_query_1.join(effusion.Annotation, effusion.Annotation.uniprot == effusion.StringLink.uniprot1).filter(effusion.Annotation.date < effusion.dataset.cutoff_date).filter(effusion.Annotation.value == True).filter(effusion.Annotation.source.in_(('direct', 'implied')))
    # max_score_query_1 = max_score_query_1.group_by(effusion.StringLink.uniprot1).subquery()
    # query_1 = effusion.dataset.database_session.query(effusion.StringLink.uniprot1, effusion.StringLink.uniprot2, effusion.StringLink.combined_score).join(max_score_query_1, and_(effusion.StringLink.uniprot1 == effusion.StringLink.uniprot1))
    # query_1 = query_1.filter(effusion.StringLink.combined_score == max_score_query_1.c.max_score)

    # uniprot1 and uniprot2 reversed, to ensure symmetrical results from stringdb
    # todo may be unnecessary or even undesired
    # commenting because i've never seen a case where it wasn't symmetrical...
    # max_score_query_2 = effusion.dataset.database_session.query(effusion.Uniprot.accession, func.max(effusion.StringLink.combined_score).label('max_score'))
    # max_score_query_2 = max_score_query_2.join(effusion.StringLink, effusion.Uniprot.accession == effusion.StringLink.uniprot2).filter(effusion.StringLink.combined_score >= min_string_score)
    # max_score_query_2 = max_score_query_2.join(effusion.Annotation).filter(effusion.Annotation.date < effusion.dataset.cutoff_date).filter(effusion.Annotation.value == True).filter(effusion.Annotation.source.in_(('direct', 'implied')))
    # max_score_query_2 = max_score_query_2.group_by(effusion.Uniprot.accession).subquery()
    # query_2 = effusion.dataset.database_session.query(effusion.StringLink.uniprot2, effusion.StringLink.uniprot1, effusion.StringLink.combined_score).join(max_score_query_2, and_(effusion.StringLink.uniprot2 == max_score_query_2.c.accession))
    # query_2 = query_2.filter(effusion.StringLink.combined_score == max_score_query_2.c.max_score)
    # query = query_1.union(query_2)
    # query = query_1



    annotated_uniprots = effusion.dataset.database_session.query(distinct(effusion.Annotation.uniprot))
    annotated_uniprots = annotated_uniprots.filter(effusion.Annotation.date <= effusion.dataset.cutoff_date).filter(
        effusion.Annotation.value == True).filter(effusion.Annotation.source.in_(('direct', 'implied')))

    max_annotated_string_links = effusion.dataset.database_session.query(effusion.StringLink.uniprot1, func.max(
        effusion.StringLink.combined_score).label('max_score'))
    max_annotated_string_links = max_annotated_string_links.filter(
        effusion.StringLink.combined_score >= min_string_score)
    max_annotated_string_links = max_annotated_string_links.filter(effusion.StringLink.uniprot1.in_(annotated_uniprots))
    max_annotated_string_links = max_annotated_string_links.filter(effusion.StringLink.uniprot2.in_(annotated_uniprots))
    max_annotated_string_links = max_annotated_string_links.group_by(effusion.StringLink.uniprot1)
    max_annotated_string_links = max_annotated_string_links.subquery()

    query = effusion.dataset.database_session.query(effusion.StringLink.uniprot1, effusion.StringLink.uniprot2,
                                                    effusion.StringLink.combined_score)
    query = query.join(max_annotated_string_links,
                       and_(max_annotated_string_links.c.uniprot1 == effusion.StringLink.uniprot1,
                            max_annotated_string_links.c.max_score == effusion.StringLink.combined_score))

    # it's inefficient to sort in order to identify the best target
    print query

    best_score_per_query = {}
    for r in query.all():
        query_uniprot, target_uniprot, combined_score = r
        # the query uniprots is passed into this function is always all uniprots with evidence for this dataset
        # except when we are sampling...
        # assert r[0] in query_uniprots

        if query_uniprot not in query_uniprots:
            continue

        previous_score = best_score_per_query.get(query_uniprot, 0)
        if combined_score > previous_score:
            retval[query_uniprot] = target_uniprot
            best_score_per_query[query_uniprot] = combined_score

    return retval


def query_string(query_uniprots, min_string_score=0, max_string_proteins=sys.maxint):
    session = effusion.dataset.database_session
    all_edges = []
    for query_uniprot in query_uniprots:
        query = session.query(effusion.StringLink).filter(effusion.StringLink.uniprot1 == query_uniprot)
        query = query.filter(effusion.StringLink.combined_score >= min_string_score)
        for row in query:
            # print row
            # {'combined_score': 542L, 'id': 603524730L, 'protein1': u'15368.BRADI3G15040.1', 'protein2': u'15368.BRADI1G75220.1'}
            all_edges.append((query_uniprot, row.uniprot2, row.combined_score))

        # todo second query might be unnecessary if string data is symmetrical
        query = session.query(effusion.StringLink).filter(effusion.StringLink.uniprot2 == query_uniprot)
        query = query.filter(effusion.StringLink.combined_score >= min_string_score)
        for row in query:
            all_edges.append((query_uniprot, row.uniprot1, row.combined_score))

    edges = []
    uniprots = set()
    for e in sorted(all_edges, key=lambda x: x[2], reverse=True):
        edges.append(e)
        uniprots.add(e[1])
        if len(uniprots) >= max_string_proteins:
            break

    return uniprots, edges


def all_by_all(uniprots):
    min_query_cover = effusion.args.min_query_cover
    # evalue = agca.config.blast_params['evalue']


    # 0. get sequences
    seq_records = get_seq_records_for_accessions_batch(effusion.args.uniprot_db, uniprots)

    query_fasta_file = tempfile.NamedTemporaryFile(delete=False, dir=effusion.args.tmp_dir)
    SeqIO.write(seq_records, query_fasta_file, "fasta")
    query_fasta_file.close()
    query_fasta_file_path = query_fasta_file.name

    # 1. make custom database
    dmnd_path_wo_ext = os.path.join(effusion.args.tmp_dir, uuid.uuid4().hex)
    dmnd_path = dmnd_path_wo_ext + '.dmnd'

    cmd = "%s makedb --in %s -d %s" % (effusion.args.diamond_bin, query_fasta_file_path, dmnd_path_wo_ext)
    if effusion.args.num_cpus is not None:
        cmd += ' --threads %d' % effusion.args.num_cpus
    logging.debug(cmd)
    subprocess.check_call(cmd, shell=True)

    # 2. diamond to get (bit)score (per reside)
    matches_file_path_wo_ext = os.path.join(effusion.args.tmp_dir, uuid.uuid4().hex)
    matches_file_path = matches_file_path_wo_ext
    # matches_file_path = matches_file_path_wo_ext + '.gz'

    cmd = (effusion.args.diamond_bin +
           " blastp" +
           " --query %s" +
           " --no-self-hits" +
           " --db %s" +
           " --max-target-seqs %d" +
           " --evalue %e" +
           " --query-cover %d" +
           " --subject-cover %d" +
           " --out %s" +
           " --outfmt 6 qseqid sseqid evalue bitscore qlen slen" +
           " --tmpdir %s"
           # + " --compress 1"
           ) % (
              query_fasta_file_path,
              dmnd_path_wo_ext,
              # these are just heuristics to give us a smaller network that we can use to give us an MST
              1000,
              # max_target_seqs. not the best heuristic. hopefully doesn't affect MST (it sometimes does), but might affect layout
              effusion.args.evalue,
              min_query_cover,
              min_query_cover,  # min_subject_cover
              matches_file_path_wo_ext,
              effusion.args.tmp_dir)
    if effusion.args.num_cpus is not None:
        cmd += ' --threads %d' % effusion.args.num_cpus
    logging.debug(cmd)
    # logging.debug('not actually running')
    subprocess.check_call(cmd, shell=True)

    logging.debug('Filtering targets')
    logging.debug('Caching sequences')

    cached_seq_records = {seq_record.id: seq_record for seq_record in seq_records}

    logging.debug('Making edges')
    # edges = []
    num_edges = 0

    # with gzip.open(matches_file_path, 'rb') as f:
    with open(matches_file_path, 'r') as f:
        for line in f:
            split_line = line.strip().split("\t")
            query_uniprot, target_uniprot, evalue, bit_score, qlen, slen = split_line[0], split_line[1], float(
                split_line[2]), float(split_line[3]), int(split_line[4]), int(split_line[5])

            # no self hits anymore
            # if query_uniprot == target_uniprot:
            #     continue

            query_length = len(cached_seq_records[query_uniprot].seq)
            target_length = len(cached_seq_records[target_uniprot].seq)
            bit_score_per_residue = bit_score / max(query_length, target_length)

            # we're just doing all x all. everything already passes the outer blast filter.
            #  now just take the better ones for MST (and also additional edges for display)
            # for display, top 1000 edges will mess with layout. use some other filter
            # aka hopefully MST edge > this bitscore edge > top 1000
            # 2x min bit score is too high esp. for longer sequences. 1.4 is based on such a network of long seqs
            if bit_score_per_residue < 1.4 * effusion.args.min_bit_score_per_residue:
                continue

            # edges.append((query_uniprot, target_uniprot, bit_score_per_residue, evalue))
            num_edges += 1
            yield (query_uniprot, target_uniprot, bit_score_per_residue, evalue)

    logging.debug('made %d edges' % num_edges)

    # edges might not necessarily pull in distinct, or new, uniprots
    # edges.sort(key=lambda x: x[2], reverse=True)#[0:max_string_proteins]

    if effusion.args.remove_tmp_files:
        os.remove(query_fasta_file_path)
        os.remove(dmnd_path)
        os.remove(matches_file_path)

        # return edges
